#include <project.h>

#include "audit_queue.h"

static void __x_audit_queue_clear(audit_q_t *queue)
{
    LIST_HEAD *entry;

    while (!__list_is_empty(&queue->audit_list)) {
        audit_comm_t *cm;

        __list_remove_tail(&queue->audit_list, entry);

        cm = OBJECT_PTR(entry, audit_comm_t, head);
        if (cm->Destroy) {
            cm->Destroy(cm);
        }
    }
    return;
}

void audit_q_clear(audit_q_t *queue)
{
    spin_lock_bh(&queue->lock);

    __x_audit_queue_clear(queue);
    atomic_set(&queue->count, 0);

    spin_unlock_bh(&queue->lock);

    return;
}

void audit_q_destroy(audit_q_t *queue)
{
    __x_audit_queue_clear(queue);
    atomic_set(&queue->count, 0);
    return;
}

int audit_q_init(audit_q_t *queue, int limit)
{
    memset(queue, 0, sizeof(audit_q_t));
    __list_init(&queue->audit_list);
    spin_lock_init(&queue->lock);
    atomic_set(&queue->count, 0);
    queue->limit = limit;
    return 0;
}

static audit_comm_t *__audit_queue_shift(audit_q_t *queue,
                                           unsigned int maxLen, BOOLEAN bAbort)
{
    audit_comm_t *retval = NULL;
    LIST_HEAD *entry;
    AUDIT_HDR * hdr;

    spin_lock_bh(&queue->lock);

    while (!__list_is_empty(&queue->audit_list)) {
        audit_comm_t *comm;
        
        tag_pr_debug("queue[%lx], &queue->audit_list[%lx]\n", queue, (unsigned int *)&queue->audit_list);

        __list_remove_head(&queue->audit_list, entry);

        atomic_dec_return(&queue->count);

        comm = OBJECT_PTR(entry, audit_comm_t, head);

        hdr = (AUDIT_HDR *)(((unsigned char *)comm) + sizeof(audit_comm_t));

        if (hdr->length  <= maxLen) {
            retval = comm;
        } else {
            if (bAbort) {
                __destroy_audit_common(comm);
                continue;
            } else {
                __list_add_head(&queue->audit_list, entry);
                atomic_inc_return(&queue->count);
                retval = NULL;
                goto DONE;
            }
        }
    }

DONE:
    spin_unlock_bh(&queue->lock);
    return retval;
}

void audit_q_push(audit_q_t *queue, audit_comm_t *common)
{
    int flag = 0;

    spin_lock_bh(&queue->lock);
    if (atomic_read(&queue->count) < queue->limit) {
        __list_add_tail(&queue->audit_list, &common->head);
        atomic_inc_return(&queue->count);
        flag = 1;
    }
    spin_unlock_bh(&queue->lock);

    if (0 == flag) {
        __destroy_audit_common(common);
    }
    return;
}

int audit_q_read(audit_q_t *queue, IOCTL_PARAM *param)
{
    int nt_status = 0;
    int length = 0;
    int count = 0;
    long outputBufferLength = 0;

    unsigned char *outPtr = NULL;
    long left;

    outputBufferLength = param->outputBufferLength;

    outPtr = (unsigned char *)param->outputBuffer;
    if (!outPtr) {
        tag_pr_info("<audit-queue> error [outputBuffer empty]\n");
        goto DONE;
    }

    left = outputBufferLength;
    // tag_pr_debug("left[%ld]\n", left);

    while (left > 0) {
        audit_comm_t *common;
        AUDIT_HDR *hdr;

        common = __audit_queue_shift(queue, left, 0 == count);
        if (!common)
            break;

        count++;

        hdr = (AUDIT_HDR *)(((unsigned char *)common) + sizeof(audit_comm_t));

        if (0 != copy_to_user(outPtr, hdr, hdr->length)) {
            __destroy_audit_common(common);
            break;
        }

        left -= hdr->length;
        outPtr += hdr->length;
        length += hdr->length;

        __destroy_audit_common(common);
    }

DONE:
    param->uRetValue = nt_status;
    param->outputBufferLength = length;

    return nt_status;
}

int audit_q_read_buf(audit_q_t *queue, char __user *ubuf)
{
    int count = 0;
    long left;
    int length = 0;

    left = PAGE_SIZE;

    while (left > 0) {
        audit_comm_t *common;
        AUDIT_HDR *hdr;

        common = __audit_queue_shift(queue, left, 0 == count);
        if (!common)
            break;

        count++;

        hdr = (AUDIT_HDR *)(((unsigned char *)common) + sizeof(audit_comm_t));

        if (0 != copy_to_user(ubuf, hdr, hdr->length)) {
            __destroy_audit_common(common);
            break;
        }

        left -= hdr->length;
        length += hdr->length;

        __destroy_audit_common(common);
    }

    return length;
}
