#include "sec_audit.h"


static void __audit_destroy_ope_item(audit_comm_t *cm)
{
    audit_ope_item_t *item = OBJECT_PTR(cm, audit_ope_item_t, common);

    kfree(item);
    return;
}

static void __audit_destroy_nac_item(audit_comm_t *cm)
{
    audit_nac_item_t *item = OBJECT_PTR(cm, audit_nac_item_t, common);

    kfree(item);
    return;
}

static void __audit_destroy_pac_item(audit_comm_t *cm)
{
    audit_pac_item_t *item = OBJECT_PTR(cm, audit_pac_item_t, common);

    kfree(item);
    return;
}

audit_comm_t *audit_q_item_new(int audit_type, int sub_type)
{
    audit_comm_t *retval = NULL;
    audit_comm_t *comm = NULL;
    
    AUDIT_HDR * hdr = NULL;
    
    int total_len = 0;
    int data_len = 0;
    audit_comm_t * content = NULL;
    int magic = 0;
    void * func_p = NULL;

    // 根据不同的审计类型，配置审计数据
    switch (audit_type) {
        case AUDIT_TYPE_IMA:
            // todo： 此处为静态度量
            // todo： 还需要动态度量
            total_len = sizeof(audit_ope_item_t);
            data_len = sizeof(audit_ope_t);
            magic = AUDIT_MAGIC;
            func_p = __audit_destroy_ope_item;
            break;
        case AUDIT_TYPE_PROTECT:
            total_len = sizeof(audit_ope_item_t);
            data_len = sizeof(audit_ope_t);
            magic = AUDIT_MAGIC;
            func_p = __audit_destroy_ope_item;
            break;
        case AUDIT_TYPE_AC:
            if (sub_type == AUDIT_SUBTYPE_MAC) {
                total_len = sizeof(audit_ope_item_t);
                data_len = sizeof(audit_ope_t);
                magic = AUDIT_MAGIC;
                func_p = __audit_destroy_ope_item;
            } else if (sub_type == AUDIT_SUBTYPE_NAC) {
                total_len = sizeof(audit_nac_item_t);
                data_len = sizeof(nac_audit_data_t);
                magic = AUDIT_MAGIC;
                func_p = __audit_destroy_nac_item;                
            } else if (sub_type == AUDIT_SUBTYPE_PAC) {
                total_len = sizeof(audit_pac_item_t);
                data_len = sizeof(pac_audit_data_t);
                magic = AUDIT_MAGIC;
                func_p = __audit_destroy_pac_item;                
            }
            break;
        default:
            return NULL;
    }

    comm = (audit_comm_t *)kmalloc(total_len, GFP_ATOMIC);
    if (!comm)
        return NULL;

    memset(comm, 0, total_len);

    hdr = (void*)(((unsigned char *)comm) + sizeof(audit_comm_t));
    content = comm + sizeof(*comm);
    hdr->magic = magic;
    hdr->hdr_type = AUDIT_HDR_TYPE_OPE;  // todo: 目前想都是用这一种,不知道是否足够
    hdr->length = data_len;
    hdr->uid = (unsigned int)get_current_uid();

    comm->Destroy = func_p;

    retval = comm;

    return retval;
}

BOOLEAN audit_write_file_ope(audit_q_t * audit_q, unsigned int *sid, const char *proc_path,
                                     long processID, unsigned int ope,
                                     unsigned int sub_type, BOOLEAN bHoldBack,
                                     unsigned int ruleType, unsigned int ruleId,
                                     unsigned int level, const char *target,
                                     unsigned char *sum, int sub_rank,
                                     int obj_rank)
{
    BOOLEAN retval = FALSE;

    audit_comm_t * comm = NULL;
    audit_ope_item_t *item = NULL;
    unsigned int maxCount = 0;

    tag_pr_debug("ruleType[%d] sub_type[%d]\n", ruleType, sub_type);

    comm = audit_q_item_new(ruleType, sub_type);
    if (!comm)
        goto DONE;

    item = (audit_ope_item_t *)comm;

    set_audit_time(&item->content.hdr.auditTime);
    item->content.hdr.result = bHoldBack;

    if (proc_path != PROCESSPATH_NULL) {
        maxCount = sizeof(item->content.process) / sizeof(char);
        strncpy(item->content.process, proc_path, maxCount);
        item->content.process[maxCount - 1] = 0;
    }

    item->content.processId = (ULONGLONG)processID;
    item->content.level = level;
    item->content.ruleId = ruleId;
    item->content.ruleType = ruleType;

    if (sid != SID_NULL) {
        item->content.hdr.uid = *sid;
    }

    item->content.sub_rank = (unsigned char)sub_rank;
    item->content.obj_rank = (unsigned char)obj_rank;

    if (target != TARGET_NULL) {
        maxCount = sizeof(item->content.target);
        strncpy(item->content.target, target, maxCount);
        item->content.target[maxCount - 1] = 0;
    }

    item->content.op_class = ope;
    item->content.sub_type = sub_type;

    if (sum != SUM_NULL) {
        memcpy(item->content.sum, sum, sizeof(item->content.sum));
    } else {
        memset(item->content.sum, 0, sizeof(item->content.sum));
    }

    audit_q_push(audit_q, &item->common);

    audit_ope_show(&item->content);

    retval = TRUE;
DONE:
    return retval;
}

BOOLEAN audit_write_nac(audit_q_t * audit_q, unsigned int srcIp, unsigned short srcPort,
                                unsigned int destIp, unsigned short destPort,
                                unsigned char protocol, char direction,
                                char action)
{
    BOOLEAN retval = FALSE;

    audit_comm_t *comm = NULL;
    nac_audit_data_t *data;

    comm = audit_q_item_new(AUDIT_TYPE_AC, AUDIT_SUBTYPE_NAC);
    if (!comm)
        goto DONE;

    data = (nac_audit_data_t *)(comm + 1);
    data->srcIp = srcIp;
    data->srcPort = srcPort;
    data->destIp = destIp;
    data->destPort = destPort;
    data->protocol = protocol;
    data->direction = direction;
    data->action = action;
    data->ruleType = AUDIT_TYPE_AC;

    data->hdr.magic = AUDIT_MAGIC;
    data->hdr.hdr_type = AUDIT_HDR_TYPE_NET;
    data->hdr.length = sizeof(nac_audit_data_t);

    set_audit_time(&data->hdr.auditTime);

    audit_q_push(audit_q, comm);

    audit_nac_show(data);

    retval = TRUE;
DONE:
    return retval;
}


BOOLEAN audit_write_pac(audit_q_t * audit_q, unsigned int dev_type, unsigned int pid, unsigned short vid,
                                unsigned int clsid, unsigned char action)
{
    BOOLEAN retval = FALSE;

    audit_comm_t *comm = NULL;
    pac_audit_data_t *data;

    comm = audit_q_item_new(AUDIT_TYPE_AC, AUDIT_SUBTYPE_PAC);
    if (!comm)
        goto DONE;

    data = (pac_audit_data_t *)(comm + 1);

    data->audit_type = AUDIT_TYPE_AC;
    data->sub_type = AUDIT_SUBTYPE_PAC;
    data->dev_type = dev_type;
    data->pid = pid;
    data->vid = vid;
    data->clsid = clsid;
    data->action = action;

    data->hdr.magic = AUDIT_MAGIC;
    data->hdr.hdr_type = AUDIT_HDR_TYPE_PAC;
    data->hdr.length = sizeof(pac_audit_data_t);

    set_audit_time(&data->hdr.auditTime);

    audit_q_push(audit_q, comm);

    audit_pac_show(data);

    retval = TRUE;
DONE:
    return retval;
}

int audit_read_ioctl(IOCTL_PARAM *param, audit_q_t * audit_q)
{
    int retval = 0;
        
    retval = audit_q_read(audit_q, param);

    return retval;
}

