#ifndef __SEC_AUDIT_H_
#define __SEC_AUDIT_H_

#include <project.h>
#include "audit_comm.h"
#include "common/general/include/mod_share.h"
#include "common/ioctrl_param.h"
#include "audit_queue.h"
#include "common/general/include/audit_mod.h"

#include <linux/sched.h>
#include <linux/cred.h>

static inline uid_t get_current_uid(void) 
{
    const struct cred *cred;
    uid_t uid;

    cred = current_cred();
    uid = cred->uid.val;

    return uid;
}


// NULL宏，用于在填充审计项时，表示无此项
// notice: 在处理审计时（显示），对于这些NULL项应忽略
#define PROCESSPATH_NULL NULL
#define PROCESSID_NULL 0
#define SID_NULL NULL
#define RULETYPE_NULL 0
#define RULLID_NULL 0
#define LEVEL_NULL 0
#define SUM_NULL NULL
#define SUBRANK_NULL 0
#define OBJRANK_NULL 0
#define TARGET_NULL NULL
#define OPE_NULL 0
#define SUBTYPE_NULL 0

// ==============================================================
#pragma pack(1)

// 适用于文件操作类型的审计
// 主要采用  主体 操作 客体 方式
typedef struct __audit_ope_item {
    audit_comm_t common;
    audit_ope_t content;
} audit_ope_item_t;

// 网络访问控制审计
typedef struct __audit_nac_item {
    audit_comm_t common;
    nac_audit_data_t content;
} audit_nac_item_t;

// 外设访问控制审计
typedef struct __audit_pac_item {
    audit_comm_t common;
    pac_audit_data_t content;
} audit_pac_item_t;


#pragma pack()
// ==============================================================

extern audit_q_t g_audit_mac;
extern audit_q_t g_audit_prot;
extern audit_q_t g_audit_ima;
extern audit_q_t g_audit_nac;
extern audit_q_t g_audit_pac;

static inline void audit_hdr_show(AUDIT_HDR * item)
{
    tag_pr_debug("magic[%d], length[%d], hdr_type[%d], result[%d], uid[%d]\n",
     item->magic, item->length, item->hdr_type, item->result, item->uid);
}

static inline void audit_ope_show(audit_ope_t * item)
{
    audit_hdr_show(&item->hdr);
    tag_pr_debug("rule_type[%s], rule_id[%d], level[%d]\n", audit_type_str(item->ruleType), item->ruleId, item->level);
    tag_pr_debug("process[%s], processId[%ld]\n", item->process, item->processId);
    tag_pr_debug("op_class[%s], sub_type[%s]\n", audit_ope_str(item->op_class), audit_subtype_str(item->sub_type));
    tag_pr_debug("sub_rank[%d], obj_rank[%d]\n", item->sub_rank, item->obj_rank);
    tag_pr_debug("target[%s], sum[%s]\n", item->target, item->sum);
}

static inline void audit_nac_show(nac_audit_data_t * item)
{
    audit_hdr_show(&item->hdr);
    tag_pr_debug("srcIp[%d], srcPort[%d]\n", item->srcIp, item->srcPort);
    tag_pr_debug("destIp[%d], destPort[%d]\n", item->destIp, item->destPort);
    tag_pr_debug("protocol[%d], sub_type[%d], direction[%d]\n", item->protocol, item->action, item->direction);
    tag_pr_debug("ruleId[%d], ruleType[%d]\n", item->ruleId, item->ruleType);
}

static inline void audit_pac_show(pac_audit_data_t * item)
{
    audit_hdr_show(&item->hdr);
    tag_pr_debug("pid[%d], vid[%d]\n", item->pid, item->vid);
    tag_pr_debug("clsid[%d]\n", item->clsid);
}

// 读内核审计 到 应用空间
int audit_read_ioctl(IOCTL_PARAM *param, audit_q_t * audit_q);

BOOLEAN audit_write_file_ope( audit_q_t * audit_q, unsigned int *sid, const char *proc_path,
                                     long processID, unsigned int ope,
                                     unsigned int sub_type, BOOLEAN result,
                                     unsigned int ruleType, unsigned int ruleId,
                                     unsigned int level, const char *target,
                                     unsigned char *sum, int sub_rank,
                                     int obj_rank);

BOOLEAN audit_write_nac(audit_q_t * audit_q, unsigned int srcIp, unsigned short srcPort,
                                unsigned int destIp, unsigned short destPort,
                                unsigned char protocol, char direction,
                                char action);

BOOLEAN audit_write_pac(audit_q_t * audit_q, unsigned int dev_type, unsigned int pid, unsigned short vid,
                                unsigned int clsid, unsigned char action);


// 系统保护  审计
#define AUDIT_SYS_PROTECT(proc_path, processID, ope, sub_type, \
                              result, target) \
do { \
    audit_write_file_ope(&g_audit_prot, SID_NULL, proc_path, processID, ope, \
                                 sub_type, result, AUDIT_TYPE_PROTECT,    \
                                 RULLID_NULL, LEVEL_NULL, target, SUM_NULL, \
                                 SUBRANK_NULL, OBJRANK_NULL);    \
} while(0);

// 静态度量 审计 
#define AUDIT_STATIC_MEASURE(proc_path, processID, ope,  \
                              result, target, sum) \
do { \
    audit_write_file_ope(&g_audit_ima, SID_NULL, proc_path, processID, ope, \
                                 AUDIT_SUBTYPE_STATIC_MEASURE, result, AUDIT_TYPE_IMA,    \
                                 RULLID_NULL, LEVEL_NULL, target, sum, \
                                 SUBRANK_NULL, OBJRANK_NULL);    \
} while(0);

// 强访 操作审计 
#define AUDIT_FILE_MAC(proc_path, processID, ope,  \
                              result, target, sub_rank, obj_rank) \
do { \
    audit_write_file_ope(&g_audit_mac, SID_NULL, proc_path, processID, ope, \
                                 AUDIT_SUBTYPE_MAC, result, AUDIT_TYPE_AC,    \
                                 RULLID_NULL, LEVEL_NULL, target, SUM_NULL, \
                                 sub_rank, obj_rank);    \
} while(0);

// todo: 动态度量 审计



// 可以让每个模块都是同名的变量，然后，每个模块往自己的队列中插入审计
// extern audit_q_t g_mod_audit;

#endif 