#include <project.h>

#include <linux/binfmts.h>
#include <linux/elf.h>
#include <linux/file.h>
#include <linux/mman.h>
#include <linux/module.h>
#include <linux/mount.h>
#include <linux/security.h>
#include <linux/slab.h>
#include <linux/version.h>

#include "common/version_type.h"
#include "common/file_ops/file_ops.h"
#include "common/general/elf_header.h"

#include "common/elf_assist.h"
#include "sec_krnl_misc.h"

// 文件类型字段的描述字符串数组
// static char *file_type[] = {"Unknown", "Relocatable", "Executable", "Shared object", "Core imgae"};

#define le16toh le16_to_cpu
#define le32toh le32_to_cpu
#define le64toh le64_to_cpu

#define be16toh be16_to_cpu
#define be32toh be32_to_cpu
#define be64toh be64_to_cpu

// 16位unsigned short int
unsigned short int uxtoh16(unsigned short int d, unsigned char bLeByteOrder)
{
    if (bLeByteOrder) {
        return be16toh(d);
    } else {
        return le16toh(d);
    }
}

// 32位unsigned int
unsigned int uxtoh32(unsigned int d, unsigned char bLeByteOrder)
{
    if (bLeByteOrder) {
        return be32toh(d);
    } else {
        return le32toh(d);
    }
}

// 32位int
int xtoh32(int d, unsigned char bLeByteOrder)
{
    if (bLeByteOrder) {
        return be32toh(d);
    } else {
        return le32toh(d);
    }
}

// 64位unsigned int
unsigned long long uxtoh64(unsigned long long d, unsigned char bLeByteOrder)
{
    if (bLeByteOrder) {
        return be64toh(d);
    } else {
        return le64toh(d);
    }
}

// 64位int
unsigned long long xtoh64(long long d, unsigned char bLeByteOrder)
{
    if (bLeByteOrder) {
        return be64toh(d);
    } else {
        return le64toh(d);
    }
}

static ssize_t __kread_ops(struct file *file, loff_t off, void *rbuf,
                           size_t count)
{
    int rbuf_len = 0;
    loff_t pos = off;
    rbuf_len = vfs_filp_read(file, (char *)rbuf, count, &pos);
    return rbuf_len;
}

ssize_t __kread(struct file *file, loff_t off, void *rbuf, size_t count,
                unsigned char bOPS)
{
    if (bOPS) {
        return __kread_ops(file, off, rbuf, count);
    } else {
        return file_ops_read(file, rbuf, count, &off);
    }
}

static int readELFHeader(elf_header *ehdr, struct file *file)
{
    int retval = -1;
    loff_t pos = 0;

    memset(ehdr, 0, sizeof(elf_header));

    if (sizeof(elf_header) !=
        file_ops_read(file, (char *)ehdr, sizeof(elf_header), &pos)) {
        goto DONE;
    }

    // -- if(strncmp(ehdr->magic, MAGIC, strlen(MAGIC)) != 0){
    if (memcmp(ehdr->magic, MAGIC, strlen(MAGIC)) != 0) {
        retval = FORMAT_ERROR;
        goto DONE;
    }

    retval = 0;
DONE:
    return retval;
}

int is_elf_file(struct file *file)
{
    int retval = 0;
    elf_header hdr;
    loff_t i_size;

    if (!get_file_dentry(file))
        goto DONE;

    i_size = i_size_read(get_file_dentry(file)->d_inode);
    if (i_size < sizeof(elf_header))
        goto DONE;

    if (0 != readELFHeader(&hdr, file))
        goto DONE;

    if ((hdr.addr_width != 1) && (hdr.addr_width != 2)) {
        // -- maybe 64bit (2)
        goto DONE;
    }
    // -- if(hdr.file_type != 2 && hdr.file_type != 3){
    // -	*.ko 涓?*.o 閮芥槸 1锛屾棤娉曞垎杈?
    // -	goto DONE;
    // -- }
    retval = 1;
DONE:
    return retval;
}

// -- 获取elf 位数、字节序
// -- 处理32位ELF文件的方式
static int read_elf32_object_file(struct file *file, unsigned char bOPS,
                                  unsigned char bLeByteOrder,
                                  Elf32_Ehdr *elf_head, BINARY_INFO *info)
{
    int retval = -1;
    Elf32_Shdr *shdr = NULL;
    // -- Elf32_Dyn dyn;
    // -- char libName[255];
    int i = 0, flag_3 = 0;
    unsigned int dynstrOff = 0;
    unsigned short _e_type = 0, _e_shnum = 0;
    unsigned int _e_shoff = 0, _sh_size = 0;
    // -- int _d_tag = 0;

    _e_type = uxtoh16(elf_head->e_type, bLeByteOrder);
    _e_shnum = uxtoh16(elf_head->e_shnum, bLeByteOrder);
    _e_shoff = uxtoh32(elf_head->e_shoff, bLeByteOrder);
    _sh_size = sizeof(Elf32_Shdr) * _e_shnum;

    shdr = (Elf32_Shdr *)kmalloc(_sh_size, GFP_KERNEL);
    if (!shdr) {
        printk("[error]read64Elf malloc fail\n");
        goto DONE;
    }

    // -- 将节头表读入shdr
    // -- _e_shoff
    if (_sh_size != __kread(file, _e_shoff, (char *)shdr, _sh_size, bOPS)) {
        printk("[error]Read Section Head ERROR!\n");
        goto DONE;
    }

    flag_3 = 0;
    for (i = 0; i < _e_shnum; i++) {
        shdr[i].sh_name = uxtoh32(shdr[i].sh_name, bLeByteOrder);
        shdr[i].sh_type = uxtoh32(shdr[i].sh_type, bLeByteOrder);
        shdr[i].sh_offset = uxtoh32(shdr[i].sh_offset, bLeByteOrder);
        if (shdr[i].sh_type == 3 && 0 == flag_3) {
            dynstrOff = shdr[i].sh_offset;  //这是 .dynstr 节的偏移量
            flag_3++;
        }
    }

    for (i = 0; i < _e_shnum; i++) {
        // -- 读取.dynamic节中下一个数组元素的内容，存在dyn所指的空间中
        if (1 == shdr[i].sh_type) {
            char buff[sizeof(".modinfo") + 5];
            int offset = dynstrOff + shdr[i].sh_name;

            if (sizeof(buff) !=
                __kread(file, offset, (char *)buff, sizeof(buff) - 1, bOPS)) {
                printk("Read .dynamic file ERROR!\n");
                goto DONE;
            }

            buff[sizeof(buff) - 1] = 0;
            if (0 == strncmp(buff, ".modinfo", sizeof(buff))) {
                info->is_kern_driver = 1;
                break;
            }
        }
    }

    retval = 0;
DONE:
    if (shdr)
        kfree(shdr);
    return retval;
}

// -- 处理64位ELF文件的方式
static int read_elf64_object_file(struct file *file, unsigned char bOPS,
                                  unsigned char bLeByteOrder,
                                  Elf64_Ehdr *elf_head, BINARY_INFO *info)
{
    int retval = -1;
    Elf64_Shdr *shdr = NULL;
    // -- Elf64_Phdr * phdr = NULL;
    // -- Elf64_Dyn dyn;
    // -- char libName[255];
    int i = 0, flag_3 = 0;
    unsigned int dynstrOff = 0;
    unsigned short _e_type = 0, _e_shnum = 0;
    unsigned long int _e_shoff = 0, _sh_size = 0;
    // -- long int _d_tag = 0;

    _e_type = uxtoh16(elf_head->e_type, bLeByteOrder);
    _e_shnum = uxtoh16(elf_head->e_shnum, bLeByteOrder);
    _e_shoff = uxtoh64(elf_head->e_shoff, bLeByteOrder);
    _sh_size = sizeof(Elf64_Shdr) * _e_shnum;

    // -- 为shdr指针动态申请相应大小的空间
    shdr = (Elf64_Shdr *)kmalloc(_sh_size, GFP_KERNEL);
    if (!shdr) {
        printk("[error]read64Elf malloc fail\n");
        goto DONE;
    }

    // -- 将节头表读入shdr
    if (_sh_size != __kread(file, _e_shoff, (char *)shdr, _sh_size, bOPS)) {
        printk("[error]Read Section Head ERROR!\n");
        goto DONE;
    }

    // -- 现在shdr指向了节头表
    // -- 找到 .dynstr 节，读出它的偏移量存在dynstrOff变量中
    flag_3 = 0;
    for (i = 0; i < _e_shnum; i++) {
        shdr[i].sh_name = uxtoh32(shdr[i].sh_name, bLeByteOrder);
        shdr[i].sh_type = uxtoh32(shdr[i].sh_type, bLeByteOrder);
        shdr[i].sh_offset = uxtoh64(shdr[i].sh_offset, bLeByteOrder);
        if (shdr[i].sh_type == 3 && 0 == flag_3) {
            dynstrOff = shdr[i].sh_offset;  //这是 .dynstr 节的偏移量
            flag_3++;
        }
    }

    for (i = 0; i < _e_shnum; i++) {
        //读取.dynamic节中下一个数组元素的内容，存在dyn所指的空间中

        if (1 == shdr[i].sh_type) {
            char buff[sizeof(".modinfo") + 5];
            int offset = dynstrOff + shdr[i].sh_name;

            if (sizeof(buff) !=
                __kread(file, offset, (char *)buff, sizeof(buff) - 1, bOPS)) {
                printk("Read .dynamic file ERROR!\n");
                goto DONE;
            }

            buff[sizeof(buff) - 1] = 0;
            if (0 == strncmp(buff, ".modinfo", sizeof(buff))) {
                info->is_kern_driver = 1;
                break;
            }
        }
    }

    retval = 0;
DONE:
    if (shdr)
        kfree(shdr);
    return retval;
}

int is_bin_file(struct file *file, unsigned char bOPS)
{
    int retval = 0;
    BINARY_INFO info;

    memset(&info, 0, sizeof(info));

    if (0 != get_elf_info(file, bOPS, 0, &info)) {
        goto DONE;
    }

    if (2 == info.file_type) {
        retval = 1;
    }
DONE:
    return retval;
}

int is_ko_file(struct file *file, unsigned char bOPS)
{
    int retval = 0;
    BINARY_INFO info;

    memset(&info, 0, sizeof(info));

    if (0 != get_elf_info(file, bOPS, 1, &info)) {
        goto DONE;
    }

    if (1 == info.file_type && info.is_kern_driver) {
        retval = 1;
    }
DONE:
    return retval;
}

int is_so_file(struct file *file, unsigned char bOPS)
{
    int retval = 0;
    BINARY_INFO info;

    memset(&info, 0, sizeof(info));

    if (0 != get_elf_info(file, bOPS, 0, &info)) {
        goto DONE;
    }

    if (3 == info.file_type) {
        retval = 1;
    }
DONE:
    return retval;
}

int is_script_file_by_head(struct file *file, unsigned char bOPS)
{
    int retval = 0;
    BINARY_INFO info;

    memset(&info, 0, sizeof(info));

    if (0 != get_elf_info(file, bOPS, 0, &info)) {
        goto DONE;
    }

    if (4 == info.file_type) {
        retval = 1;
    }
DONE:
    return retval;
}

int get_elf_info(struct file *file, unsigned char bOPS, unsigned char bReadKO,
                 BINARY_INFO *elfInfo)
{
    int retval = FORMAT_ERROR;
    elf_header hdr;
    loff_t i_size;
    unsigned char bLeByteOrder = 0;
    char * buf = NULL;
    int hdr_left = sizeof(elf_header);

    if (!get_file_dentry(file))
        goto DONE;

    i_size = i_size_read(get_file_dentry(file)->d_inode);

    // 脚本头判断 -----------------
    if (i_size <= 4)
        goto DONE;
    buf = (char *)&hdr;
    if (4 != __kread(file, 0, buf, 4, bOPS))
        goto DONE;

    if ((buf[0] == '#' && buf[1] == '!' && buf[2] == '/') 
        || (buf[0] == '#' && buf[1] == '!' && buf[2] == ' ' && buf[2] == '/')) {
            retval = 0;
            elfInfo->file_type = 4;
            goto DONE;
    }   
    hdr_left -= 4; 
    // -----------------

    if (i_size < sizeof(elf_header))
        goto DONE;

    if (hdr_left != __kread(file, 0, (char *)&hdr, hdr_left , bOPS)) {
        goto DONE;
    }

    if (memcmp(hdr.magic, MAGIC, strlen(MAGIC)) != 0) {
        retval = FORMAT_ERROR;
        goto DONE;
    }

    elfInfo->bits = hdr.addr_width;       // -- ch[4];
    elfInfo->byte_order = hdr.byteorder;  // -- ch[5];
    if (1 == elfInfo->bits) {
        Elf32_Ehdr elf_head;
        
        if (elfInfo->byte_order == 2) {
            bLeByteOrder = 1;  // -- 大端
        }

        if (sizeof(Elf32_Ehdr) !=
            __kread(file, 0, (char *)&elf_head, sizeof(Elf32_Ehdr), bOPS)) {
            goto DONE;
        }

        elfInfo->file_type = uxtoh16(elf_head.e_type, bLeByteOrder);
        elfInfo->architecture = uxtoh16(elf_head.e_machine, bLeByteOrder);

        if (1 == elfInfo->file_type) {
            if (bReadKO) {
                retval = read_elf32_object_file(file, bOPS, bLeByteOrder,
                                                &elf_head, elfInfo);
                if (-1 == retval) {
                    printk("read 32 bit ELF file ERROR!\n");
                }
            }
        }
    } else if (2 == elfInfo->bits) {
        Elf64_Ehdr elf_head;

        if (elfInfo->byte_order == 2) {
            bLeByteOrder = 1;
        }

        if (sizeof(Elf64_Ehdr) !=
            __kread(file, 0, (char *)&elf_head, sizeof(Elf64_Ehdr), bOPS)) {
            goto DONE;
        }

        elfInfo->file_type = uxtoh16(elf_head.e_type, bLeByteOrder);
        elfInfo->architecture = uxtoh16(elf_head.e_machine, bLeByteOrder);
        if (1 == elfInfo->file_type) {
            if (bReadKO) {
                retval = read_elf64_object_file(file, bOPS, bLeByteOrder,
                                                &elf_head, elfInfo);
                if (-1 == retval) {
                    printk("read 32 bit ELF file ERROR!\n");
                }
            }
        }
    } else {
        goto DONE;
    }

    if (1 != elfInfo->file_type) {
        retval = 0;
        goto DONE;
    }

    retval = 0;
DONE:
    return retval;
}

static int readElfHeaderWithOps(elf_header *ehdr, struct file *file)
{
    int retval = -1;
    fmode_t f_mode = file->f_mode;

    loff_t pos = 0;
    int rbuf_len = 0;
    size_t count = 0;

    char *rbuf = kzalloc(PAGE_SIZE, GFP_KERNEL);  // -- __CALC_SIZE

    // -- file->f_mode = file->f_mode | FMODE_READ | FMODE_CAN_READ;

    /*
    if (!(file->f_mode & FMODE_READ))
        return -EBADF;
    if (!(file->f_mode & FMODE_CAN_READ))
        return -EINVAL;
    */
    if (!rbuf)
        goto DONE;

    count = PAGE_SIZE;

    // 这里不知道为什么用fp接口，会导致启动bash死机，暂时换成和读文件一样的接口尝试
    // rbuf_len = vfs_filp_read(file, (char *)rbuf, count, &pos);
    rbuf_len = __kread(file, pos, rbuf, count, 0);
    if (rbuf_len < sizeof(elf_header)) {
        goto DONE;
    }
    memset(ehdr, 0, sizeof(elf_header));

    if (sizeof(elf_header) > rbuf_len) {
        goto DONE;
    }

    memcpy(ehdr, rbuf, sizeof(elf_header));

    if (memcmp(ehdr->magic, MAGIC, strlen(MAGIC)) != 0) {
        retval = FORMAT_ERROR;
        goto DONE;
    }
    retval = 0;
DONE:
    file->f_mode = f_mode;
    if (rbuf) 
        kfree(rbuf);
    return retval;
}

int is_elf_file_ex(struct file *file)
{
    int retval = 0;
    elf_header hdr;
    loff_t i_size;

    i_size = i_size_read(get_file_dentry(file)->d_inode);
    if (i_size < sizeof(elf_header)) {
        goto DONE;
    }

    if (0 != readElfHeaderWithOps(&hdr, file))
        goto DONE;

    if ((hdr.addr_width != 1) && (hdr.addr_width != 2)) {
        // -- maybe 64bit (2)
        goto DONE;
    }
    // -- if(hdr.file_type != 2 && hdr.file_type != 3){
    // -	*.ko 涓?*.o 閮芥槸 1锛屾棤娉曞垎杈?
    // -	goto DONE;
    // -- }
    retval = 1;
DONE:
    return retval;
}
