#include <linux/uio.h>
#include "common/version_type.h"
#include "file_ops.h"

struct file *file_ops_open(const char *filename, int flags, int mode)
{
    return filp_open(filename, flags, mode);
}

int file_ops_close(struct file *filp, fl_owner_t id)
{
    return filp_close(filp, id);
}

loff_t file_ops_get_len(struct file *file)
{
    return i_size_read(get_file_dentry(file)->d_inode);
}

static ssize_t vfs_write_no_check(struct file *file, const char __user *buf,
                                  size_t count, loff_t *pos)
{
    ssize_t ret;
    if (!(file->f_mode & FMODE_WRITE))
        return -EBADF;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 2, 0))
    if (!file->f_op || (!file->f_op->write && !file->f_op->aio_write))
        return -EINVAL;
    if (file->f_op->write)
        ret = file->f_op->write(file, buf, count, pos);
    else
        ret = do_sync_write(file, buf, count, pos);

#elif (LINUX_VERSION_CODE > KERNEL_VERSION(4, 19, 0))

#if defined(ARM64)
    if (unlikely(!access_ok(buf, count)))
        return -EFAULT;
    if (count > MAX_RW_COUNT)
        count = MAX_RW_COUNT;
    ret = vfs_write(file, (__force const char __user *)buf, count, pos);

#else  // 龙芯V10服务器，不需要调用此函数
    ret = 0;
#endif

#else

    /* The cast to a user pointer is valid due to tp set_fs() */
    //   if (!(file->f_mode & FMODE_CAN_WRITE))
    //       return -EINVAL;
    if (unlikely(!access_ok(VERIFY_READ, buf, count)))
        return -EFAULT;

    if (count > MAX_RW_COUNT)
        count = MAX_RW_COUNT;
    ret = __vfs_write_n(file, (const char __user *)buf, count, pos);
#endif
    return ret;
}

size_t file_ops_write(struct file *filp, char *buf, int len)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 10, 0)
    mm_segment_t old_fs;
#endif
    ssize_t ret;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0)  // && defined (MIPS64EL)
    ret = kernel_write(filp, (const char __user *)buf, len, &filp->f_pos);
#else
    old_fs = get_fs();
    set_fs(KERNEL_DS_VERSION);
    ret = vfs_write_no_check(filp, (const char __user *)buf, len, &filp->f_pos);
    set_fs(old_fs);
#endif

    return ret;
}

// 封装的写文件函数
size_t file_ops_writes(struct file *file, const char *buf, size_t count)
{
    int ret = 0;
    char *buf_p = buf;
    size_t left = count;
    size_t len = 0;
    size_t writed_len = 0;
    int buf_max = 1024;

    while (left > 0) {
        len = (left > buf_max) ? buf_max : left;

        ret = file_ops_write(file, buf_p, len);
        if (ret < len)
            goto DONE;

        buf_p += len;
        writed_len += len;
        left -= len;
    }

DONE:
    return writed_len;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(4, 1, 18)

static ssize_t new_sync_read(struct file *filp, char __user *buf, size_t len,
                             loff_t *ppos)
{
    struct iovec iov = {.iov_base = buf, .iov_len = len};
    struct kiocb kiocb;
    struct iov_iter iter;
    ssize_t ret;

    init_sync_kiocb(&kiocb, filp);
    kiocb.ki_pos = *ppos;
    iov_iter_init(&iter, READ, &iov, 1, len);

    //--ret = call_read_iter(filp, &kiocb, &iter);
    ret = filp->f_op->read_iter(&kiocb, &iter);
    BUG_ON(ret == -EIOCBQUEUED);
    *ppos = (loff_t)kiocb.ki_pos;
    return ret;
}

static ssize_t __vfs_read_(struct file *file, char __user *buf, size_t count,
                           loff_t *pos)
{
    if (file->f_op->read)
        return file->f_op->read(file, buf, count, pos);
    else if (file->f_op->read_iter)
        return new_sync_read(file, buf, count, pos);
    else
        return -EINVAL;
}

static ssize_t new_sync_write(struct file *filp, const char __user *buf,
                              size_t len, loff_t *ppos)
{
    struct iovec iov = {.iov_base = (void __user *)buf, .iov_len = len};
    struct kiocb kiocb;
    struct iov_iter iter;
    ssize_t ret;

    init_sync_kiocb(&kiocb, filp);
    kiocb.ki_pos = *ppos;
    iov_iter_init(&iter, WRITE, &iov, 1, len);

    //--ret = call_write_iter(filp, &kiocb, &iter);
    ret = filp->f_op->write_iter(&kiocb, &iter);
    BUG_ON(ret == -EIOCBQUEUED);
    if (ret > 0)
        *ppos = kiocb.ki_pos;
    return ret;
}

ssize_t __vfs_write_n(struct file *file, const char __user *p, size_t count,
                    loff_t *pos)
{
    if (file->f_op->write)
        return file->f_op->write(file, p, count, pos);
    else if (file->f_op->write_iter)
        return new_sync_write(file, p, count, pos);
    else
        return -EINVAL;
}
#endif

// ssize_t nfs_vfs_read(struct file *file, char *buf, size_t count, loff_t *pos)
// {
//     ssize_t ret;
//     mm_segment_t old_fs;

//     old_fs = get_fs();
// #if LINUX_VERSION_CODE <= KERNEL_VERSION(5,4,0)
//     set_fs(get_ds());
// #else
//     set_fs(KERNEL_DS);
// #endif
//     if (count > MAX_RW_COUNT) {
//         count = MAX_RW_COUNT;
//     }
// #if LINUX_VERSION_CODE > KERNEL_VERSION(3, 18, 139)
//     ret = __vfs_read_(file, buf, count, pos);
// #else
//     ret = vfs_read(file, buf, count, pos);
// #endif
//     set_fs(old_fs);
//     return ret;
// }

// ssize_t nfs_vfs_write(struct file *file, char *buf, size_t count, loff_t
// *pos)
// {
//     ssize_t ret;
//     mm_segment_t old_fs;

//     old_fs = get_fs();
// #if LINUX_VERSION_CODE <= KERNEL_VERSION(5,4,0)
//     set_fs(get_ds());
// #else
//     set_fs(KERNEL_DS);
// #endif
//     if (count > MAX_RW_COUNT) {
//         count =  MAX_RW_COUNT;
//     }
// #if LINUX_VERSION_CODE > KERNEL_VERSION(3, 18, 139)
//     ret = __vfs_write_n(file, buf, count, pos);
// #else
//     ret = vfs_write(file, buf, count, pos);
// #endif
//     set_fs(old_fs);
//     return ret;
// }

ssize_t file_ops_read(struct file *file, char *buf, size_t count, loff_t *pos)
{
    // return nfs_vfs_read(file, buf, count, pos);
    return kernel_read(file, buf, count, pos);
}

ssize_t file_ops_read_old(struct file *file, loff_t off, void *buf,
                          size_t count)
{
    ssize_t retval;
    loff_t pos;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 14, 0)
    retval = kernel_read(file, off, buf, count);
#else
    pos = off;
    retval = kernel_read(file, buf, count, &pos);
#endif
    return retval;
}
