
#include <linux/sched.h>
#include <linux/slab.h>
#include "version_type.h"
#include "project.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0)
#include <linux/sched/mm.h>
#endif
#include <linux/mount.h>
#include "fullpath.h"

static const char deleted_str[] = " (deleted)";
static const size_t deleted_size = sizeof(deleted_str) - 1;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 15, 0)
#include <linux/seqlock.h>
seqlock_t *pvfsmount_lock = NULL;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
#include <linux/lglock.h>
struct lglock *pvfsmount_lock = NULL;
#endif

#if (LINUX_VERSION_CODE > KERNEL_VERSION(3, 15, 0))
//#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 0))

#include <linux/path.h>
struct mount {
    struct hlist_node mnt_hash;
    struct mount *mnt_parent;
    struct dentry *mnt_mountpoint;
    struct vfsmount mnt;
    struct rcu_head mnt_rcu;
#ifdef CONFIG_SMP
    struct mnt_pcp __percpu *mnt_pcp;
#else
    int mnt_count;
    int mnt_writers;
#endif
    struct list_head mnt_mounts;   /* list of children, anchored here */
    struct list_head mnt_child;    /* and going through their mnt_child */
    struct list_head mnt_instance; /* mount instance on sb->s_mounts */
    const char *mnt_devname;       /* Name of device e.g. /dev/dsk/hda1 */
};

#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 0))

#include <linux/path.h>
struct mount {
    struct hlist_node mnt_hash;
    struct mount *mnt_parent;
    struct dentry *mnt_mountpoint;
    struct vfsmount mnt;
#ifdef CONFIG_SMP
    struct mnt_pcp __percpu *mnt_pcp;
#else
    int mnt_count;
    int mnt_writers;
#endif
    struct list_head mnt_mounts;   /* list of children, anchored here */
    struct list_head mnt_child;    /* and going through their mnt_child */
    struct list_head mnt_instance; /* mount instance on sb->s_mounts */
    const char *mnt_devname;       /* Name of device e.g. /dev/dsk/hda1 */
};

#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 9, 0))
struct mount {
    struct list_head mnt_hash;
    struct mount *mnt_parent;
    struct dentry *mnt_mountpoint;
    struct vfsmount mnt;
#ifdef CONFIG_SMP
    struct mnt_pcp __percpu *mnt_pcp;
    atomic_t mnt_longterm;         /* how many of the refs are longterm */
#else
    int mnt_count;
    int mnt_writers;
#endif
    struct list_head mnt_mounts;   /* list of children, anchored here */
    struct list_head mnt_child;    /* and going through their mnt_child */
    struct list_head mnt_instance; /* mount instance on sb->s_mounts */
    const char *mnt_devname;       /* Name of device e.g. /dev/dsk/hda1 */
};
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0))
struct mount {
    struct list_head mnt_hash;
    struct mount *mnt_parent;
    struct dentry *mnt_mountpoint;
    struct vfsmount mnt;
#ifdef CONFIG_SMP
    struct mnt_pcp __percpu *mnt_pcp;
#else
    int mnt_count;
    int mnt_writers;
#endif
    struct list_head mnt_mounts;   /* list of children, anchored here */
    struct list_head mnt_child;    /* and going through their mnt_child */
    struct list_head mnt_instance; /* mount instance on sb->s_mounts */
    const char *mnt_devname;       /* Name of device e.g. /dev/dsk/hda1 */
};
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0))
struct mount {
    struct list_head mnt_hash;
    struct mount *mnt_parent;
    struct dentry *mnt_mountpoint;
    struct vfsmount mnt;
#ifdef CONFIG_SMP
    struct mnt_pcp __percpu *mnt_pcp;
    atomic_t mnt_longterm;         /* how many of the refs are longterm */
#else
    int mnt_count;
    int mnt_writers;
#endif
    struct list_head mnt_mounts;   /* list of children, anchored here */
    struct list_head mnt_child;    /* and going through their mnt_child */
    struct list_head mnt_instance; /* mount instance on sb->s_mounts */
    const char *mnt_devname;       /* Name of device e.g. /dev/dsk/hda1 */
};
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0))
static struct vfsmount *do_find_mount(struct mount *cmnt, struct dentry *d_root)
{
    struct mount *mount;
    struct vfsmount *tmnt;
    //  tag_pr_debug("cmnt->mnt.mnt_root %p,d_root %p\n",cmnt->mnt.mnt_root,d_root);
    if (cmnt->mnt.mnt_root == d_root) {
        return &cmnt->mnt;
    }

    list_for_each_entry (mount, &cmnt->mnt_mounts, mnt_child) {
        if ((tmnt = do_find_mount(mount, d_root))) {
            return tmnt;
        }
    }
}
#else
static struct vfsmount *do_find_mount(struct vfsmount *cmnt,
                                      struct dentry *d_root)
{
    struct vfsmount *mnt, *tmnt;
    // tag_pr_debug("cmnt->mnt_root=%p,d_root=%p\n",cmnt->mnt_root,d_root);struct
    // mount *mount = eal_mount(cmnt);
    if (cmnt->mnt_root == d_root) {
        return cmnt;
    }
    list_for_each_entry (mnt, &cmnt->mnt_mounts, mnt_child) {
        if ((tmnt = do_find_mount(mnt, d_root))) {
            return tmnt;
        }
    }
    return NULL;
}
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 15, 0))
// extern seqlock_t *pvfsmount_lock;
static inline void lock_mount_hash(void)
{
    if (pvfsmount_lock) {
        write_seqlock(pvfsmount_lock);
    }
}

static inline void unlock_mount_hash(void)
{
    if (pvfsmount_lock) {
        write_sequnlock(pvfsmount_lock);
    }
}

static inline struct vfsmount *find_vfsmount(struct dentry *dentry)
{
    struct mount *mnt;
    struct vfsmount *vfsmount = 0;
    lock_mount_hash();
    list_for_each_entry (mnt, &dentry->d_sb->s_mounts, mnt_instance) {
        if (dentry->d_sb->s_root == mnt->mnt.mnt_root) {
            vfsmount = &mnt->mnt;
            break;
        }
    }
    unlock_mount_hash();
    return vfsmount;
}
#elif (LINUX_VERSION_CODE > KERNEL_VERSION(3, 5, 0))

// extern struct lglock *pvfsmount_lock;
static inline void vfsmount_read_lock(void)
{
    if (pvfsmount_lock)
        br_read_lock(pvfsmount_lock);
}
static inline void vfsmount_read_unlock(void)
{
    if (pvfsmount_lock)
        br_read_unlock(pvfsmount_lock);
}
static inline struct vfsmount *find_vfsmount(struct dentry *dentry)
{
    struct mount *mnt;
    struct vfsmount *vfsmount = 0;
    vfsmount_read_lock();
    list_for_each_entry (mnt, &dentry->d_sb->s_mounts, mnt_instance) {
        if (dentry->d_sb->s_root == mnt->mnt.mnt_root) {
            vfsmount = &mnt->mnt;
            break;
        }
    }
    vfsmount_read_unlock();
    return vfsmount;
}
#else
static inline struct vfsmount *find_vfsmount(struct dentry *dentry)
{
    return do_find_mount(current->fs->rootmnt, dentry->d_sb->s_root);
}
#endif

static inline char *nfssec_version_dpath(struct dentry *dentry,
                                         struct vfsmount *vfsmnt, char *buf,
                                         int buflen)
{
    struct path path = {.mnt = vfsmnt, .dentry = dentry};
    return d_path(&path, buf, buflen);
}

// todo: 这个方式不一定正确，带到功能验证完成后，才能确定
static char *get_fullpath_by_dentry(struct dentry *dentry, char *kbuffer)
{
    char *pathroot = NULL;
    struct vfsmount *mnt;
    unsigned int size;

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(3, 5, 0))
    dget(dentry);
#endif
    mnt = find_vfsmount(dentry);
    if (!mnt)
        goto err;

    pathroot = nfssec_version_dpath(dentry, mnt, kbuffer, PAGE_SIZE);
    if (IS_ERR(pathroot)) {
        pathroot = NULL;
    } else {
        size = strlen(pathroot);
        if (!IS_ROOT(dentry) && d_unhashed(dentry) && size > deleted_size &&
            strcmp(pathroot + size - deleted_size, deleted_str) == 0)
            pathroot[size - deleted_size] = '\0';
    }

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(3, 5, 0))
    dput(dentry);
#endif

    return pathroot;
err:
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(3, 5, 0))
    dput(dentry);
#endif
    return NULL;
}

char *fullpath_by_file(struct file *file, char *buffer, unsigned int length)
{
    char *ptr = NULL;

    if (!get_file_dentry(file))
        goto DONE;

        // ptr = get_fullpath_by_dentry(get_file_dentry(file), buffer);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 0, 0)
    ptr = dentry_path_raw(get_file_dentry(file), buffer, length - 1);
#else

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 18)
    ptr = d_path(get_file_dentry(file), buffer, length - 1);
#else
    if (!get_file_vfsmount(file))
        goto DONE;
    ptr = d_path(file->f_dentry, file->f_vfsmnt, buffer, length - 1);
#endif
#endif
    if (IS_ERR(ptr)) {
        ptr = NULL;
    }

DONE:
    return ptr;
}

void fullpath_put(char *path)
{
    if (path)
        kfree(path);
}

char *get_fullpath_by_file(struct file *file)
{
    char *path = NULL;
    char *buffer = NULL;

    buffer = kzalloc(FULLPATH_MAX, GFP_ATOMIC);
    if (!buffer)
        goto DONE;

    path = fullpath_by_file(file, buffer, FULLPATH_MAX);
    if (!path)
        goto DONE;

    // strcpy(buffer, path);

DONE:
    if (buffer)
        kfree(buffer);

    return path;
}

char *get_fullpath_by_task(struct task_struct *task)
{
    struct mm_struct *mm = NULL;
    char *buffer = NULL;

    mm = get_task_mm(task);
    if (!mm || !mm->exe_file)
        return NULL;

    buffer = get_fullpath_by_file(mm->exe_file);

DONE:
    if (mm)
        mmput(mm);

    return buffer;
}

// #include "../common/realpath.h"
char *get_fullpath_by_dentry_mnt(struct dentry *dentry, struct vfsmount *mnt)
{
    char *fullPath;
    struct path _path = {};
    char * name_buf = NULL;
    
    name_buf = kmalloc(PAGE_SIZE, GFP_KERNEL);
	if (!name_buf)
        return NULL;
	memset(name_buf, 0x0, PAGE_SIZE) ;

    _path.mnt = mnt;
    _path.dentry = dentry;

    fullPath = d_path(&_path, name_buf, PAGE_SIZE);
    
    if (name_buf)
        kfree(name_buf);

    return fullPath;
}

#include <linux/nsproxy.h>
#include <linux/pid_namespace.h>

char *get_fullpath_by_pid(long pid)
{
    char *fullpath = NULL;
    struct task_struct *task;

    // notice: get like this or use /proc/pid/exec, which one is faster?
    rcu_read_lock();
    task = pid_task(find_vpid(pid), PIDTYPE_PID);
    rcu_read_unlock();
    if (!task)
        goto DONE;

    fullpath = get_fullpath_by_task(task);

DONE:
    return fullpath;
}
