#include <project.h>
#include "comm_hash_table.h"

static size_t __bkdrhash(const char *str)
{
    register size_t hash = 0;
    size_t ch;

    // -- while (ch = (size_t)*str++)
    for (; (*str); str++) {
        ch = (size_t)*str;
        hash = hash * 131 + ch;
    }
    return hash;
}

static size_t __bkdrhash_with_length(const char *str, unsigned int length)
{
    register size_t hash = 0;
    size_t ch;
    unsigned int i;

    for (i = 0; i < (length / sizeof(char)); i++) {
        ch = (size_t)str[i];
        hash = hash * 131 + ch;
    }
    return hash;
}

void __hash_table_clear(sec_hash_table_t *table)
{
    int i = 0;

    for (i = 0; i < (int)table->bucket_count; i++) {
        LIST_HEAD *hdr = &table->hash_list[i];
        while (!__list_is_empty(hdr)) {
            LIST_HEAD *entry;
            __list_remove_head(hdr, entry);
            table->hash_release(entry);
        }
    }
    return;
}

void comm_hash_table_destroy(sec_hash_table_t *table)
{
    __hash_table_clear(table);
    if (table->hash_list) {
        kfree(table->hash_list);
        table->hash_list = NULL;
    }
}

int comm_hash_table_remove_item(sec_hash_table_t *table, LIST_HEAD *entry)
{
    if ((0 == table->bucket_count))
        return -1;

    __list_remove_item(entry);

    __list_init(entry);

    table->hash_release(entry);
    table->item_count--;

    return 0;
}

int comm_hash_table_init(sec_hash_table_t *table, hash_tb_value_func hash_value,
                         hash_tb_release_func hash_release)
{
    table->bucket_count = 0;
    table->hash_list = NULL;
    table->hash_value = hash_value;
    table->hash_release = hash_release;
    return 0;
}

static int comm_hash_table_equal(sec_hash_table_t *table, LIST_HEAD *item,
                                 const char *ptrVal)
{
    if (0 == strcmp((char *)table->hash_value(item), (char *)ptrVal)) {
        return 0;
    }
    return -1;
}

static int __x_hash_table_equal_by_length(sec_hash_table_t *table, LIST_HEAD *item,
                                          const char *ptrVal,
                                          unsigned int length)
{
    if (0 == memcmp((char *)table->hash_value(item), (char *)ptrVal, length)) {
        return 0;
    }
    return -1;
}

static int __comm_hash_table_insert(sec_hash_table_t *table, LIST_HEAD *item,
                                    const char *val_ptr)
{
    unsigned int index;
    LIST_HEAD *entry;

    if ((0 == table->bucket_count)) {
        return -1;
    }

    index = __bkdrhash(val_ptr);

    index = index % table->bucket_count;

    entry = table->hash_list[index].next;
    while (entry != &table->hash_list[index]) {
        if (0 == comm_hash_table_equal(table, entry, val_ptr)) {
            return -1;
        }
        entry = entry->next;
    }
    __list_add_tail(&table->hash_list[index], item);

    return 0;
}

int comm_hash_table_insert(sec_hash_table_t *table, LIST_HEAD *item,
                           const char *val_ptr)
{
    int retval;
    
    retval = __comm_hash_table_insert(table, item, val_ptr);
    if (0 == retval) {
        table->item_count++;
    }
    return retval;
}

int __x_hash_table_set_size(sec_hash_table_t *table, int bucket_count)
{
    int i = 0;
    int real_count = bucket_count + 1;
    LIST_HEAD *retval = NULL;

    if ((bucket_count <= 0)) {
        return -1;
    }

    if (((int)table->bucket_count == bucket_count)) {
        return -1;
    }

    retval = (LIST_HEAD *)kmalloc(sizeof(LIST_HEAD) * (real_count), GFP_ATOMIC);

    table->hash_list = retval;
    if ((!table->hash_list)) {
        table->bucket_count = 0;
        return -1;
    }
    // -- table->bucket_count = bucket_count;

    table->bucket_count = bucket_count;
    for (i = 0; i < real_count; i++) {
        __list_init(&table->hash_list[i]);
    }

    return 0;
}

void comm_hash_table_remove_all(sec_hash_table_t *table)
{
    int i = 0;
    if ((0 == table->bucket_count)) {
        return;
    }
    for (; i < (int)table->bucket_count; i++) {
        LIST_HEAD *header = &table->hash_list[i];
        while (!__list_is_empty(header)) {
            LIST_HEAD *entry = header->next;
            __list_remove_item(entry);
            table->item_count--;
            __list_init(entry);
            table->hash_release(entry);
        }
    }

    return;
}

int comm_hash_table_set_buckets(sec_hash_table_t *table, int bucket_count)
{
    int i;
    int old_bucket_count = table->bucket_count;
    LIST_HEAD *old_hash_list = table->hash_list;

    if (old_bucket_count == bucket_count)
        return 0;

    if (0 != __x_hash_table_set_size(table, bucket_count)) {
        table->bucket_count = old_bucket_count;
        table->hash_list = old_hash_list;
        return -1;
    }
    for (i = 0; i < old_bucket_count; i++) {
        LIST_HEAD *head = &old_hash_list[i];
        while (!__list_is_empty(head)) {
            LIST_HEAD *entry = head->next;
            __list_remove_item(entry);
            __comm_hash_table_insert(table, entry, table->hash_value(entry));
        }
    }

    if (old_hash_list) 
        kfree(old_hash_list);
    return 0;
}

LIST_HEAD *comm_hash_table_find_ex(sec_hash_table_t *table, const char *val_ptr,
                                   unsigned int length)
{
    unsigned int index;
    LIST_HEAD *sublist;
    LIST_HEAD *target = NULL, *entry;

    if ((0 == table->bucket_count)) {
        return NULL;
    }

    index = __bkdrhash_with_length(val_ptr, length);

    index = index % table->bucket_count;

    sublist = &table->hash_list[index];
    entry = sublist->next;
    while (entry != sublist) {
        if (0 ==
            __x_hash_table_equal_by_length(table, entry, val_ptr, length)) {
            target = entry;
            break;
        }
        entry = entry->next;
    }

    return target;
}

LIST_HEAD *comm_hash_table_find(sec_hash_table_t *table, const char *val_ptr)
{
    unsigned int index;
    LIST_HEAD *sublist;
    LIST_HEAD *target = NULL, *entry;

    if ((0 == table->bucket_count)) {
        return NULL;
    }

    index = __bkdrhash(val_ptr);
    index = index % table->bucket_count;

    sublist = &table->hash_list[index];
    entry = sublist->next;
    while (entry != sublist) {
        if (0 == comm_hash_table_equal(table, entry, val_ptr)) {
            target = entry;
            break;
        }
        entry = entry->next;
    }

    return target;
}

int comm_hash_table_remove(sec_hash_table_t *table, const char *val_ptr)
{
    unsigned int index;
    LIST_HEAD *list;
    LIST_HEAD *entry;

    if ((0 == table->bucket_count))
        return -1;

    index = __bkdrhash(val_ptr);
    index = index % table->bucket_count;

    list = &table->hash_list[index];

    entry = list->next;
    while (entry != list) {
        if (0 == comm_hash_table_equal(table, entry, val_ptr)) {
            __list_remove_item(entry);
            table->item_count--;
            __list_init(entry);
            table->hash_release(entry);
            return 0;
        }
        entry = entry->next;
    }

    return -1;
}
