#include <project.h>
#include "include/misc/hash_code.h"

#include "hashtb.h"

#define SEC_HASH_INIT_VALUE (0x12345678)
static unsigned int hashtb_hash(HASHTB *table, unsigned int *val_ptr)
{
    unsigned int retval;
    unsigned int count = table->valueLen / sizeof(int);
    switch (count) {
        case 0:
            return 0;
            break;
        case 1:
            retval = _sec_hash_1word(val_ptr[0]);
            break;
        case 2:
            retval = _sec_hash_3words(val_ptr[0], val_ptr[1], 0, SEC_HASH_INIT_VALUE);
            break;
        case 3:
            retval = _sec_hash_3words(val_ptr[0], val_ptr[1], val_ptr[2],
                                   SEC_HASH_INIT_VALUE);
            break;
        default:
            retval = _sec_hash2(val_ptr, count, SEC_HASH_INIT_VALUE);
            ;
    }
    retval = (retval % (unsigned int)table->bucket_count);
    return retval;
}

static __inline__ void __hashtb_free(HASHTB *table, void *__data)
{
    if (table->_free) {
        (*table->_free)(__data);
    } else {
        kfree(__data);
    }
    return;
}

static __inline__ void *__hashtb_alloc(HASHTB *table, unsigned int size)
{
    void *__data = NULL;
    if (table->_alloc) {
        __data = (*table->_alloc)(size);
    } else {
        __data = (LIST_HEAD *)kmalloc(size, GFP_ATOMIC);
    }
    return __data;
}

static BOOLEAN __hashtb_equal(HASHTB *table, LIST_HEAD *item,
                              unsigned int *ptrVal)
{
    return (0 == memcmp((char *)table->get_value(item), (char *)ptrVal,
                        table->valueLen));
}

int hashtb_init_ex(HASHTB *table, HAHSTB_VALUE __get_value_ptr,
                   HAHSTB_RELEASE release_item, unsigned int valueLen,
                   HASHTB_ALLOC _alloc, HASHTB_FREE _free)
{
    int retval;
    retval = hashtb_init(table, __get_value_ptr, release_item, valueLen);
    if (0 == retval) {
        table->_free = _free;
        table->_alloc = _alloc;
    }
    return retval;
}

int hashtb_init(HASHTB *table, HAHSTB_VALUE __get_value_ptr,
                HAHSTB_RELEASE release_item, unsigned int valueLen)
{
    if (0 != (valueLen % sizeof(int))) {
        return -1;
    }
    table->bucket_count = 0;
    table->hash_list = NULL;
    table->release_item = release_item;
    table->get_value = __get_value_ptr;
    table->valueLen = valueLen;
    table->_free = NULL;
    table->_alloc = NULL;
    return 0;
}

void hashtb_remove_all(HASHTB *table)
{
    int i = 0;
    if ((0 == table->bucket_count)) {
        return;
    }
    for (; i < (int)table->bucket_count; i++) {
        LIST_HEAD *header = &table->hash_list[i];
        while (!__list_is_empty(header)) {
            LIST_HEAD *entry = header->next;
            __list_remove_item(entry);
            table->item_count--;
            __list_init(entry);
            table->release_item(entry);
        }
    }

    return;
}

void hashtb_destroy(HASHTB *table)
{
    int i = 0;
    if ((0 == table->bucket_count)) {
        return;
    }

    if (table->hash_list) {
        for (; i < (int)table->bucket_count; i++) {
            LIST_HEAD *header = &table->hash_list[i];
            while (!__list_is_empty(header)) {
                LIST_HEAD *entry = header->next;
                __list_remove_item(entry);
                table->item_count--;
                __list_init(entry);
                table->release_item(entry);
            }
        }

        __hashtb_free(table, table->hash_list);
    }
    table->bucket_count = 0;
    table->hash_list = NULL;
    return;
}

LIST_HEAD *hashtb_find(HASHTB *table, unsigned int *val_ptr)
{
    unsigned int index;
    LIST_HEAD *sublist;
    LIST_HEAD *target = NULL, *entry;

    if ((0 == table->bucket_count)) 
        return NULL;

    index = hashtb_hash(table, val_ptr);
    sublist = &table->hash_list[index];
    entry = sublist->next;
    while (entry != sublist) {
        if (__hashtb_equal(table, entry, val_ptr)) {
            target = entry;
            break;
        }
        entry = entry->next;
    }

    return target;
}

static BOOLEAN __hashtb_insert_item(HASHTB *table, LIST_HEAD *item,
                                    unsigned int *val_ptr)
{
    unsigned int index;
    LIST_HEAD *entry;

    if ((0 == table->bucket_count)) {
        return FALSE;
    }

    index = hashtb_hash(table, val_ptr);
    entry = table->hash_list[index].next;
    while (entry != &table->hash_list[index]) {
        if (__hashtb_equal(table, entry, val_ptr)) {
            return FALSE;
        }
        entry = entry->next;
    }

    __list_add_tail(&table->hash_list[index], item);

    return TRUE;
}

BOOLEAN hashtb_insert_item(HASHTB *table, LIST_HEAD *item,
                           unsigned int *val_ptr)
{
    BOOLEAN retval;
    retval = __hashtb_insert_item(table, item, val_ptr);
    if (retval) {
        table->item_count++;
    }
    return TRUE;
}

BOOLEAN hashtb_remove_item_entry(HASHTB *table, LIST_HEAD *entry)
{
    if ((0 == table->bucket_count))
        return FALSE;

    if (__list_is_empty(entry)) {
        return FALSE;
    }

    table->item_count--;
    __list_remove_item_safe(entry);
    table->release_item(entry);

    return TRUE;
}

BOOLEAN hashtb_remove_item(HASHTB *table, unsigned int *val_ptr)
{
    // -- LIST_HEAD * item;
    unsigned int index;
    LIST_HEAD *list;
    LIST_HEAD *entry;

    if ((0 == table->bucket_count))
        return FALSE;

    index = hashtb_hash(table, val_ptr);
    list = &table->hash_list[index];

    entry = list->next;
    while (entry != list) {
        if (__hashtb_equal(table, entry, val_ptr)) {
            __list_remove_item(entry);
            table->item_count--;
            __list_init(entry);
            table->release_item(entry);
            return TRUE;
        }
        entry = entry->next;
    }

    return FALSE;
}

BOOLEAN __set_hashtb_size(HASHTB *table, int bucket_count)
{
    int i = 0;
    int real_count = bucket_count + 1;
    LIST_HEAD *retval = NULL;

    if ((bucket_count <= 0)) {
        return FALSE;
    }

    if (((int)table->bucket_count == bucket_count)) {
        return FALSE;
    }

    // -- table->hash_list = (LIST_HEAD * )Malloc(sizeof(LIST_HEAD) *
    // (real_count));
    // -- NdisAllocateMemoryWithTag(&retval, sizeof(LIST_HEAD) * (real_count),
    // TAG);

    retval =
        (LIST_HEAD *)__hashtb_alloc(table, sizeof(LIST_HEAD) * (real_count));

    table->hash_list = retval;
    if ((!table->hash_list)) {
        table->bucket_count = 0;
        return FALSE;
    }
    // -- table->bucket_count = bucket_count;

    table->bucket_count = bucket_count;
    for (i = 0; i < real_count; i++) {
        __list_init(&table->hash_list[i]);
    }

    return TRUE;
}

int hashtb_get_buckets(HASHTB *table)
{
    return table->bucket_count;
}

// -- 支持动态增加桶的数目；
BOOLEAN hashtb_set_buckets(HASHTB *table, int bucket_count)
{
    int i;
    int old_bucket_count = table->bucket_count;
    LIST_HEAD *old_hash_list = table->hash_list;

    if (old_bucket_count == bucket_count)
        return TRUE;
    if ((FALSE == __set_hashtb_size(table, bucket_count))) {
        table->bucket_count = old_bucket_count;
        table->hash_list = old_hash_list;
        return FALSE;
    }
    for (i = 0; i < old_bucket_count; i++) {
        LIST_HEAD *head = &old_hash_list[i];
        while (!__list_is_empty(head)) {
            LIST_HEAD *entry = head->next;
            __list_remove_item(entry);
            __hashtb_insert_item(table, entry,
                                 (unsigned int *)table->get_value(entry));
        }
    }

    if (old_hash_list) {
        __hashtb_free(table, old_hash_list);
    }
    return TRUE;
}
