#include <project.h>
#include "object_pool/object_pool.h"

#include "scn_machine.h"

#include "key_tree.h"
#include "scn_content_search.h"  // -- 测试使用
#include "scn_node.h"


// 向关键树中插入一个节点
// 参数 parent: 父节点
// 参数 bit4: 4位数据
// 参数 stop_flag: 停止标志，1表示停止，0表示不停止
// 参数 key_ctx: 上下文
// 返回值: 插入的节点
scan_node_t *__insert_node(scan_machine_t *ma, scan_node_t *parent, unsigned char bit4,
                        int stop_flag, void *key_ctx)
{
    scan_node_t *son;

    // 遍历父节点的子节点链表，判断是否有相同值的节点
    son = parent->_u_.relation.son_node;
    
    while (son) {
        if (son->value == bit4) {
            parent = son;
            break;
        }
        son = son->_u_.relation.brother_node;
    }
    
    // 未找到，则创建新节点，将其插入到父节点的子节点链表中
    if (!son) {
        son = new_scn_node(ma);
        if (son) {
            son->_u_.relation.brother_node = parent->_u_.relation.son_node;
            parent->_u_.relation.son_node = son;
            son->value = bit4;
            son->stop_flag = (!!stop_flag);

            // 更新父节点的未构建子节点计数器
            parent->t.unbuilt.son_count++; 

            // 先将新节点添加到未构建节点链表中
            __list_add_tail(&ma->unbuild_nodes, &son->t.unbuilt.ub_list);
        }
    }

    return son;
}

// 向关键树中插入一个值（两个节点）
// 参数 parent: 父节点
// 参数 val: 要插入的值
// 参数 pnext: 下一个节点的指针
// 参数 key_ctx: 上下文
// 返回值: 0/-1 成功/失败
__inline__ int __insert_value(scan_machine_t *ma, scan_node_t *parent,
                              unsigned char val, scan_node_t **pnext,
                              void *key_ctx)
{
    int retval = -1;
    scan_node_t *node;
    unsigned char bit4;

    // 取值低4位，插入为一个节点
    bit4 = (val & 0x0F);
    node = __insert_node(ma, parent, bit4, 0, NULL);
    if (unlikely(!node))
        goto DONE;
        
    // 取值高4位，插入位一个节点
    bit4 = val >> 4;
    parent = node;
    node = __insert_node(ma, parent, bit4, 1, key_ctx);
    *pnext = node;
    if (node != NULL)
        retval = 0;

DONE:
    return retval;
}

int scan_scn_key_tree_insert_item(scan_machine_t *ma, char *key, int len,
                                  void *key_ctx)
{
    int ret = 0;
    scan_node_t *cur = ma->root;
    scan_node_t *next = NULL;
    char *ptr = key;

    // 依次将key中的值插入
    while (len > 0 && cur) {
        next = NULL;

        ret = __insert_value(ma, cur, *ptr, &next, key_ctx);
        if (0 != ret)
            break;

        len--;
        ptr++;
        cur = next;
    }

    // 插入上下文
    if (len == 0 && next) {
        if (ma->op.insert_key) {
           ma->op.insert_key(next, key_ctx);
        }
    }

    return 0;
}

// 构建节点的转移关系
// 参数 node: 要构建转移关系的节点
// 返回值: 0/-1 成功/失败
__inline__ int __build_node(scan_machine_t *ma, scan_node_t *node)
{
    int retval = -1;
    // -- unsigned short son_count = node->t.unbuilt.son_count;
    scan_node_t *son;

    son = node->_u_.relation.son_node;

    while (son) {
        node->t.transfer.nodes[son->value] = son;
        son = son->_u_.relation.brother_node;
    }

    __list_add_tail(&ma->built_nodes, &node->_u_.list);
    retval = 0;

    return retval;
}

// 构建 DFA 节点的转移关系：将节点的子节点按值映射到节点的转移表中，并将节点添加到已构建节点链表中
// 参数 node: 要构建转移关系的 DFA 节点
// 返回值: 0/-1 成功/失败
__inline__ int __build_dfa_node(scan_machine_t *ma, scan_node_t *node)
{
    int retval = -1;

    scan_node_t *son;

    son = node->_u_.relation.son_node;

    while (son) {
        node->t.transfer.nodes[son->value] = son;
        son = son->_u_.relation.brother_node;
    }

    __list_add_tail(&ma->built_nodes, &node->_u_.list);
    retval = 0;
    // -- DONE:

    return retval;
}

int scan_scn_key_tree_build(scan_machine_t *ma)
{
    int retval = 0;
    LIST_HEAD *entry;
    
    if (unlikely(ma->is_built))
        return -1;

    // 遍历未构建节点链表
    while (!__list_is_empty(&ma->unbuild_nodes)) {
        scan_node_t *node;

        __list_remove_head(&ma->unbuild_nodes, entry);

        node = OBJECT_PTR(entry, scan_node_t, t.unbuilt.ub_list);

        memset(&node->t.unbuilt, 0, sizeof(node->t.unbuilt));
        
        // 依次构建节点的转移关系
        if (ma->is_dfa) {
            __build_dfa_node(ma, node);
        } else {
            __build_node(ma, node);
        }
    }

    // 标记状态机已构建
    ma->is_built = 1;
    
    return retval;
}
