#include <project.h>

#include "scn_machine.h"
#include "key_tree.h"
#include "scn_content_search.h"
#include "scn_multi_match.h"
#include "scn_node.h"
#include "statistic.h"

// #define __toup(c) ((c) >= 'a' && (c) <= 'z' ? (c) + ('A' - 'a') : (c))

SCN_PATTERN *new_scn_pattern(unsigned int patExtlen)
{
    SCN_PATTERN *pattern;
    pattern = (SCN_PATTERN *)kmalloc(sizeof(SCN_PATTERN) + patExtlen, GFP_ATOMIC);

    if (unlikely(!pattern))
        return NULL;

    memset(pattern, 0, sizeof(SCN_PATTERN));
#ifdef __CNT_PATTERN_DEBUG__
    pattern->magic = SCN_PATTERN_MAGIC;
#endif
    INC_OBJ_COUNT(cnt_pattern);
    return pattern;
}

void delete_scn_pattern(SCN_PATTERN *pattern)
{
#ifdef __STOR_PATTERN_FREE_CONTEXT
    if (pattern->context && pattern->free_context) {
        pattern->free_context(pattern->context);
    }
#endif
    kfree(pattern);
    DEC_OBJ_COUNT(cnt_pattern);
    return;
}

static void delete_cnt_search_am(cnt_search_am_t *am)
{
    LIST_HEAD *entry;

    if (am->state_engine) 
        comm_object_release(am->state_engine);
    
    while (!__list_is_empty(&am->pat_list)) {
        SCN_PATTERN *pat;

        __list_remove_tail(&am->pat_list, entry);
        
        pat = OBJECT_PTR(entry, SCN_PATTERN, list_head);
        delete_scn_pattern(pat);
    }
    kfree(am);
    DEC_OBJ_COUNT(cnt_search_am);
    return;
}

static __inline__ cnt_search_am_t *new_cnt_search_am(BOOL is_dfa)
{
    cnt_search_am_t *am = (cnt_search_am_t *)kmalloc(sizeof(cnt_search_am_t), GFP_ATOMIC);
    if (unlikely(!am))
        return NULL;

    memset(am, 0, sizeof(cnt_search_am_t));
    
    am->state_engine = scan_scn_create_multi_machine(is_dfa);
    if (unlikely(!am->state_engine)) {
        delete_cnt_search_am(am);
        return NULL;
    }

    __list_init(&am->pat_list);
    am->header.destroy_ptr = (destroy_atom_ptr)delete_cnt_search_am;
    INC_OBJ_COUNT(cnt_search_am);
    return am;
}

__attribute__((visibility("default"))) 
void * scan_scn_trie_create(unsigned int patExtLen)
{
    cnt_search_am_t *am = new_cnt_search_am(TRUE);
    if (unlikely(!am))
        return NULL;
        
    comm_object_addref(am);
    am->patExtLen = patExtLen;
    // -- am->check = _check;
    return am;
}

__attribute__((visibility("default"))) void *
scan_scn_am_create(BOOL is_dfa, unsigned int patExtLen)
{
    cnt_search_am_t *am = new_cnt_search_am(is_dfa);
    if (unlikely(!am))
        return NULL;
        
    comm_object_addref(am);
    am->patExtLen = patExtLen;
    // -- am->check = _check;
    return am;
}

#ifdef __STOR_PATTERN_FREE_CONTEXT
__attribute__((visibility("default"))) SCN_PATTERN *
scan_scn_am_insert_pattern(void *am_ptr, unsigned char *pat,
                           unsigned short pat_len, void (*free_ctx)(void *),
                           void *ctx)
#else
__attribute__((visibility("default"))) SCN_PATTERN *
scan_scn_am_insert_pattern(void *am_ptr, unsigned char *pat,
                           unsigned short pat_len, void *ctx)
#endif
{
    cnt_search_am_t *am = (cnt_search_am_t *)am_ptr;
    SCN_PATTERN *pattern = new_scn_pattern(am->patExtLen);

    tag_pr_debug("==============am->patExtLenL[%d]\n", am->patExtLen);

    if (unlikely(!pattern))
        return NULL;
            tag_pr_debug("\n");

#ifdef __STOR_PATTERN_INFO
    pattern->pat = pat;
#endif
    pattern->pat_len = pat_len;
    // -- pattern->check = check;

#ifdef __STOR_PATTERN_FREE_CONTEXT
    pattern->free_context = free_ctx;
#endif
    pattern->context = ctx;
    tag_pr_debug("\n");

    if (0 != scan_scn_key_tree_insert_item(am->state_engine, (char *)pat,
                                           pat_len, pattern)) {
        tag_pr_debug("\n");
                               
        delete_scn_pattern(pattern);
        return NULL;
    }
    tag_pr_debug("\n");

    atomic_inc_return(&am->pat_count);
    __list_add_tail(&am->pat_list, &pattern->list_head);

    return pattern;
}

#ifdef __STOR_PATTERN_FREE_CONTEXT
__attribute__((visibility("default"))) SCN_PATTERN *
scan_scn_trie_insert_key(void *am_ptr, unsigned char *pat,
                         unsigned short pat_len, void (*free_ctx)(void *),
                         void *ctx)
#else
__attribute__((visibility("default"))) SCN_PATTERN *
scan_scn_trie_insert_key(void *am_ptr, unsigned char *pat,
                         unsigned short pat_len, void *ctx)
#endif
{
#ifdef __STOR_PATTERN_FREE_CONTEXT
    return scan_scn_am_insert_pattern(am_ptr, pat, pat_len, free_ctx, ctx);
#else
    return scan_scn_am_insert_pattern(am_ptr, pat, pat_len, ctx);
#endif
}

// 类似于 keyword tree，没有自动化，只能通过一种简单的方式实现
__attribute__((visibility("default"))) int scan_scn_trie_build(void *am)
{
    return scan_scn_trie_compile(((cnt_search_am_t *)am)->state_engine);
}

__attribute__((visibility("default"))) int scan_scn_am_build(void *am)
{
    return scan_scn_multi_machine_compile(((cnt_search_am_t *)am)->state_engine);
}

// __inline__ CNT_NODE *__cnt_dfa_go(cnt_search_am_t *am, unsigned char value,
//                                   CNT_NODE *parent)
// {
//     CNT_NODE *node;
//     unsigned char bit4;

//     bit4 = value & 0x0F;
//     node = (CNT_NODE *)parent->scan_scn_node.t.transfer.nodes[bit4];
    
//     parent = node;

//     bit4 = value >> 4;
//     node = (CNT_NODE *)parent->scan_scn_node.t.transfer.nodes[bit4];

//     return node;
// }

__inline__ CNT_NODE *__trie_dfa_go(cnt_search_am_t *am, unsigned char value,
                                   CNT_NODE *parent)
{
    CNT_NODE *node;
    unsigned char bit4;

    bit4 = value & 0x0F;
    node = (CNT_NODE *)parent->scan_scn_node.t.transfer.nodes[bit4];
    
    parent = node;
    if (!parent)
        return NULL;

    bit4 = value >> 4;
    node = (CNT_NODE *)parent->scan_scn_node.t.transfer.nodes[bit4];

    return node;
}

// __inline__ CNT_NODE *__cnt_nfa_go(cnt_search_am_t *am, unsigned char value,
//                                   CNT_NODE *parent)
// {
//     CNT_NODE *node;
//     unsigned char bit4;
    
//     bit4 = value & 0x0F;
//     node =
//         (CNT_NODE *)scan_scn_multi_machine_go(am->state_engine, parent, bit4);
//     if (unlikely(!node))
//         return NULL;

//     bit4 = value >> 4;
//     node = (CNT_NODE *)scan_scn_multi_machine_go(am->state_engine, node, bit4);

//     return node;
// }

// 当返回值为0时表示匹配成功
__inline__ int __cnt_match(cnt_search_am_t *am, SCAN_CHECK_ROUTINE _check,
                           CNT_NODE *node, const unsigned char *base,
                           unsigned char *ptr, int left, void *search_ctx)
{
    SCN_MUL_CONTAINER *contain;
    contain = node->match_keys;
    // -- __check_patterns(node);

    while (contain) {
        SCN_PATTERN *pattern = (SCN_PATTERN *)contain->key_ctx;
        char *pattern_pos;

        // -- ix_assert(pattern && pattern->check);

#ifdef __CNT_PATTERN_DEBUG__
        if (pattern->magic != SCN_PATTERN_MAGIC) {
            // DbgPrint("why?\n");
        }
#endif

        // 传入参数:
        // 匹配的前缀字符串，最大匹配长度和匹配模式
        pattern_pos = (char *)(ptr - pattern->pat_len + 1);
        // 返回匹配结果
        if (SCN_CHECK_BREAK ==
            (*_check)(pattern, pattern_pos,
                      (unsigned int)(pattern_pos - (char *)base), left - 1,
                      search_ctx)) {
            // -- CNT_CHECK_CONTINUE
            return 0;
        }

        contain = contain->next;
    }
    return -1;
}

__attribute__((visibility("default"))) int
scan_scn_trie_traverse(void *am_ptr, unsigned char *content, int len,
                       SCAN_CHECK_ROUTINE _check, void *search_ctx)
{
    cnt_search_am_t *am = (cnt_search_am_t *)am_ptr;
    register unsigned char *ptr = (unsigned char *)content;
    register int left = len;
    CNT_NODE *node;
    CNT_NODE *root = (CNT_NODE *)am->state_engine->root;

    if (unlikely(len <= 0))
        return 0;
    if (unlikely(!am->state_engine->is_built))
        return -1;

    node = root;
    if (am->state_engine->is_dfa) {
        while (left) {
            node = __trie_dfa_go(am, *ptr, node);
            if (!node) {
                goto DONE;
            }
            if (node->match_keys && (0 == __cnt_match(am, _check, node, content,
                                                      ptr, left, search_ctx)))
                goto DONE;
            ptr++;
            left--;
        }
    } else {
		// while(left){	
		// 	node = __cnt_nfa_go(am, __toup(* ptr), node);
		// 	if(node->match_keys && (0 == __cnt_match(am, node, content, ptr, left, search_ctx))) goto DONE;
 		// 	ptr ++;
		// 	left --;
		// }
    }

DONE:
    return 0;
}

// __attribute__((visibility("default"))) int
// scan_scn_am_traverse(void *am_ptr, unsigned char *content, int len,
//                      SCAN_CHECK_ROUTINE _check, void *search_ctx)
// {
//     cnt_search_am_t *am = (cnt_search_am_t *)am_ptr;
//     register unsigned char *ptr = (unsigned char *)content;
//     register int left = len;
//     CNT_NODE *node;
//     CNT_NODE *root = (CNT_NODE *)am->state_engine->root;

//     if (unlikely(len <= 0))
//         return 0;
//     if (unlikely(!am->state_engine->is_built))
//         return -1;
    
//     node = root;
//     if (am->state_engine->is_dfa) {
//         while (left) {
//             node = __cnt_dfa_go(am, *ptr, node);
//             if (node->match_keys && (0 == __cnt_match(am, _check, node, content,
//                                                       ptr, left, search_ctx)))
//                 goto DONE;
//             ptr++;
//             left--;
//         }
//     } else {
//         while (left) {
//             node = __cnt_nfa_go(am, *ptr, node);
//             if (node->match_keys && (0 == __cnt_match(am, _check, node, content,
//                                                       ptr, left, search_ctx)))
//                 goto DONE;

//             ptr++;
//             left--;
//         }
//     }

// DONE:
//     return 0;
// }

// scn_am_go 函数用于根据传入的数据包和位置信息，移动当前数据位置到base和baseOffset处;
// 也可以理解为从 &base[baseOffset] 开始的扫描，如果当前_node==NULL，则baseOffset应为0;
// 在这之后需要将 base 设置为新的位置，以便下次扫描从该位置开始(这可能需要匹配器来处理，具体取决于实现)
// 返回值是当前节点，一个数据包，应该将其设置为初始状态以准备下一次扫描;
// __attribute__((visibility("default"))) void *
// scan_scn_am_go(void *am_ptr, unsigned char *base, unsigned int baseOffset,
//                int len, SCAN_CHECK_ROUTINE _check, void *search_ctx,
//                void *_node)
// {
//     cnt_search_am_t *am = (cnt_search_am_t *)am_ptr;
//     register unsigned char *ptr = (unsigned char *)&base[baseOffset];
//     register int left = len;
//     CNT_NODE *node = (CNT_NODE *)_node;
//     // -- CNT_NODE * root = (CNT_NODE *)am->state_engine->root;
   
//     if (unlikely(len <= 0))
//         return node;

//     if (!node) {
//         node = (CNT_NODE *)am->state_engine->root;
//     }

//     if (am->state_engine->is_dfa) {
//         while (left) {
//             node = __cnt_dfa_go(am, *ptr, node);
//             if (node->match_keys && (0 == __cnt_match(am, _check, node, base,
//                                                       ptr, left, search_ctx)))
//                 goto DONE;
//             ptr++;
//             left--;
//         }
//     } else {
//         while (left) {
//             node = __cnt_nfa_go(am, *ptr, node);
//             if (node->match_keys && (0 == __cnt_match(am, _check, node, base,
//                                                       ptr, left, search_ctx)))
//                 goto DONE;

//             ptr++;
//             left--;
//         }
//     }

// DONE:
//     return node;
// }
