#include <linux/binfmts.h>
#include <linux/file.h>
#include <linux/mman.h>
#include <linux/module.h>
#include <linux/mount.h>
#include <linux/security.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/version.h>
#include <project.h>

#include "common/file_ops/file_ops.h"
#include "common/general/sha1.h"
#include "common/hash/digest_ctx.h"
#include "hash_calc.h"
#include "common/sec_krnl_misc.h"
#include "common/version_type.h"
#include "common/elf_assist.h"

#define __CALC_SIZE (1024)
static int __wl_sec_hash_calc(struct file *file, char *digest);

int wlSecHashCalcFile(const char *path, char *digest)
{
    int retval = -1;
    struct file *file = NULL;

    file = filp_open(path, O_RDONLY | O_SYNC, 0);
    if (IS_ERR(file)) {
        file = NULL;
        // -- printk("[wlSecHashCalcFile] open (%s) error!\n", path);
        goto DONE;
    }
    retval = __wl_sec_hash_calc(file, digest);
DONE:
    if (file) {
        filp_close(file, NULL);
    }
    return retval;
}

static int __wl_sec_hash_calc(struct file *file, char *digest)
{
    DIGEST_CTX ctx;  // -- struct hash_desc desc;
    // -- struct scatterlist sg[1];
    loff_t i_size, offset = 0;
    char *rbuf = NULL;
    int retval = -1;

    digest_ctx_init(&ctx);

    rbuf = kzalloc(__CALC_SIZE, GFP_KERNEL);
    if (!rbuf) {
        retval = -ENOMEM;
        goto DONE;
    }

    if (!get_file_dentry(file))
        goto DONE;

    i_size = i_size_read(get_file_dentry(file)->d_inode);

    while (offset < i_size) {
        int rbuf_len;

        // -- #if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)
        // -- loff_t pos;
        // -- #endif

        rbuf_len = file_ops_read(file, rbuf, __CALC_SIZE, &offset);

        // -- #if LINUX_VERSION_CODE < KERNEL_VERSION(4, 14, 0)
        // -- 	rbuf_len = kernel_read(file, offset, rbuf, __CALC_SIZE);
        // -- #else
        // -- 	pos = offset;
        // -- 	rbuf_len = kernel_read(file, rbuf, __CALC_SIZE, &pos);
        // -- #endif

        if (rbuf_len < 0) {
            retval = rbuf_len;
            goto DONE;
        }
        if (rbuf_len == 0) {
            break;
        }
        // offset += rbuf_len;

        digest_ctx_update(&ctx, rbuf, (int)rbuf_len);
    }

    digest_ctx_finish(&ctx, digest);
    retval = 0;
DONE:
    if (rbuf)
        kfree(rbuf);

    return retval;
}

extern ssize_t __kread(struct file *file, loff_t off, void *rbuf, size_t count,
                       unsigned char bOPS);

// -- calculate in file_permission, kernel_read will invoke another
// file_permission,
// -- so call f_op->read to avoid kernel_read;
int wlSecHashCalcWithOps(struct file *file, char *digest)
{
    DIGEST_CTX ctx;  // -- struct hash_desc desc;
    // -- struct scatterlist sg[1];
    loff_t i_size, offset = 0;
    char *rbuf = NULL;
    int retval = -1;
    size_t left = 0;

    digest_ctx_init(&ctx);

    rbuf = kzalloc(PAGE_SIZE, GFP_KERNEL);  // -- __CALC_SIZE
    if (!rbuf) {
        retval = -ENOMEM;
        // -- printk("wlSecHashCalcWithOps 1 error\n");
        goto DONE;
    }

    i_size = i_size_read(get_file_dentry(file)->d_inode);
    left = i_size;
    while (left > 0) {
        int rbuf_len = 0;
        size_t count = 0;
        loff_t pos = offset;


        // -- rbuf_len = __sec_kern_read(file, (char *)rbuf, count, &pos);

        // -- count = (__CALC_SIZE > left ? left : __CALC_SIZE);
        count = PAGE_SIZE;

        // 这里不知道为什么用fp接口，会导致启动bash死机，暂时换成和读文件一样的接口尝试
        // rbuf_len = vfs_filp_read(file, (char *)rbuf, count, &pos);
        rbuf_len = __kread(file, pos, rbuf, count, 0);

        // -- rbuf_len = kernel_read(file, offset, rbuf, __CALC_SIZE);
        if (rbuf_len < 0) {
            retval = rbuf_len;
            // -- printk("wlSecHashCalcWithOps 4 error
            // (read:%d)(offset:%d)(count:%d)\n", rbuf_len, (int) offset,
            // (int)count);
            goto DONE;
        }

        // -- printk("<+> ++ vfs_filp_read count:%d rbuf_len:%d (total:%d)\n",
        // (int)count,  (int)rbuf_len, (int)i_size);
        if (rbuf_len == 0) {
            break;
        }
        offset += rbuf_len;
        left -= rbuf_len;

        digest_ctx_update(&ctx, rbuf, (int)rbuf_len);
    }

    digest_ctx_finish(&ctx, digest);
    retval = 0;
DONE:
    if (rbuf)
        kfree(rbuf);

    return retval;
}

// -- digest - length: 32 bytes;
int wlSecHashCalc(struct file *file, char *digest)
{
    int retval;
    retval = __wl_sec_hash_calc(file, digest);
    return retval;
}

int wlSelcHashCalcUserBuffer(void __user *umod, unsigned long len, char *digest)
{
    char *rbuf = NULL;
    int retval = -1;
    DIGEST_CTX ctx;  // -- struct hash_desc desc;
    unsigned int left = len;

    digest_ctx_init(&ctx);

    rbuf = kzalloc(__CALC_SIZE, GFP_KERNEL);
    if (!rbuf)
        goto DONE;
    while (left > 0) {
        int _cl = (left > __CALC_SIZE) ? __CALC_SIZE : left;
        if (0 != copy_from_user(rbuf, umod, _cl)) {
            // -- printk("[__wlSelcHashCalcUserBuffer] copy_from_user
            // failed......\n");
            retval = -EIO;
            goto DONE;
        }
        digest_ctx_update(&ctx, rbuf, (int)_cl);
        left -= _cl;
        umod = ((char *)umod + _cl);
    }
    digest_ctx_finish(&ctx, digest);
    retval = 0;
DONE:
    if (rbuf) {
        kfree(rbuf);
    }
    return retval;
}
