#include <project.h>
#include "common/general/comm_hash_table.h"
#include "image_table.h"
#include "common/hookmem.h"

// --------------------------------------------------------------------------
// -- static STR_HASH_TABLE m_imageTable;
// --------------------------------------------------------------------------

#pragma pack(1)
typedef struct __HASH_CONTENT {
    unsigned char imageIntegrity[INTEGRITY_LENGTH];
    // -- unsigned char path[260];
} HASH_CONTENT;
#pragma pack()

typedef struct __IMAGE_HASH_ITEM {
    LIST_HEAD head;
    HASH_CONTENT content;
    unsigned char type;  // -- 0
} IMAGE_HASH_ITEM;

static unsigned char *__hashItemValue(LIST_HEAD *head);
static void __hashItemRelease(LIST_HEAD *head);
static IMAGE_HASH_ITEM *__create_white_item(void);
static void __destroyHashItem(IMAGE_HASH_ITEM *hashItem);

// ---------------------------------------------------------------------------------------------------------

BOOLEAN wlImageHashTableRemove(IMAGE_HASH_TABLE *imageTable,
                               unsigned char *imageIntegrity)
{
    // -- IMAGE_HASH_ITEM * item;
    BOOLEAN retval = FALSE;

    spin_lock_bh(&imageTable->lock);

    if (imageTable->hash_table) {
        // -- BOOLEAN hashtb_remove_item(HASHTB * table, unsigned int * valPtr);
        retval = hashtb_remove_item(imageTable->hash_table,
                                    (unsigned int *)imageIntegrity);
    }

    spin_unlock_bh(&imageTable->lock);

    return retval;
}

BOOLEAN wlImageHashTableInsert(IMAGE_HASH_TABLE *imageTable,
                               unsigned char *imageIntegrity,
                               unsigned char type)
{
    BOOLEAN retval = FALSE;

    spin_lock_bh(&imageTable->lock);
    retval = wlImageHashTBInsert(imageTable->hash_table, imageIntegrity, type);
    spin_unlock_bh(&imageTable->lock);

    return retval;
}

BOOLEAN wlImageHashTBInsert(HASHTB *hashTable, unsigned char *imageIntegrity,
                            unsigned char type)
{
    IMAGE_HASH_ITEM *item;
    BOOLEAN retval = FALSE;

    item = __create_white_item();
    if (!item) {
        return FALSE;
    }

    memcpy(item->content.imageIntegrity, imageIntegrity, INTEGRITY_LENGTH);
    item->type = type;
    // -- strncpy(item->content.path, path, sizeof(item->content.path));
    retval = hashtb_insert_item(hashTable, &item->head,
                                (unsigned int *)imageIntegrity);
    if (!retval) {
        __destroyHashItem(item);
    } else {
        // -- printk("insert:%s\n", path);
        // -- __print_hex_data_only(imageIntegrity, INTEGRITY_LENGTH);
    }
    return retval;
}

BOOLEAN wlImageHashTableClearAll(IMAGE_HASH_TABLE *imageTable)
{
    spin_lock_bh(&imageTable->lock);
    if (imageTable->hash_table) {
        hashtb_remove_all(imageTable->hash_table);
    }
    spin_unlock_bh(&imageTable->lock);

    return TRUE;
}

static void __white_list_remove_type(IMAGE_HASH_TABLE *imageTable,
                                     unsigned char type)
{
    int i = 0;
    HASHTB *table = imageTable->hash_table;
    if (!table)
        return;

    if ((0 == table->bucket_count)) {
        return;
    }
    for (; i < (int)table->bucket_count; i++) {
        LIST_HEAD *header = &table->hash_list[i];
        LIST_HEAD *entry;

        entry = header->next;

        while (entry != header) {
            IMAGE_HASH_ITEM *hashItem =
                OBJECT_PTR(entry, IMAGE_HASH_ITEM, head);
            entry = entry->next;
            if (type == hashItem->type) {
                __list_remove_item(&hashItem->head);
                table->item_count--;
                __list_init(&hashItem->head);
                table->release_item(&hashItem->head);
            }
        }
    }

    return;
}

BOOLEAN wlImageHashTableClear(IMAGE_HASH_TABLE *imageTable, unsigned char type)
{
    spin_lock_bh(&imageTable->lock);
    __white_list_remove_type(imageTable, type);
    spin_unlock_bh(&imageTable->lock);

    return TRUE;
}

BOOLEAN image_table_find(HASHTB *ply_head, unsigned char *digest)
{
    IMAGE_HASH_ITEM *item;
    BOOLEAN retval = FALSE;
    LIST_HEAD *entry;

    if (ply_head) {
        entry = hashtb_find(ply_head, (unsigned int *)digest);
        if (entry) {
           retval = TRUE;
            item = OBJECT_PTR(entry, IMAGE_HASH_ITEM, head);
        }
    }

    return retval;    
}
                               

BOOLEAN wlImageHashTableLookup(IMAGE_HASH_TABLE *imageTable,
                               unsigned char *imageIntegrity)
{
    IMAGE_HASH_ITEM *item;
    BOOLEAN retval = FALSE;
    LIST_HEAD *entry;

    spin_lock_bh(&imageTable->lock);

    if (imageTable->hash_table) {
        entry =
            hashtb_find(imageTable->hash_table, (unsigned int *)imageIntegrity);
        if (entry) {
            retval = TRUE;
            item = OBJECT_PTR(entry, IMAGE_HASH_ITEM, head);
        }
    }

    spin_unlock_bh(&imageTable->lock);

    return retval;
}

static HASHTB *__imageHashTBCreate(unsigned int __bucketCount)
{
    HASHTB *retval = NULL;

    retval = (HASHTB *)kmalloc(sizeof(HASHTB), GFP_ATOMIC);
    if (!retval) {
        return NULL;
    }

#ifdef __LINUX_KMALLOC_CRASH_BUGS___
    if (0 != hashtb_init_ex(retval, __hashItemValue, __hashItemRelease,
                            INTEGRITY_LENGTH, rvm_alloc, rvm_free)) {
#else
    if (0 != hashtb_init_ex(retval, __hashItemValue, __hashItemRelease,
                            INTEGRITY_LENGTH, sec_vmalloc, sec_vfree)) {
#endif

        kfree(retval);

        retval = NULL;
        return retval;
    }
    while (__bucketCount > 10) {
        if (!hashtb_set_buckets(retval, __bucketCount)) {
        } else {
            break;
        }
        __bucketCount = (__bucketCount / 2);
    }
    if (__bucketCount <= 10) {
        kfree(retval);
        retval = NULL;
    }

    return retval;
}

HASHTB *wlImageHashTBCreateExt(unsigned int __bucketCount)
{
    return __imageHashTBCreate(__bucketCount);
}

HASHTB *wlImageHashTBCreate(void)
{
    HASHTB *retval = NULL;
    unsigned int __bucketCount = (4096 * 4);
    retval = __imageHashTBCreate(__bucketCount);
    return retval;
}

BOOLEAN wlImageHashTBUpdate(IMAGE_HASH_TABLE *imageTable, HASHTB *tb)
{
    HASHTB *old = NULL;


    spin_lock_bh(&imageTable->lock);

    if (imageTable->hash_table) {
        old = imageTable->hash_table;
    }
    imageTable->hash_table = tb;

    spin_unlock_bh(&imageTable->lock);

    if (old) {
        hashtb_destroy(old);
        kfree(old);
        old = NULL;
    }
    return TRUE;
}

void wlImageHashTBDestroy(HASHTB *tb)
{
    hashtb_destroy(tb);
    kfree(tb);
}

static BOOLEAN __imageHashTableInit(IMAGE_HASH_TABLE *imageTable,
                                    unsigned int __bucketCount)
{
    spin_lock_init(&imageTable->lock);
    imageTable->hash_table = wlImageHashTBCreateExt(__bucketCount);

    if (!imageTable->hash_table) {
        return FALSE;
    }
    return TRUE;
}

BOOLEAN wlImageHashTableInit(IMAGE_HASH_TABLE *imageTable)
{
    BOOLEAN retval;
    retval = __imageHashTableInit(imageTable, 4096 * 4);
    return retval;
}

BOOLEAN wlImageHashTableInitExt(IMAGE_HASH_TABLE *imageTable,
                                unsigned int __bucketCount)
{
    BOOLEAN retval;
    retval = __imageHashTableInit(imageTable, __bucketCount);
    return retval;
}

void wlImageHashTableDestroy(IMAGE_HASH_TABLE *imageTable)
{
    if (imageTable->hash_table) {
        hashtb_destroy(imageTable->hash_table);

        kfree(imageTable->hash_table);
        imageTable->hash_table = NULL;
    }

    return;
}

void __destroyHashItem(IMAGE_HASH_ITEM *hashItem)
{
    kfree(hashItem);
}

IMAGE_HASH_ITEM *__create_white_item(void)
{
    IMAGE_HASH_ITEM *retval = NULL;

    retval = (IMAGE_HASH_ITEM *)kmalloc(sizeof(IMAGE_HASH_ITEM), GFP_ATOMIC);
    if (!retval)
        return NULL;
    memset(retval, 0, sizeof(IMAGE_HASH_ITEM));
    return retval;
}

void __hashItemRelease(LIST_HEAD *head)
{
    IMAGE_HASH_ITEM *hashItem = OBJECT_PTR(head, IMAGE_HASH_ITEM, head);
    __destroyHashItem(hashItem);
    return;
}

unsigned char *__hashItemValue(LIST_HEAD *head)
{
    IMAGE_HASH_ITEM *hashItem = OBJECT_PTR(head, IMAGE_HASH_ITEM, head);
    unsigned char *ptr = (unsigned char *)hashItem->content.imageIntegrity;
    return (unsigned char *)ptr;
}
