#include <linux/binfmts.h>
#include <linux/file.h>
#include <linux/fsnotify.h>
#include <linux/mman.h>
#include <linux/module.h>
#include <linux/mount.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <project.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 10, 0)
#include <linux/syscalls.h>
#endif

#include <linux/dcache.h>
#include <linux/namei.h>

#include <linux/fs.h>
#include <linux/fs_struct.h>
#include <linux/uio.h>

#include "../common/kedbgout.h"
#include "../common/nfs_table.h"
#include "../common/pid_ctx.h"
#include "../common/sec_user_table.h"
#include "common/general/include/mod_share.h"
#include "include/misc/list_head.h"
#include "sec_krnl_misc.h"
#include "fullpath.h"

static int __get_process_name(char *exepath, int length, long pid);

#ifdef rcu_dereference_protected

#define __task_cred(_tsk) rcu_dereference_protected((_tsk)->cred, 1)
#else
#define __task_cred(_tsk) (_tsk)->cred
#endif

// -- #if LINUX_VERSION_CODE < KERNEL_VERSION(4,1,18)
void sec_put_task_struct(struct task_struct *t)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 0, 0)
    put_task_struct(t);
#else
    if (atomic_dec_and_test(&t->usage)) {
        if (NULL != __put_task_struct_func) {
            __put_task_struct_func(t);
        }
    }
#endif
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 1, 12)
// -- linux 4.2.1
// -- ssize_t __vfs_read_n(struct file *file, char __user *buf, size_t count,
// loff_t *pos)

static ssize_t __vfs_read_n(struct file *file, char __user *buf, size_t count,
                          loff_t *pos)
{
    ssize_t ret;

    if (file->f_op->read) {
        ret = file->f_op->read(file, buf, count, pos);
    } else {
        ret = do_sync_read(file, buf, count, pos);
    }
    return ret;
}

int vfs_filp_read(struct file *file, const char __user *buf, size_t count,
                  loff_t *pos)
{
    mm_segment_t old_fs;
    ssize_t ret;

    old_fs = get_fs();
    set_fs(get_ds());
    ret = __vfs_read_n(file, buf, count, pos);
    set_fs(old_fs);
    return ret;
}

#else

static ssize_t __new_sync_read(struct file *filp, char __user *buf, size_t len,
                               loff_t *ppos)
{
    struct iovec iov = {.iov_base = buf, .iov_len = len};
    struct kiocb kiocb;
    struct iov_iter iter;
    ssize_t ret;

    init_sync_kiocb(&kiocb, filp);
    kiocb.ki_pos = *ppos;
    iov_iter_init(&iter, READ, &iov, 1, len);

    //--ret = call_read_iter(filp, &kiocb, &iter);
    ret = filp->f_op->read_iter(&kiocb, &iter);
    BUG_ON(ret == -EIOCBQUEUED);
    *ppos = kiocb.ki_pos;

    return ret;
}

ssize_t __vfs_read_n(struct file *file, char __user *buf, size_t count,
                   loff_t *pos)
{
    if (file->f_op->read) {
        return file->f_op->read(file, buf, count, pos);
    } else if (file->f_op->read_iter) {
        return __new_sync_read(file, buf, count, pos);
    } else {
        return -EINVAL;
    }
}

// -- #define vfs_filp_read kernel_read
// -- ssize_t kernel_read(struct file *file, void *buf, size_t count, loff_t
// *pos)
int vfs_filp_read(struct file *file, const char __user *buf, size_t count,
                  loff_t *pos)
{

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 10, 0)
    mm_segment_t old_fs;
#endif
    ssize_t ret;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
    return kernel_read(file, buf, count, pos);

#else
    old_fs = get_fs();
// shy
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0)
    set_fs(KERNEL_DS);
#else
    set_fs(get_ds());
#endif
    ret = __vfs_read_n(file, buf, count, pos);
    set_fs(old_fs);

#endif
    return ret;
}

#endif

PID_CONTEXT *attachProcess(long pid)
{
    sec_pid_t pidObj;
    PID_CONTEXT *ctx = NULL;
    char * fullpath = NULL;

#ifdef __USE_PID_STRUCT_AS_KEY
    pidObj = sec_find_vpid(pid);
    if (!pidObj) {
        return NULL;
    }
#else
    pidObj = pid;
#endif
    ctx = sec_gen_insert(pidObj);
    if (ctx) {
        // __get_process_name(ctx->appFullName, sizeof(ctx->appFullName) - 1, pid);
        fullpath = get_fullpath_by_pid(pid);
        if (fullpath) {
//            tag_pr_info("===== got fullpath [%d-%s]\n", pid, fullpath);
            ctx->processID = pid;
            ctx->isThread = 0;
            query_user_info(ctx);
        }
    }

    return ctx;
}

static int __get_process_name(char *exepath, int length, long pid)
{
    int retval = -1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 28)
    struct path path;
#else
    struct nameidata nd;
#endif

    char link[128];
    char *ptr;
    int err = 0;

    exepath[0] = 0;
    snprintf(link, sizeof(link), "/proc/%lu/exe", pid);

    // maybe used in kern_path, but without check inside
    if (IS_ERR_OR_NULL(current->fs))
       return 0;
    if (IS_ERR_OR_NULL(current->fs->root.dentry))
       return 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 28)
    err = kern_path(link, LOOKUP_FOLLOW, &path);
#else
    err = path_lookup(link, LOOKUP_FOLLOW, &nd);
#endif
    if (err) {
        return -1;
    }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 28)
    ptr = d_path(&path, exepath, length);
    path_put(&path);

#else
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 18)
    struct path path;
    path.mnt = nd.mnt;
    path.dentry = nd.dentry;
    ptr = d_path(&path, exepath, length);
#else
    ptr = d_path(nd.dentry, nd.mnt, exepath, length);
#endif
    path_release(&nd);
#endif
    if (!IS_ERR(ptr)) {
        retval = 0;
        exepath[length - 1] = 0;
        strcpy(exepath, ptr);
    }

    return retval;
}


static char *__get_path_from_file_object(struct file *file, char *buffer,
                                         unsigned int length)
{
    char *ptr = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 0, 0)
// -- path = dentry_path_raw(vma->vm_file->f_dentry, page_ptr, PAGE_SIZE);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 0, 0)
    if (!file->f_path.dentry) {
        goto DONE;
    }
    ptr = dentry_path_raw(file->f_path.dentry, buffer, length - 1);
#else
    if (!file->f_dentry) {
        goto DONE;
    }
    ptr = dentry_path_raw(file->f_dentry, buffer, length - 1);
#endif
#else

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 18)
    // -- char *d_path(const struct path *path, char *buf, int buflen)
    ptr = d_path(&file->f_path, buffer, length - 1);
#else
    if (file->f_dentry == NULL || file->f_vfsmnt == NULL) {
        goto DONE;
    }
    // -- char * d_path(struct dentry *dentry, struct vfsmount *vfsmnt, char
    // *buf, int buflen)
    ptr = d_path(file->f_dentry, file->f_vfsmnt, buffer, length - 1);
#endif
#endif
    if (IS_ERR(ptr)) {
        ptr = NULL;
    }

DONE:
    return ptr;
}

char *get_path_from_file_object(struct file *file, char *buffer,
                                unsigned int length)
{
    return __get_path_from_file_object(file, buffer, length);
}
