#include <project.h>
#include <linux/uidgid.h>

#include "../common/spin_lock.h"
#include "sec_user_table.h"

typedef struct __USER_TABLE {
    spinlock_t lock;

    HASHTB *lookupTB;
} sec_user_table_t;

static sec_user_item_t *create_user_item(void);
// ------------------------------------------------------------------------------
sec_user_table_t g_userTable;

BOOLEAN sec_user_table_module_init(void)
{
    spin_lock_init(&g_userTable.lock);

    g_userTable.lookupTB = sec_user_table_create_hash_tb();
    return TRUE;
}

void sec_user_table_module_destroy(void)
{
    if (g_userTable.lookupTB) {
        sec_user_table_destroy_hash_tb(g_userTable.lookupTB);
        g_userTable.lookupTB = NULL;
    }
    return;
}

void sec_user_table_destroy_hash_tb(HASHTB *lookupTB)
{
    hashtb_remove_all(lookupTB);

    kfree(lookupTB);
    return;
}

static void __user_item_release(LIST_HEAD *entry)
{
    sec_user_item_t *ctx = OBJECT_PTR(entry, sec_user_item_t, head);
    comm_object_release(ctx);
    return;
}

static unsigned char *__user_item_value(LIST_HEAD *entry)
{
    sec_user_item_t *ctx = OBJECT_PTR(entry, sec_user_item_t, head);
    unsigned char *ptr = (unsigned char *)&ctx->sid_data;
    return (unsigned char *)ptr;
}

void sec_user_table_update_hash_tb(HASHTB *lookupTB)
{
    HASHTB *old = NULL;

    spin_lock(&g_userTable.lock);

    old = g_userTable.lookupTB;
    g_userTable.lookupTB = lookupTB;

    spin_unlock(&g_userTable.lock);

    sec_user_table_destroy_hash_tb(old);
    return;
}

HASHTB *sec_user_table_create_hash_tb(void)
{
    int bucket_count = 400;  // -- (4096 * 4);
    HASHTB *lookupTB = NULL;

    lookupTB = (HASHTB *)kmalloc(sizeof(HASHTB), GFP_ATOMIC);
    if (!lookupTB)
        return NULL;

    hashtb_init(lookupTB, __user_item_value, __user_item_release,
                sizeof(SID_STOR));

    while (bucket_count > 10) {
        if (!hashtb_set_buckets(lookupTB, bucket_count)) {
        } else {
            break;
        }
        bucket_count = (bucket_count / 2);
    }
    if (bucket_count <= 10) {
        kfree(lookupTB);
        return NULL;
    }

    return lookupTB;
}

void __destroy_user_item(sec_user_item_t *usrItem)
{
    kfree(usrItem);
    return;
}

// -- 返回值带一个引用计数；
sec_user_item_t *create_user_item(void)
{
    sec_user_item_t *usrItem;

    usrItem = (sec_user_item_t *)kmalloc(sizeof(sec_user_item_t), GFP_ATOMIC);

    if (!usrItem)
        return NULL;

    memset(usrItem, 0, sizeof(sec_user_item_t));

    usrItem->objHdr.destroy_ptr = (destroy_atom_ptr)__destroy_user_item;
    comm_object_addref(usrItem);
    return usrItem;
}

sec_user_item_t *sec_user_table_lookup_hash_tb(HASHTB *lookupTB, SID_STOR *sid)
{
    sec_user_item_t *retval = NULL;

    LIST_HEAD *entry;

    if (lookupTB == NULL)
        return NULL;

    entry = hashtb_find(lookupTB, (unsigned int *)sid);
    if (entry) {
        retval = OBJECT_PTR(entry, sec_user_item_t, head);
        comm_object_addref(retval);
    }

    return retval;
}

sec_user_item_t *sec_user_table_insert_hash_tb(HASHTB *lookupTB,
                                          SEC_USER_ITEM_STOR *sto)
{
    sec_user_item_t *item = NULL;
    SID_STOR *sid;

    item = create_user_item();
    if (!item)
        return NULL;

    sid = (SID_STOR *)&item->sid_data;

    sid->id = sto->sid.id;
    sid->model = sto->sid.model;
    item->rank = sto->rank;
    item->user_id = sto->user_id;
    item->integrityRank = sto->integrityRank;
    // item->model = sto->model;

    if (NULL != hashtb_find(lookupTB, (unsigned int *)&item->sid_data)) {
        comm_object_release(item);
        item = NULL;
    } else {
        if (!hashtb_insert_item(lookupTB, &item->head,
                                (unsigned int *)&item->sid_data)) {
            comm_object_release(item);
            item = NULL;
        } else {
            comm_object_addref(item);  // -- 给lookupTB使用;
        }
    }

    return item;
}

// -- 带有引用计数
sec_user_item_t *sec_user_table_insert(SEC_USER_ITEM_STOR *sto)
{
    sec_user_item_t *item = NULL;

    spin_lock(&g_userTable.lock);

    if (g_userTable.lookupTB) {
        item = sec_user_table_insert_hash_tb(g_userTable.lookupTB, sto);
    }

    spin_unlock(&g_userTable.lock);

    return item;
}

sec_user_item_t *sec_user_table_lookup(SID_STOR *sid)
{
    sec_user_item_t *retval = NULL;
    LIST_HEAD *entry;

    if (!g_userTable.lookupTB)
        return NULL;

    spin_lock(&g_userTable.lock);

    entry = hashtb_find(g_userTable.lookupTB, (unsigned int *)sid);
    if (entry) {
        retval = OBJECT_PTR(entry, sec_user_item_t, head);
        comm_object_addref(retval);
    }

    spin_unlock(&g_userTable.lock);

    return retval;
}

void sec_user_table_clear(void)
{
    if (!g_userTable.lookupTB)
        return;

    spin_lock(&g_userTable.lock);
    hashtb_remove_all(g_userTable.lookupTB);
    spin_unlock(&g_userTable.lock);

    return;
}

static __inline__ unsigned int __get_task_sid(struct task_struct *task)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 0)
#if __KUID_STRUCT__
    return __task_cred(task)->uid.val;
#else
    return __task_cred(task)->uid;
#endif
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 18)
    return task->uid;
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 27)
    return __task_cred(task)->uid;
#else
    return task_uid(task);
#endif
}

static __inline__ BOOLEAN __query_user_info(PID_CONTEXT *ctx)
{
    // -- 检查SID，判断是否是服务
    BOOLEAN bRet = TRUE;
    sec_user_item_t *userItem = NULL;

    // -- current->cred->euid;
    // -- current_sid()

    ctx->sid.id = __get_current_sid();

    // -- if(0 == ctx->sid.id){
    // --	ctx->sid.id = __get_current_sid();
    // -- }

    userItem = sec_user_table_lookup(&ctx->sid);
    if (userItem) {
        ctx->rank = userItem->rank;
        ctx->user_id = userItem->user_id;
        comm_object_release(userItem);
        userItem = NULL;
    }

    return bRet;
}

BOOLEAN query_user_info(PID_CONTEXT *ctx)
{
    return __query_user_info(ctx);
}

BOOLEAN query_specified_user_info(PID_CONTEXT *ctx, struct task_struct *tsk)
{
    BOOLEAN bRet = TRUE;
    sec_user_item_t *userItem = NULL;

    ctx->sid.id = __get_task_sid(tsk);

    userItem = sec_user_table_lookup(&ctx->sid);

    if (userItem) {
        ctx->rank = userItem->rank;
        ctx->user_id = userItem->user_id;
        comm_object_release(userItem);
        userItem = NULL;
    }

    return bRet;
}