#include <project.h>
#include "kern_symbol.h"
#include "lsm_if.h"
#include "probe.h"


#if defined(CONFIG_STRICT_KERNEL_RWX) && !defined(SECURITY_WRITABLE_HOOKS)
#include <linux/uaccess.h> /* probe_kernel_write() */
#define NEED_TO_CHECK_HOOKS_ARE_WRITABLE

#if defined(CONFIG_X86)
#define MAX_RO_PAGES 1024
static struct page *ro_pages[MAX_RO_PAGES];
static unsigned int ro_pages_len = 0;

static bool lsm_test_page_ro(void *addr)
{
    unsigned int i;
    int unused;
    struct page *page;

    page = (struct page *) lookup_address((unsigned long) addr, &unused);
    if (!page)
        return false;
    if (test_bit(_PAGE_BIT_RW, &(page->flags)))
        return true;
    for (i = 0; i < ro_pages_len; i++)
        if (page == ro_pages[i])
            return true;
    if (ro_pages_len == MAX_RO_PAGES)
        return false;
    ro_pages[ro_pages_len++] = page;
    return true;
}

static bool check_ro_pages(struct security_hook_heads *hooks, struct security_hook_list * lsm_hook, int nfssec_hook_size)
{
    int i;
    struct hlist_node *ilist, *last = NULL;
    struct hlist_head *list = (struct hlist_head *) hooks;

    if (list->first) {

    #if LINUX_VERSION_CODE >= KERNEL_VERSION(5,10,0)
        void *temp;
        copy_to_user(&temp, &list->first, sizeof(void *));
        if (temp == list->first)
            return true;
    #else 
        // means copy data from kernel to user safety in atomic
        if (!probe_kernel_write(&list->first, list->first, sizeof(void *)))
            return true;
    #endif 
    }

    for (i = 0; i < nfssec_hook_size; i++) {
        const unsigned int idx =
            ((unsigned long) lsm_hook[i].head - (unsigned long) hooks) / sizeof(struct list_head);
        struct hlist_head *self = &list[idx];

        last = NULL;
        for (ilist = self->first; ilist; ilist = ilist->next) {
            last = ilist;
        }

        if (last) {
            if (!lsm_test_page_ro(&last->next))
                return false;
        } else {
            if (!lsm_test_page_ro(&self->first))
                return false;
        }
    }
    return true;
}
#else
static bool check_ro_pages(struct security_hook_heads *hooks, struct security_hook_list * lsm_hook, int nfssec_hook_size)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,10,0)
    copy_to_user(&((struct list_head *) hooks)->next,
                 ((struct list_head *) hooks)->next,
                 sizeof(void *));
    return true;
#else
     return !probe_kernel_write(&((struct list_head *) hooks)->next,
                                ((struct list_head *) hooks)->next,
                                sizeof(void *));
#endif
}
#endif
#endif

// tood： 当编入内核时，想使用直接的lsm注册接口，但是行不通，会导致死机，未调通
// #ifdef NFSSEC_KMOD
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
int lsm_if_init(struct security_hook_list * lsm_hooks, int nfssec_hook_size, char *lsm)
{
    int idx;
    struct security_hook_heads * hooks = get_symbol("security_hook_heads");

    if (!hooks) {
        tag_pr_err("failed to probe_security_hook_heads\n");
        goto out;
    }

    for (idx = 0; idx < nfssec_hook_size; idx++) {
        lsm_hooks[idx].head = ((void *) hooks)
                                + ((unsigned long) lsm_hooks[idx].head)
                                - ((unsigned long) &probe_dummy_security_hook_heads);
    }
#if defined(NEED_TO_CHECK_HOOKS_ARE_WRITABLE)
    if (!check_ro_pages(hooks, lsm_hooks, nfssec_hook_size)) {
        // -- printk(KERN_INFO "Can't update security_hook_heads due to write protected. Retry with rodata=0 kernel command line option added.\n");
        tag_pr_err("failed to check_ro_pages\n");
        return -EINVAL;
    }
#endif

#if defined(NEED_TO_CHECK_HOOKS_ARE_WRITABLE) && defined(CONFIG_X86)
    for (idx = 0; idx < ro_pages_len; idx++)
        set_bit(_PAGE_BIT_RW, &(ro_pages[idx]->flags));
#endif
    for (idx = 0; idx < nfssec_hook_size; idx++) {
        add_hook(&lsm_hooks[idx]);
    }

#if defined(NEED_TO_CHECK_HOOKS_ARE_WRITABLE) && defined(CONFIG_X86)
    for (idx = 0; idx < ro_pages_len; idx++)
        clear_bit(_PAGE_BIT_RW, &(ro_pages[idx]->flags));
#endif
   
    return 0;
out:
    return -EINVAL;
}
#endif

int lsm_if_exit(struct security_hook_list *lsm_hooks, int nfssec_hook_size)
{
    int idx;

#if defined(NEED_TO_CHECK_HOOKS_ARE_WRITABLE) && defined(CONFIG_X86)
    for (idx = 0; idx < ro_pages_len; idx++)
        set_bit(_PAGE_BIT_RW, &(ro_pages[idx]->flags));
#endif

    for (idx = 0; idx < nfssec_hook_size; idx++){
        hlist_del_rcu(&lsm_hooks[idx].list);
    }   

#if defined(NEED_TO_CHECK_HOOKS_ARE_WRITABLE) && defined(CONFIG_X86)
    for (idx = 0; idx < ro_pages_len; idx++)
        clear_bit(_PAGE_BIT_RW, &(ro_pages[idx]->flags));
#endif

    return 0;   
}
// #else
// int __init lsm_if_init(struct security_hook_list * lsm_hooks, int nfssec_hook_size, char *lsm)
// {
//     tag_pr_debug("in lsm_if_init\n");
//         tag_pr_debug("in nfssec_hook_size[%d], lsm[%s]\n", nfssec_hook_size, lsm);

//     // security_add_hooks(lsm_hooks, nfssec_hook_size, lsm);
//     tag_pr_debug("after lsm_if_init\n");

//     return 0;
// }
// int lsm_if_exit(struct security_hook_list * lsm_hooks, int nfssec_hook_size)
// {
//     // security_delete_hooks(lsm_hooks, nfssec_hook_size);
//     return 0;
// }
// #endif