#include <project.h>
#include "inode_cache.h"

inode_map_t g_inode_map;

static void __destroy_inode_cache_item(inode_cache_item_t *item)
{
    kfree(item);
    return;
}

// 返回值带一个引用计数
inode_cache_item_t *inode_cache_item_new(void)
{
    inode_cache_item_t *item;

    item = (inode_cache_item_t *)kmalloc(sizeof(inode_cache_item_t), GFP_ATOMIC);
    if (!item)
        return NULL;

    memset(item, 0, sizeof(inode_cache_item_t));

    spin_lock_init(&item->lock);

    item->objHdr.destroy_ptr = (destroy_atom_ptr)__destroy_inode_cache_item;
    __list_init(&item->hhead);
    __list_init(&item->chead);
    comm_object_addref(item);

    return item;
}

static void __i_node_cache_item_release(LIST_HEAD *entry)
{
    inode_cache_item_t *ctx = OBJECT_PTR(entry, inode_cache_item_t, hhead);
    comm_object_release(ctx);
    return;
}

static unsigned char *__i_node_cache_item_value(LIST_HEAD *entry)
{
    inode_cache_item_t *ctx = OBJECT_PTR(entry, inode_cache_item_t, hhead);
    unsigned char *ptr = (unsigned char *)&ctx->key;

    return (unsigned char *)ptr;
}

static BOOLEAN __i_node_map_init(inode_map_t *table)
{
    BOOLEAN retval = FALSE;
    int bucket_count = (4096 * 4);

    spin_lock_init(&table->lock);

    __list_init(&table->cacheList);

    hashtb_init(&table->lookupTB, __i_node_cache_item_value,
                __i_node_cache_item_release, sizeof(inode_key_t));

    while (bucket_count > 10) {
        if (!hashtb_set_buckets(&table->lookupTB, bucket_count)) {
        } else {
            break;
        }
        bucket_count = (bucket_count / 2);
    }
    
    if (bucket_count <= 10)
        return FALSE;
    retval = TRUE;

    return retval;
}

static void __i_node_map_destroy(inode_map_t *table)
{
    while (!__list_is_empty(&table->cacheList)) {
        LIST_HEAD *entry;
        inode_cache_item_t *item;
        __list_remove_tail(&table->cacheList, entry);
        item = OBJECT_PTR(entry, inode_cache_item_t, chead);
        comm_object_release(item);
    }

    hashtb_destroy(&table->lookupTB);
    return;
}

static void __i_node_map_update(inode_map_t *table, inode_cache_item_t *item)
{
    __list_remove_item(&item->chead);
    __list_add_tail(&table->cacheList, &item->chead);
    return;
}

static inode_cache_item_t *__i_node_map_lookup(inode_map_t *table, dev_t diskDev,
                                             unsigned long i_ino)
{
    inode_cache_item_t *item = NULL;
    inode_key_t key;
    LIST_HEAD *entry;

    key.diskDev = diskDev;
    key.i_ino = i_ino;

    spin_lock_bh(&table->lock);

    entry = hashtb_find(&table->lookupTB, (unsigned int *)&key);
    if (entry) {
        item = OBJECT_PTR(entry, inode_cache_item_t, hhead);
        comm_object_addref(item);
        __i_node_map_update(table, item);
    }

    // delete only one item
    if (hash_get_item_count(&table->lookupTB) > __MAX_CACHE_COUNT) {
        inode_cache_item_t *abortItem = NULL;
        LIST_HEAD *__abort_entry;
        __list_remove_head(&table->cacheList, __abort_entry);

        abortItem = OBJECT_PTR(__abort_entry, inode_cache_item_t, chead);
        hashtb_remove_item_entry(&table->lookupTB, &abortItem->hhead);
        comm_object_release(abortItem);
    }

    spin_unlock_bh(&table->lock);

    return item;
}

inode_cache_item_t *__i_node_map_insert(inode_map_t *table, dev_t diskDev,
                                      unsigned long i_ino, loff_t i_size,
                                      timespec_t i_mtime,
                                      timespec_t i_ctime,
                                      unsigned char bytes[INTEGRITY_LENGTH])
{
    inode_cache_item_t *item = NULL;
    LIST_HEAD *entry;
    inode_cache_item_t *oldItem = NULL;

    item = inode_cache_item_new();
    if (!item)
        return NULL;

    item->key.diskDev = diskDev;
    item->key.i_ino = i_ino;
    item->i_size = i_size;
    item->i_mtime = i_mtime;
    item->i_ctime = i_ctime;
    memcpy(item->bytes, bytes, INTEGRITY_LENGTH);

    spin_lock_bh(&table->lock);
    entry = hashtb_find(&table->lookupTB, (unsigned int *)&item->key);
    if (entry) {
        oldItem = OBJECT_PTR(entry, inode_cache_item_t, hhead);

        // release a ref-count
        hashtb_remove_item_entry(&table->lookupTB, entry);  

        // a ref-count keep for oldItem itself!
        __list_remove_item(&oldItem->chead);  
    }

    if (!hashtb_insert_item(&table->lookupTB, &item->hhead,
                            (unsigned int *)&item->key)) {
        comm_object_release(item);
        item = NULL;
        goto DONE;
    } else {
        comm_object_addref(item);  // 给lookupTB使用;
    }
    __list_add_tail(&table->cacheList, &item->chead);
    comm_object_addref(item);  // 给cacheList使用;

DONE:
    spin_unlock_bh(&table->lock);

    if (oldItem) 
        comm_object_release(oldItem);  // ref for cacheList;

    return item;
}

inode_cache_item_t *inode_cache_lookup(dev_t diskDev, unsigned long i_ino)
{
    return __i_node_map_lookup(&g_inode_map, diskDev, i_ino);
}

inode_cache_item_t *inode_cache_insert(dev_t diskDev, unsigned long i_ino,
                                     loff_t i_size, timespec_t i_mtime,
                                     timespec_t i_ctime,
                                     unsigned char bytes[INTEGRITY_LENGTH])
{
    inode_cache_item_t *item;
    item = __i_node_map_insert(&g_inode_map, diskDev, i_ino, i_size, i_mtime,
                               i_ctime, bytes);
    return item;
}

BOOLEAN inode_cache_module_init(void)
{
    return __i_node_map_init(&g_inode_map);
}

void inode_cache_module_destroy(void)
{
    __i_node_map_destroy(&g_inode_map);
    return;
}
