#include <project.h>
#include <linux/xattr.h>

#include "lsm_impl.h"
#include "common/version_type.h"
#include "common/sym_hook/lsm_if.h"
#include "common/config/sec_ply_if.h"
#include "common/elf_assist.h"
#include "common/hash_calc.h"
#include "common/image_table.h"
#include "common/script_type.h"
#include "common/fullpath.h"
#include "common/audit/sec_audit.h"
#include "inode_cache.h"

// 基于扩展标签 实现 完整性策略
// else， 表示基于二进制策略实现
#define IMA_PLY_BY_ATTR


#ifdef IMA_PLY_BY_ATTR
#include "ima_xattr.h"
#endif

// check for file ima
// return 0/-1 succecc/failed
static int ima_file_digest(struct inode *inode, struct file *file, int ima_type, char *digest)
{
    int retval = 0;
    int ret = 0;

    unsigned char bCacheValid = 0;

#ifdef __USE_IMAGE_TABLE__
    timespec_t mtime;
    timespec_t ctime;

    if (inode->i_sb) {
        inode_cache_item_t *cacheItem = NULL;
        cacheItem = inode_cache_lookup(inode->i_sb->s_dev, inode->i_ino);
        if (cacheItem) {
            if (cacheItem->i_size == (loff_t)inode->i_size) {
                mtime = inode_get_mtime(inode);
                ctime = inode_get_ctime(inode);

                if (__IS_TIMESPEC_EQUAL(&mtime, &cacheItem->i_mtime) &&
                    __IS_TIMESPEC_EQUAL(&ctime, &cacheItem->i_ctime)) {
                    memcpy(digest, cacheItem->bytes, INTEGRITY_LENGTH);
                    bCacheValid = 1;
                }
            }
            comm_object_release(cacheItem);
        }
    }
#endif

    if (!bCacheValid) {
        if (ima_type == IMA_FILE_SO) {
            if (!is_so_file(file, 0)) {
                goto DONE;
            }
        } else if (ima_type == IMA_FILE_KO) {
            if (!is_ko_file(file, 0)) {
                goto DONE;
            }
        } else if (ima_type == IMA_FILE_SCRIPT) {
            if (!is_script_file(file)) {
                goto DONE;
            }
        } else { // IMA_FILE_BIN
            if (!is_bin_file(file, 0)) {
                goto DONE;
            }
        }

        // todo: 计算文件hash，待封装
        ret = wlSecHashCalc(file, digest);
        if (ret) {
            goto DONE;
        }

#ifdef __USE_IMAGE_TABLE__
        if (inode->i_sb) {
            inode_cache_item_t *cacheItem = NULL;

            cacheItem = inode_cache_insert(
                inode->i_sb->s_dev, inode->i_ino, inode->i_size, mtime,
                ctime, (unsigned char *)digest);
            if (cacheItem)
                comm_object_release(cacheItem);
        }
#endif
    }

DONE:
    return retval;
}

static int impl_bprm_check_security(struct linux_binprm *bprm)
{
    int retval = 0;
    int ret = 0;

    struct file *file = bprm->file;
    struct inode *inode;
    char digest[INTEGRITY_LENGTH] = {0};
    int result = 0;

#ifdef IMA_PLY_BY_ATTR
    ima_xattr_t xattr;
#endif

    sec_ply_if_t * ply_ima = NULL;

    // 获取全局的plf_if
    ply_ima = get_ply_if(SEC_PLY_IMA);
    if (!ply_ima) 
        goto DONE;

    if (!is_sw_enable(&ply_ima->sw)) 
        goto DONE;

#ifndef IMA_PLY_BY_ATTR
    if (is_sec_ply_empty(ply_ima))
        goto DONE;
#endif

    if (!file || !get_file_dentry(file) || !get_file_dentry(file)->d_inode)
        goto DONE;

    inode = get_file_dentry(file)->d_inode;

    if (!S_ISREG(inode->i_mode))
        goto DONE;

#ifdef IMA_PLY_BY_ATTR
    // todo: 判断文件类型

    ret = ima_file_digest(inode, file, IMA_FILE_BIN, digest);
    if(ret || digest[0] == '\0') {
        ret = ima_file_digest(inode, file, IMA_FILE_SCRIPT, digest);
        if(ret || digest[0] == '\0') 
            goto DONE;
    }

	ret = __vfs_getxattr(get_file_dentry(file), inode, NFSSEC_IMA_XATTR_NAME, &xattr, sizeof(ima_xattr_t));
    if (ret != sizeof(ima_xattr_t)) {
        tag_pr_info("get xattr error!\n");
        result = -1;
        goto DONE;
    }

    // digest_show(xattr.hash, sizeof(xattr.hash));

	ret = memcmp(xattr.hash, digest, sizeof(digest));
    if (ret) {
        tag_pr_info("digest differ!\n");
        result = -1;
        goto DONE;
    }

#else
    ret = ima_file_digest(inode, file, IMA_FILE_BIN, digest);
    if(ret || digest[0] == '\0') {
        ret = ima_file_digest(inode, file, IMA_FILE_SCRIPT, digest);
        if(ret || digest[0] == '\0') 
            goto DONE;
    }

    // digest_show(digest, sizeof(digest));

    mutex_lock(&ply_ima->sec_ply_lock);
    ret = image_table_find(ply_ima->ply_head, digest);
    mutex_unlock(&ply_ima->sec_ply_lock);
    if (ret == FALSE) {
        tag_pr_debug("bin digest not found!!!\n");
        result = -1;
        goto DONE;
    }
#endif

DONE:
    // 开关为开，且判断结果为失败，则返回错误
    if (result && ply_ima->sw.b_ctrl) {
        retval = -1;
    }

    if (result && ply_ima->sw.b_audit) {
        char *page_ptr = NULL;
        char *fullpath = NULL;

        page_ptr = (char *)__get_free_page(GFP_ATOMIC);
        fullpath = fullpath_by_file(file, page_ptr, PAGE_SIZE);
        if (fullpath) {
            tag_pr_info("fullpath[%s]\n", fullpath);
            AUDIT_STATIC_MEASURE(current->comm, current->pid, AUDIT_OPE_FILE_EXEC,  \
                              retval ? AUDIT_RESULT_FAILED : AUDIT_RESULT_SUCCESS, fullpath, digest);
        }
            
        if (page_ptr)
            free_page((unsigned long)page_ptr);
    }

    return retval;
}

static int impl_mmap_file(struct file *file, unsigned long reqprot,
                         unsigned long prot, unsigned long flags)
{
    int retval = 0;
    int ret = 0;

    struct inode *inode;
    char digest[INTEGRITY_LENGTH] = {0};
    int result = 0;

#ifdef IMA_PLY_BY_ATTR
    ima_xattr_t xattr;
#endif

    sec_ply_if_t * ply_ima = NULL;

    // 获取全局的plf_if
    ply_ima = get_ply_if(SEC_PLY_IMA);
    if (!ply_ima) 
        goto DONE;

    if (!is_sw_enable(&ply_ima->sw)) 
        goto DONE;

#ifndef IMA_PLY_BY_ATTR
    if (is_sec_ply_empty(ply_ima))
        goto DONE;
#endif

    if (!file || !get_file_dentry(file) || !get_file_dentry(file)->d_inode)
        goto DONE;

    inode = get_file_dentry(file)->d_inode;

    if (!S_ISREG(inode->i_mode))
        goto DONE;

#ifdef IMA_PLY_BY_ATTR
    ret = ima_file_digest(inode, file, IMA_FILE_SO, digest);
    if(ret || digest[0] == '\0') {
        goto DONE;
    }

	ret = __vfs_getxattr(get_file_dentry(file), inode, NFSSEC_IMA_XATTR_NAME, &xattr, sizeof(ima_xattr_t));
    if (ret != sizeof(ima_xattr_t)) {
        tag_pr_debug("get xattr error!\n");
        result = -1;
        goto DONE;
    }

	ret = memcmp(xattr.hash, digest, sizeof(digest));
    if (ret) {
        result = -1;
        goto DONE;
    }

#else
    ret = ima_file_digest(inode, file, IMA_FILE_SO, digest);
    if(ret || digest[0] == '\0') {
        goto DONE;
    }

    // digest_show(digest, sizeof(digest));

    mutex_lock(&ply_ima->sec_ply_lock);
    ret = image_table_find(ply_ima->ply_head, digest);
    mutex_unlock(&ply_ima->sec_ply_lock);
    if (ret == FALSE) {
        tag_pr_debug("bin digest not found!!!\n");
        retval = -1;
        goto DONE;
    }
#endif

DONE:
    // 开关为开，且判断结果为失败，则返回错误
    if (result && ply_ima->sw.b_ctrl) {
        retval = -1;
    }
    
    if (result && ply_ima->sw.b_audit) {
        char *page_ptr = NULL;
        char *fullpath = NULL;

        page_ptr = (char *)__get_free_page(GFP_ATOMIC);
        fullpath = fullpath_by_file(file, page_ptr, PAGE_SIZE);
        if (fullpath) {
            tag_pr_info("fullpath[%s]\n", fullpath);
            AUDIT_STATIC_MEASURE(current->comm, current->pid, AUDIT_OPE_FILE_LOAD_LIB,  \
                              retval ? AUDIT_RESULT_FAILED : AUDIT_RESULT_SUCCESS, fullpath, digest);
        }
            
        if (page_ptr)
            free_page((unsigned long)page_ptr);
    }

    return retval;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0)
static int impl_kernel_module_from_file(struct file *file, enum kernel_read_file_id id,
                                bool contents)
#else  // 4.19 ok
static int impl_kernel_module_from_file(struct file *file, enum kernel_read_file_id id)
#endif
{
    int retval = 0;
    int ret = 0;

    struct inode *inode;
    char digest[INTEGRITY_LENGTH] = {0};
    int result = 0;

#ifdef IMA_PLY_BY_ATTR
    ima_xattr_t xattr;
#endif

    sec_ply_if_t * ply_ima = NULL;

    // 获取全局的plf_if
    ply_ima = get_ply_if(SEC_PLY_IMA);
    if (!ply_ima) 
        goto DONE;

    if (!is_sw_enable(&ply_ima->sw)) 
        goto DONE;

#ifndef IMA_PLY_BY_ATTR
    if (is_sec_ply_empty(ply_ima))
        goto DONE;
#endif

    if (!file || !get_file_dentry(file) || !get_file_dentry(file)->d_inode)
        goto DONE;

    inode = get_file_dentry(file)->d_inode;

    if (!S_ISREG(inode->i_mode))
        goto DONE;

#ifdef IMA_PLY_BY_ATTR
    ret = wlSecHashCalc(file, digest);
    if (ret) {
        goto DONE;
    }

	ret = __vfs_getxattr(get_file_dentry(file), inode, NFSSEC_IMA_XATTR_NAME, &xattr, sizeof(ima_xattr_t));
    if (ret != sizeof(ima_xattr_t)) {
        tag_pr_debug("get xattr error!\n");
        result = -1;
        goto DONE;
    }

	ret = memcmp(xattr.hash, digest, sizeof(digest));
    if (ret) {
        result = -1;
        goto DONE;
    }

#else
    // todo: 计算文件hash，待封装
    ret = wlSecHashCalc(file, digest);
    if (ret) {
        goto DONE;
    }

    // todo: 对ko的分析似乎不准确，暂时先直接计算
    // ret = ima_file_digest(inode, file, IMA_FILE_KO, digest);
    // if(ret || digest[0] == '\0') {
    //     goto DONE;
    // }

    // digest_show(digest, sizeof(digest));

    mutex_lock(&ply_ima->sec_ply_lock);
    ret = image_table_find(ply_ima->ply_head, digest);
    mutex_unlock(&ply_ima->sec_ply_lock);
    if (ret == FALSE) {
        tag_pr_debug("bin digest not found!!!\n");
        retval = -1;
        goto DONE;
    }
#endif

DONE:
    // 开关为开，且判断结果为失败，则返回错误
    if (result && ply_ima->sw.b_ctrl) {
        retval = -1;
    }
  
    if (result && ply_ima->sw.b_audit) {
        char *page_ptr = NULL;
        char *fullpath = NULL;

        page_ptr = (char *)__get_free_page(GFP_ATOMIC);
        fullpath = fullpath_by_file(file, page_ptr, PAGE_SIZE);
        if (fullpath) {
            tag_pr_info("fullpath[%s]\n", fullpath);
            AUDIT_STATIC_MEASURE(current->comm, current->pid, AUDIT_OPE_FILE_INSMOD,  \
                              retval ? AUDIT_RESULT_FAILED : AUDIT_RESULT_SUCCESS, fullpath, digest);
        }
            
        if (page_ptr)
            free_page((unsigned long)page_ptr);
    }

    return retval;
}

#ifdef __SELF_ATTR_PROTECT__
// do check for self xattr
// return: [0/-1] success/failed
static int __self_xattr_check(const char * name)
{
    int retval = 0;
    sec_ply_if_t * ply_ima = NULL;

    ply_ima = get_ply_if(SEC_PLY_IMA);
    if (!ply_ima || !is_sw_enable(&ply_ima->sw)) 
        goto DONE;

    if (!strncmp(name, NFSSEC_IMA_XATTR_NAME, strlen(NFSSEC_IMA_XATTR_NAME)) &&
        strcmp(current->comm, "nfssecd"))
        retval = -1;
DONE:
    return retval;
}

static int impl_inode_getxattr(struct dentry *dentry, const char *name)
{
    return __self_xattr_check(name);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
static int impl_inode_setxattr(struct mnt_idmap *idmap,
	 struct dentry *dentry, const char *name, const void *value,
	 size_t size, int flags)
#else
static int impl_inode_setxattr(struct dentry *dentry, const char *name,
                     const void *value, size_t size, int flags)
#endif
{
    return __self_xattr_check(name);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
static int impl_inode_removexattr(struct mnt_idmap *idmap,
	 struct dentry *dentry, const char *name)
#else
static int impl_inode_removexattr(struct dentry *dentry, const char *name)
#endif
{
    return __self_xattr_check(name);
}
#endif

struct security_hook_list nfssec_ima_hooks[] = {

	MY_HOOK_INIT(bprm_check_security, impl_bprm_check_security),
    MY_HOOK_INIT(mmap_file, impl_mmap_file),
    MY_HOOK_INIT(kernel_read_file, impl_kernel_module_from_file),

#ifdef __SELF_ATTR_PROTECT__
	MY_HOOK_INIT(inode_setxattr, impl_inode_setxattr),
	MY_HOOK_INIT(inode_getxattr, impl_inode_getxattr),
	MY_HOOK_INIT(inode_removexattr, impl_inode_removexattr),
#endif

};

int nfssec_ima_hook_size = sizeof(nfssec_ima_hooks) / sizeof(nfssec_ima_hooks[0]);
