#include "list.h"

#include <linux/slab.h>
#include "debug_log_queue.h"

#define USE_DEBUG (0)
#define USE_DEBUG_FOR_EACH (0)

#ifndef BOOL
#define BOOL int
#endif

#ifndef FALSE
#define FALSE (0)
#define TRUE (!FALSE)
#endif

#define lock(list, flags) if (list->is_thread_safe_) \
                            spin_lock_irqsave(&list->lock_, flags)

#define unlock(list, flags) if (list->is_thread_safe_) \
                            spin_unlock_irqrestore(&list->lock_, flags);


static int xc_list_pop_nolock(struct xc_list *list, char **data);

static int xc_list_node_init(struct xc_list_node *node)
{
    node->next_ = node;
    node->rear_ = node;
    node->data_ = NULL;

    return 0;
}

int xc_list_init(struct xc_list *list
                , unsigned int maximum)
{
    struct xc_list_node *head = NULL;

    list->head_             = NULL;
    list->maximum_          = maximum;
    list->len_              = 0;
    list->name_[0]          = 0;
    list->is_thread_safe_   = 1;

    head = kmalloc(sizeof(struct xc_list_node), GFP_ATOMIC);
	if(!head) {
        printk(KERN_WARNING "[net_audit_module] Failed to allocate memory.\n");
        return -1;
    }
    xc_list_node_init(head);
    list->head_     = head;

    spin_lock_init(&list->lock_);

    return 0;
}

int xc_list_set_name(struct xc_list *list, const char *name)
{
    unsigned long flags = 0;

    lock(list, flags);
    
    strncpy(list->name_, name, sizeof (list->name_) - 1);

    unlock(list, flags);

    return 0;
}

int xc_list_set_thread_safe(struct xc_list *list, int enable)
{
    list->is_thread_safe_ = enable;

    return 0;
}

static int xc_list_is_initialized(struct xc_list *list)
{
    if (!list->head_) {
        printk(KERN_WARNING "[net_audit_module] List initialization failed!.\n");
        return -1;
    }

    return 0;
}

static int xc_list_is_initialized_and_unlock(struct xc_list *list)
{
    int ret = xc_list_is_initialized(list);
    unsigned long flags = 0;

    if (ret) {
        unlock(list, flags);
    }

    return ret;
}

int xc_list_push(struct xc_list *list, const void *data, unsigned int len)
{
    struct xc_list_node *q = NULL;
    struct xc_list_node *p = NULL;
    char *pop_data = NULL;
    unsigned long flags = 0;

    lock(list, flags);

    if (xc_list_is_initialized_and_unlock(list)) {
        return -1;
    }

    q = kmalloc(sizeof(struct xc_list_node), GFP_ATOMIC);
	if(!q) {
        unlock(list, flags);
        printk(KERN_WARNING "[net_audit_module] Failed to allocate memory.\n");
        return -1;
    }

    q->data_ = kmalloc(len + 1, GFP_ATOMIC);
    if(!(q->data_)) {
        unlock(list, flags);
        printk(KERN_WARNING "[net_audit_module] Failed to allocate memory.\n");
        return -1;
    }
    *(q->data_ + len) = 0;
    memcpy(q->data_, data, len);

    // full
    if (list->maximum_ && list->len_ >= list->maximum_) {
        printk_log(USE_DEBUG, "[net_audit_module] List:%s list full! len:%u.\n",list->name_ , list->len_);
        xc_list_pop_nolock(list, &pop_data);
        if (pop_data) {
            kfree(pop_data);
        }
    }

    p = list->head_->rear_;
    q->next_ = p->next_;
    p->next_->rear_ = q;
    p->next_ = q;
    q->rear_ = p;

    ++list->len_;
    printk_log(USE_DEBUG, "[net_audit_module] list:%s Push, len:%u.\n",list->name_ , list->len_);

    unlock(list, flags);

    return 0;
}

int xc_list_push_shallow_copy(struct xc_list *list, void *data)
{
    struct xc_list_node *q = NULL;
    struct xc_list_node *p = NULL;
    char *pop_data = NULL;
    unsigned long flags = 0;

    lock(list, flags);

    if (xc_list_is_initialized_and_unlock(list)) {
        return -1;
    }

    q = kmalloc(sizeof(struct xc_list_node), GFP_ATOMIC);
	if(!q) {
        unlock(list, flags);
        printk(KERN_WARNING "[net_audit_module] Failed to allocate memory.\n");
        return -1;
    }

    q->data_ = data;

    // full
    if (list->maximum_ && list->len_ >= list->maximum_) {
        printk_log(USE_DEBUG, "[net_audit_module] List:%s list full! len:%u.\n",list->name_ , list->len_);
        xc_list_pop_nolock(list, &pop_data);
        if (pop_data) {
            kfree(pop_data);
        }
    }

    p = list->head_->rear_;
    q->next_ = p->next_;
    p->next_->rear_ = q;
    p->next_ = q;
    q->rear_ = p;

    ++list->len_;
    printk_log(USE_DEBUG, "[net_audit_module] list:%s Push, len:%u.\n",list->name_ , list->len_);

    unlock(list, flags);

    return 0;
}

int xc_list_push_str(struct xc_list *list, const char *str)
{
    unsigned int len = strlen(str);
    return xc_list_push(list, str, len);
}

static int xc_list_pop_nolock(struct xc_list *list, char **data)
{
    struct xc_list_node *q = NULL;
    struct xc_list_node *p = NULL;

    *data = NULL;

    if (xc_list_is_initialized(list)) {
        return -1;
    }

    q = list->head_->next_;
    p = list->head_;

    if (list->len_ <= 0 || q->next_ == q) {
        // printk_log(USE_DEBUG, "[net_audit_module] list:%s List is empty.\n", list->name_);
        return -1; // empty
    }

    p->next_ = q->next_;
    q->next_->rear_ = p;

    *data = q->data_;
    --list->len_;

    kfree(q);
    printk_log(USE_DEBUG, "[net_audit_module] list:%s Pop, len:%u.\n", list->name_ , list->len_);

    return 0;
}

int xc_list_pop(struct xc_list *list, char **data)
{
    int ret = 0;
    unsigned long flags = 0;

    lock(list, flags);
    ret = xc_list_pop_nolock(list, data);
    unlock(list, flags);

    return ret;
}

int xc_list_clear(struct xc_list *list)
{
    struct xc_list_node *p = NULL;
    struct xc_list_node *q = NULL;
    unsigned long flags = 0;

    lock(list, flags);

    if (xc_list_is_initialized_and_unlock(list)) {
        return -1;
    }

    p = list->head_->next_;

    while (p != list->head_) {
        q = p->next_;
        if (p->data_) {
            kfree(p->data_);
        }
        kfree(p);
        p = q;
    }

    xc_list_node_init(list->head_);
    list->len_ = 0;
    printk_log(USE_DEBUG, "[net_audit_module] list:%s Clear, len:%u.\n", list->name_ , list->len_);

    unlock(list, flags);

    return 0;
}

static int xc_list_remove_if_internal(struct xc_list *list
                                    , xc_list_find_fun fun
                                    , const char *condition
                                    , BOOL remove_one)
{
    struct xc_list_node *p = NULL;
    struct xc_list_node *q = NULL;
    unsigned long flags = 0;

    lock(list, flags);

    if (xc_list_is_initialized_and_unlock(list)) {
        return -1;
    }

    p = list->head_;

    while(p->next_ != list->head_) {
        if(!fun(p->next_->data_, condition)) {
            q = p->next_;
            p->next_ = q->next_;
            q->next_->rear_ = p;
            --list->len_;
            kfree(q->data_);
            kfree(q);
            printk_log(USE_DEBUG, "[net_audit_module] list:%s Remove, len:%u.\n", list->name_ , list->len_);

            if (remove_one) {
                unlock(list, flags);
                return 0;
            }
        }

        p = p->next_;
    }

    unlock(list, flags);

    return -1;
}

int xc_list_remove_if(struct xc_list *list, xc_list_find_fun fun, const char *condition)
{
    return xc_list_remove_if_internal(list, fun, condition, TRUE);
}

int xc_list_remove_if_all(struct xc_list *list, xc_list_find_fun fun, const char *condition)
{
    return xc_list_remove_if_internal(list, fun, condition, FALSE);
}

int xc_list_remove_back(struct xc_list *list)
{
    struct xc_list_node *p = NULL;
    struct xc_list_node *q = NULL;
    unsigned long flags = 0;

    lock(list, flags);

    if (xc_list_is_initialized_and_unlock(list)) {
        return -1;
    }

    // empty
    if (!list->len_) {
        unlock(list, flags);
        return -1;
    }

    p = list->head_->rear_->rear_;
    q = list->head_->rear_;
    p->next_ = q->next_;
    q->next_->rear_ = p;
    --list->len_;
    kfree(q->data_);
    kfree(q);

    printk_log(USE_DEBUG, "[net_audit_module] list:%s Remove back, len:%u.\n", list->name_ , list->len_);
    unlock(list, flags);

    return 0;
}

int xc_list_for_each(struct xc_list *list, xc_list_for_each_fun fun
                    , char *opt, unsigned int *len)
{
    struct xc_list_node *p = NULL;
    unsigned long flags = 0;
    
    lock(list, flags);

    if (xc_list_is_initialized_and_unlock(list)) {
        return -1;
    }

    *len = list->len_;
    if (!list->len_) {
        unlock(list, flags);
        return -1;
    }

    p = list->head_->next_;

#if USE_DEBUG_FOR_EACH
    int pos = 0;
    char txt[128];
#endif
    while(p != list->head_){
#if USE_DEBUG_FOR_EACH
        ++pos;
        snprintf(txt, sizeof (txt) - 1, "list:%s For each, pos:%d len:%u.\n"
        , list->name_, pos, list->len_);
        debug_log_queue_push_add_title(txt);
#endif
        if (!fun(p->data_, opt)) {
            unlock(list, flags);
            return 0;
        }

        p = p->next_;
    }

    unlock(list, flags);

    return 0;
}

int xc_list_destroy(struct xc_list *list)
{
    if (xc_list_is_initialized(list)) {
        return -1;
    }

    xc_list_clear(list);
    kfree(list->head_);
    list->head_ = NULL;

    printk_log(USE_DEBUG, "[net_audit_module] list:%s Destroy!\n", list->name_);
    
    return 0;
}
