
/**
 * @brief 网络审计策略
 * 
 */

#include "net_audit.h"
#include "debug_log_queue.h"

#include <linux/slab.h>
#include <linux/in.h>
#include <net/ipv6.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ide.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/mutex.h>
#include <linux/spinlock_types.h>

#define USE_DEBUG_ADD_POLICY (0)


// 查询缓存
struct policy_buf {
    char *policy;
    unsigned int len;
};

// 审计策略
static struct xc_list audit_policy;

// 五元组审计：查表机制 双缓存机制降低中断中的阻塞时间
static struct xc_list audit_policy_tables_1[PROTOCOL_TYPE_SIZE][NETFLOW_TYPE_SIZE][IP_VERSION_SIZE];
static struct xc_list audit_policy_tables_2[PROTOCOL_TYPE_SIZE][NETFLOW_TYPE_SIZE][IP_VERSION_SIZE];

// 五元组审计：操作策略的锁
static struct mutex audit_policy_thread_lock;
static spinlock_t audit_policy_interrupt_lock;

// 五元组审计：缓存区指针
static struct xc_list (*audit_thread_table)[NETFLOW_TYPE_SIZE][IP_VERSION_SIZE] = audit_policy_tables_1;
static struct xc_list (*audit_interrupt_table)[NETFLOW_TYPE_SIZE][IP_VERSION_SIZE] = audit_policy_tables_2;

// 进程审计队列
static struct xc_list audit_proc_policy_1[NETFLOW_TYPE_SIZE], audit_proc_policy_2[NETFLOW_TYPE_SIZE];
static struct xc_list *proc_policy_interrupt = audit_proc_policy_1;
static struct xc_list *proc_policy_thread    = audit_proc_policy_2;
static struct mutex proc_policy_thread_lock;
static spinlock_t proc_policy_interrupt_lock;


static int check_dir(const int direction)
{
    if (NETFLOW_INBOUND != direction
    && NETFLOW_OUTBOUND != direction
    && NETFLOW_ALL != direction) {
        return -1;
    }

    return 0;
}

// 获得策略表ID
static int get_tuple_policy_table_id(const struct netflow_audit_policy *policy, struct policy_table_id *id)
{
    char txt[512] = {0};

    // 进程策略不放在这个表里
    if (NETFLOW_PROCESS == policy->type) {
        debug_log_queue_push_add_title("netflow_drop_add_policy failed! Cannot directly add a process!\n");
        return -1;
    }

    // 检查协议
    uint8_t protocol = policy->tuple.protocol;
    if (IPPROTO_TCP != protocol && IPPROTO_UDP != protocol) {
        snprintf(txt, sizeof (txt), "netflow_audit_add_policy failed! Unknown protocol:%d\n", protocol);
        debug_log_queue_push_add_title(txt);
        return -1;
    }

    // 检查方向
    int direction = (int)policy->direction;
    if (check_dir(direction)) {
        snprintf(txt, sizeof (txt), "netflow_audit_add_policy failed! Unknown direction:%d\n", direction);
        debug_log_queue_push_add_title(txt);
        return -1;
    }

    // 检查IP版本
    int ip_version = (int)policy->tuple.ip_version;
    if (IP_VERSION_IPV4 != ip_version && IP_VERSION_IPV6 != ip_version) {
        snprintf(txt, sizeof (txt), "netflow_audit_add_policy failed! Unknown IP version:%d\n", ip_version);
        debug_log_queue_push_add_title(txt);
        return -1;
    }

    // 检查IP范围
    if (IP_VERSION_IPV4 == ip_version) {
        if (ntohl(policy->tuple.src_ip_range.start.ipv4.s_addr)
        > ntohl(policy->tuple.src_ip_range.end.ipv4.s_addr)) {
            debug_log_queue_push_add_title("netflow_audit_add_policy failed! Wrong IP range!\n");
            return -1;
        }
    }
    else if (IP_VERSION_IPV6 == ip_version) {
        if (ipv6_addr_cmp(&(policy->tuple.src_ip_range.start.ipv6)
        , &(policy->tuple.src_ip_range.end.ipv6)) > 0) {
            debug_log_queue_push_add_title("netflow_audit_add_policy failed! Wrong IP range!\n");
            return -1;
        }
    }

    // 检查端口范围
    if ((policy->tuple.src_port_range.start > policy->tuple.src_port_range.end)
    || (policy->tuple.dst_port_range.start > policy->tuple.dst_port_range.end)) {
        debug_log_queue_push_add_title("netflow_audit_add_policy failed! Wrong port range!\n");
        return -1;
    }

    id->protocol   = (int)((IPPROTO_TCP == protocol) ? (PROTOCOL_TCP) : (PROTOCOL_UDP));
    id->direction  = direction;
    id->ip_version = ip_version;

    return 0;
}

// 策略表遍历
static int policy_table_for_each(struct xc_list (*policy_tables)[NETFLOW_TYPE_SIZE][IP_VERSION_SIZE], xc_list_fun fun, void *arg)
{
    int i,j,k;

    for (i = 0; i < PROTOCOL_TYPE_SIZE; ++i) {
        for (j = 0; j < NETFLOW_TYPE_SIZE; ++j) {
            for (k = 0; k < IP_VERSION_SIZE; ++k) {
                fun(&(policy_tables[i][j][k]), arg);
            }
        }
    }

    return 0;
}

static int init_list_callback(struct xc_list *list, void *arg)
{
    char name[64];
    static int count = 0;

    xc_list_init(list, ((unsigned int)(unsigned long)arg));
    snprintf(name, sizeof (name) - 1, "audit_tb_%d", ++count);
    xc_list_set_name(list, name);
    xc_list_set_thread_safe(list, 0);

    return 0;
}

static int clear_list_callback(struct xc_list *list, void *arg)
{
    int ret = 0;

    (void) arg;
    ret = xc_list_clear(list);

    return ret;
}

int netflow_audit_init(void)
{
    const unsigned int maximum = 0; // 没有大小限制

    // 五元组审计初始化
    mutex_init(&audit_policy_thread_lock);
    spin_lock_init(&audit_policy_interrupt_lock);
    xc_list_init(&audit_policy, maximum);
    xc_list_set_name(&audit_policy, "audit_policy");
    xc_list_set_thread_safe(&audit_policy, 0);
    policy_table_for_each(audit_policy_tables_1, init_list_callback, (void *)(unsigned long)maximum);
    policy_table_for_each(audit_policy_tables_2, init_list_callback, (void *)(unsigned long)maximum);

    // 进程审计初始化
    mutex_init(&proc_policy_thread_lock);
    spin_lock_init(&proc_policy_interrupt_lock);
    for (int i = 0; i < NETFLOW_TYPE_SIZE; ++i) {
        xc_list_init(&audit_proc_policy_1[i], maximum);
        xc_list_set_name(&audit_proc_policy_1[i], "audit_proc_policy_1");
        xc_list_set_thread_safe(&audit_proc_policy_1[i], 0);
        xc_list_init(&audit_proc_policy_2[i], maximum);
        xc_list_set_name(&audit_proc_policy_2[i], "audit_proc_policy_2");
        xc_list_set_thread_safe(&audit_proc_policy_2[i], 0);
    }

    return 0;
}

// 调用前必须调用mutex_lock(&audit_policy_thread_lock)
static void swap_table_ptr(void)
{
    unsigned long flags;
    spin_lock_irqsave(&audit_policy_interrupt_lock, flags);

    struct xc_list (*temp)[NETFLOW_TYPE_SIZE][IP_VERSION_SIZE] = audit_thread_table;
    audit_thread_table = audit_interrupt_table;
    audit_interrupt_table = temp;

    spin_unlock_irqrestore(&audit_policy_interrupt_lock, flags);
}

int netflow_audit_add_policy(struct netflow_audit_policy * policy)
{
    int ret = 0;
    struct policy_table_id table_id = {0};
#if USE_DEBUG_ADD_POLICY
    char txt[128];
#endif

    if (NETFLOW_TUPLE == policy->type) {
        // 获得表id
        ret = get_tuple_policy_table_id(policy, &table_id);
        if (ret) {
            return -1;
        }
    }
    // 进程策略
    else {
        // 检查方向
        int direction = (int)policy->direction;
        if (check_dir(direction)) {
#if USE_DEBUG_ADD_POLICY
            snprintf(txt, sizeof (txt), "netflow_audit_add_policy failed! Unknown direction:%d\n", direction);
            debug_log_queue_push_add_title(txt);
#endif
            return -1;
        }
    }
    
#if USE_DEBUG_ADD_POLICY
    snprintf(txt, sizeof (txt) - 1, "table_id: protocol:%d direction:%d ip_version:%d\n"
    , table_id.protocol, table_id.direction, table_id.ip_version);
    debug_log_queue_push_add_title(txt);
#endif

    mutex_lock(&audit_policy_thread_lock);

    // 保存策略
    ret = xc_list_push(&audit_policy, policy, sizeof (struct netflow_audit_policy));
    if (ret) {
        mutex_unlock(&audit_policy_thread_lock);
        debug_log_queue_push_add_title("netflow_audit_add_policy failed!");
        return -1;
    }

    // 进程策略到这里就结束了
    if (NETFLOW_TUPLE != policy->type) {
        mutex_unlock(&audit_policy_thread_lock);
        return 0;
    }

    // 保存到表1
    struct xc_list *table = &audit_thread_table[table_id.protocol][table_id.direction][table_id.ip_version];
    ret = xc_list_push(table, policy, sizeof (struct netflow_audit_policy));
    if (ret) {
        xc_list_remove_back(&audit_policy);
        mutex_unlock(&audit_policy_thread_lock);
        debug_log_queue_push_add_title("netflow_audit_add_policy failed!");
        return -1;
    }

    // 交换指针
    swap_table_ptr();

    // 保存到表2
    table = &audit_thread_table[table_id.protocol][table_id.direction][table_id.ip_version];
    ret = xc_list_push(table, policy, sizeof (struct netflow_audit_policy));
    if (ret) {
        xc_list_remove_back(&audit_policy);

        unsigned long flags;
        spin_lock_irqsave(&audit_policy_interrupt_lock, flags);
        xc_list_remove_back(&audit_interrupt_table[table_id.protocol][table_id.direction][table_id.ip_version]);
        spin_unlock_irqrestore(&audit_policy_interrupt_lock, flags);

        mutex_unlock(&audit_policy_thread_lock);
        debug_log_queue_push_add_title("netflow_audit_add_policy failed!");
        return -1;
    }

    mutex_unlock(&audit_policy_thread_lock);

    return 0;
}

/**
 * @brief 添加进程审计策略
 * 
 * @param policy 
 * @return 0：成功 -1：失败
 */
int netflow_audit_add_proc_policy(struct netflow_audit_proc_policy *policy)
{
#if USE_DEBUG_ADD_POLICY
    char txt[128];
#endif
    // 检查方向
    int direction = (int)policy->direction;
    if (check_dir(direction)) {
#if USE_DEBUG_ADD_POLICY
        snprintf(txt, sizeof (txt), "netflow_audit_add_proc_policy failed! Unknown direction:%d\n", direction);
        debug_log_queue_push_add_title(txt);
#endif
        return -1;
    }

    mutex_lock(&proc_policy_thread_lock);

    int ret = xc_list_push(&(proc_policy_thread[direction]), policy
    , sizeof (struct netflow_audit_proc_policy));

    mutex_unlock(&proc_policy_thread_lock);

    return ret;
}

/**
 * @brief 添加进程审计策略完成
 * 
 * @return 0：成功 -1：失败
 */
int netflow_audit_add_policy_complete(void)
{
    mutex_lock(&proc_policy_thread_lock);
    unsigned long flags;
    spin_lock_irqsave(&proc_policy_interrupt_lock, flags);

    // 交换指针
    struct xc_list *tmp = proc_policy_interrupt;
    proc_policy_interrupt = proc_policy_thread;
    proc_policy_thread = tmp;

    spin_unlock_irqrestore(&proc_policy_interrupt_lock, flags);

    // 为下次接收做准备
    for (int i = 0; i < NETFLOW_TYPE_SIZE; ++i) {
        xc_list_clear(&(proc_policy_thread[i]));
    }

    mutex_unlock(&proc_policy_thread_lock);

    return 0;
}

static int find_policy_id_of_table(const char *data, const char *condition)
{
    const struct netflow_audit_policy *policy = (const struct netflow_audit_policy *)data;
    uint16_t policy_id = (uint16_t)(unsigned long)condition;

    if (policy->policy_id == policy_id) {
        return 0;
    }

    return -1;
}

static int find_policy_id(const char *data, const char *condition)
{
    const struct netflow_audit_policy *policy = (const struct netflow_audit_policy *)data;
    uint16_t policy_id = (uint16_t)(unsigned long)condition;

    if (policy->policy_id == policy_id) {
        // 进程策略到这里就结束了
        if (NETFLOW_TUPLE != policy->type) {
            return 0;
        }

        int ret = 0;
        struct policy_table_id table_id = {0};

        // 获得表id
        ret = get_tuple_policy_table_id(policy, &table_id);
        if (ret) {
            // 不应该出现这种情况
            debug_log_queue_push_add_title("Error: abnormal table ID!\n");
            return 0;
        }

        // 删除表1中的项
        struct xc_list *table = &audit_thread_table[table_id.protocol][table_id.direction][table_id.ip_version];
        ret = xc_list_remove_if(table, find_policy_id_of_table, (const char *)(unsigned long)policy_id);
        if (ret) {
            debug_log_queue_push_add_title("Warning: node does not exist!");
        }

        // 交换指针
        swap_table_ptr();

        // 删除表2中的项
        table = &audit_thread_table[table_id.protocol][table_id.direction][table_id.ip_version];
        ret = xc_list_remove_if(table, find_policy_id_of_table, (const char *)(unsigned long)policy_id);
        if (ret) {
            debug_log_queue_push_add_title("Warning: node does not exist!");
        }

        return 0;
    }

    return -1;
}

int netflow_audit_delete_policy(uint16_t policy_id)
{
    int ret = 0;

    mutex_lock(&audit_policy_thread_lock);

    ret = xc_list_remove_if(&audit_policy, find_policy_id, (const char *)(unsigned long)policy_id);
    if (ret) {
        debug_log_queue_push_add_title("Warning: node does not exist!");
    }

    mutex_unlock(&audit_policy_thread_lock);

    return ret;
}

int netflow_audit_clear_policy(void)
{
    mutex_lock(&audit_policy_thread_lock);

    // 清空策略表
    xc_list_clear(&audit_policy);
    // 清空表1
    policy_table_for_each(audit_thread_table, clear_list_callback, NULL);
    // 交换指针
    swap_table_ptr();
    // 清空表2
    policy_table_for_each(audit_thread_table, clear_list_callback, NULL);

    mutex_unlock(&audit_policy_thread_lock);

    return 0;
}

static int get_audit_policies(const char *data, char *opt)
{
    struct xc_list *policy_query = (struct xc_list *)opt;

    xc_list_push(policy_query, data, sizeof (struct netflow_audit_policy));
    
    return -1;
}

static int fill_policy_buf_query(const char *data, char *opt)
{
    struct policy_buf *buf = (struct policy_buf *)opt;

    memcpy(buf->policy + buf->len * sizeof(struct netflow_audit_policy)
    , data, sizeof(struct netflow_audit_policy));
    ++buf->len;

    return -1;
}

struct netflow_audit_policy *netflow_audit_policies(int *count)
{
    unsigned int len = 0;
    int ret = 0;
    struct netflow_audit_policy *policy = NULL;
    struct xc_list policy_query; // 查询得到的审计策略
    struct policy_buf policy_buf_query = {NULL, 0}; // 查询策略buf

    xc_list_init(&policy_query, 0); // 链表长度大小不限制
    xc_list_set_name(&policy_query, "policy_query");
    xc_list_set_thread_safe(&policy_query, 0);
    mutex_lock(&audit_policy_thread_lock);

    ret = xc_list_for_each(&audit_policy, get_audit_policies, (char *)(&policy_query), &len);
    if (!ret && len) {
        *count = len;
        policy = kmalloc(len * sizeof(struct netflow_audit_policy), GFP_ATOMIC);
        if(!policy) {
            mutex_unlock(&audit_policy_thread_lock);
            printk(KERN_WARNING "[net_audit_module] Failed to allocate memory.\n");
            *count = -1;
            return NULL;
        }

        policy_buf_query.len = 0;
        policy_buf_query.policy = (char *)policy;
        ret = xc_list_for_each(&policy_query
                                , fill_policy_buf_query
                                , (char *)(&policy_buf_query), &len);
        xc_list_destroy(&policy_query);
        if (!ret) {
            mutex_unlock(&audit_policy_thread_lock);
            return policy;
        }
        else {
            mutex_unlock(&audit_policy_thread_lock);
            *count = 0;
            kfree(policy);
            return NULL;
        }
    }
    else {
        mutex_unlock(&audit_policy_thread_lock);
        xc_list_destroy(&policy_query);
        *count = 0;
        return NULL;
    }
}

int netflow_audit_policies_tables_for_each(struct policy_table_id *table_id
                                            , xc_list_for_each_fun fun
                                            , void *arg)
{
    if (NETFLOW_TYPE_SIZE <= table_id->direction
    || IP_VERSION_SIZE <= table_id->ip_version
    || PROTOCOL_TYPE_SIZE <= table_id->protocol) {
        return -1;
    }

    unsigned long flags;
    spin_lock_irqsave(&audit_policy_interrupt_lock, flags);

    struct xc_list *table = &audit_interrupt_table[table_id->protocol][table_id->direction][table_id->ip_version];
    int len = 0;
    int ret = xc_list_for_each(table, fun, arg, &len);

    spin_unlock_irqrestore(&audit_policy_interrupt_lock, flags);

    return ret;
}

int netflow_audit_proc_policies_tables_for_each(int direction
                                                , xc_list_for_each_fun fun
                                                , void *arg)
{
    // 检查方向
    if (check_dir(direction)) {
        return -1;
    }

    unsigned long flags;
    spin_lock_irqsave(&proc_policy_interrupt_lock, flags);

    int len = 0;
    int ret = xc_list_for_each(&(proc_policy_interrupt[direction]), fun, arg, &len);

    spin_unlock_irqrestore(&proc_policy_interrupt_lock, flags);

    return ret;
}
