#include <project.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/usb.h>

#include "usb_ctrl.h"
#include "usb_wlist.h"
#include "usb_sw.h"
#include "common/audit/sec_audit.h"

struct proc_dir_entry *nfssec_devctl_dir;
struct proc_dir_entry *devctrl_policy_file;
struct proc_dir_entry *devctl_switch_file;
struct proc_dir_entry *devctl_audit_file;

audit_q_t g_audit_pac;

//遍历所有usb设备的interface找到并释放指定的data的interface驱动
static int foreach_udev_release_driver(struct usb_device *udev, void *data)
{
    struct usb_interface *intf;
    struct usb_attr *attr;
    int i;

    attr = (struct usb_attr *)data;
    if (attr->vid != udev->descriptor.idVendor ||
        attr->pid != udev->descriptor.idProduct)
        return 0;

    for (i = 0; i < udev->actconfig->desc.bNumInterfaces; i++) {
        intf = udev->actconfig->interface[i];
        if (intf && attr->clsid == intf->cur_altsetting->desc.bInterfaceClass && intf->dev.driver) {
            tag_pr_debug("[usb_ctrl] release vid=%04x pid=%04x clsid=%04x\n",
                   attr->vid, attr->pid, attr->clsid);
            usb_driver_release_interface(to_usb_driver(intf->dev.driver), intf);
        }
    }

    return 0;
}

//释放attr对应的interface驱动
static int find_usb_release_driver(struct usb_attr *attr)
{
    usb_for_each_dev(attr, foreach_udev_release_driver);

    return 0;
}

//当有usb设备插入时，触发此函数，检测插入的usb设备的存储类中不在白名单的设备驱动卸载，只放行白名单设备
static int notify_dev_add_func(struct usb_device *udev)
{
    int i;
    struct usb_interface *intf;
    struct usb_attr attr = {0};
    int find = 0;
    struct dev_switch sw = {0};

	sw.type = DEV_TYPE_USB;
	sw.sub_type = USB_CLASS_MASS_STORAGE;

	type_switch_get(&sw);
	if (sw.mode == SW_MODE_OFF)
		return 0;

    attr.vid = udev->descriptor.idVendor;
    attr.pid = udev->descriptor.idProduct;

    for (i = 0; i < udev->actconfig->desc.bNumInterfaces; i++) {

        intf = udev->actconfig->interface[i];
		if (!intf)
			continue;

		attr.clsid = intf->cur_altsetting->desc.bInterfaceClass;

		switch (attr.clsid) {
			case USB_CLASS_MASS_STORAGE:
				find = usb_wlist_find(&attr);
				if (find) {
					tag_pr_debug("[usb_ctrl] got vid=%04x pid=%04x clsid=%04x in white list: pass\n",
							attr.vid, attr.pid, attr.clsid);
					break;
				}
				
				// not found in whitelist
				tag_pr_debug("[usb_ctrl] got vid=%04x pid=%04x clsid=%04x not found in white list: stop\n",
						attr.vid, attr.pid, attr.clsid);

				if (sw.mode == SW_MODE_CTRL || sw.mode == SW_MODE_AUDIT_CTRL) {
					tag_pr_debug("usb do ctrl\n");
					if (intf->dev.driver)
						usb_driver_release_interface(to_usb_driver(intf->dev.driver), intf);
				}

				if (sw.mode == SW_MODE_AUDIT || sw.mode == SW_MODE_AUDIT_CTRL) {
					tag_pr_debug("usb do audit\n");
                    audit_write_pac(&g_audit_pac, PAC_DEV_TYPE_USB, attr.pid, attr.vid, attr.clsid, PAC_ACTION_DEV_GOT);
				}
				break;

			default:
				break;
		}
    }

    return 0;
}

//设置usb notify
static int usb_notify(struct notifier_block *self, unsigned long action,
                      void *dev)
{
    switch (action) {
        case USB_DEVICE_ADD:
            notify_dev_add_func(dev);
            break;
        case USB_DEVICE_REMOVE:
            break;
    }
    return NOTIFY_OK;
}

static struct notifier_block usb_notify_block = {
    .notifier_call = usb_notify,
};

// data中struct all_usb_data为数据头，将查询结果放在数据体部分
static int all_usb_foreach(struct usb_device *udev, void *data)
{
    struct usb_interface *intf;
    struct usb_attr *attr_item;
    struct all_usb_data *all_usb;
    int i;

    if (!data)
        return 0;

    all_usb = (struct all_usb_data *)data;
    if (all_usb->index >= all_usb->data_len)
        return -ENOMEM;

    for (i = 0; i < udev->actconfig->desc.bNumInterfaces; i++) {
        intf = udev->actconfig->interface[i];
        if (intf) {
            if (all_usb->index + sizeof(struct usb_attr) > all_usb->data_len)
                return -ENOMEM;

            if (intf->cur_altsetting->desc.bInterfaceClass ==
                USB_CLASS_MASS_STORAGE) {
                tag_pr_debug("[usb_ctrl] gto USB_CLASS_MASS_STORAGE: vid=%04x pid=%04x \n",
                       udev->descriptor.idVendor, udev->descriptor.idProduct);
                attr_item = (struct usb_attr *)(all_usb->data + all_usb->index);
                attr_item->vid = udev->descriptor.idVendor;
                attr_item->pid = udev->descriptor.idProduct;
                attr_item->clsid = intf->cur_altsetting->desc.bInterfaceClass;
                all_usb->index += sizeof(struct usb_attr);
            }
        }
    }

    return 0;
}

//读取所有usb数据和usb白名单数据
static ssize_t devctl_ply_read(struct file *file, char __user *ubuf, size_t size,
                              loff_t *off)
{
    int retval = 0;
    char *buf;
    struct all_usb_data *all_usb;
    // int i;

    buf = kzalloc(PAGE_SIZE, GFP_KERNEL);
    if (!buf)
        return -ENOMEM;

    all_usb = (struct all_usb_data *)buf;
    all_usb->data_len = PAGE_SIZE - sizeof(struct all_usb_data);
    all_usb->index = 0;

    usb_for_each_dev(all_usb, all_usb_foreach);

    // tag_pr_debug("all_usb->index[%d]\n", all_usb->index);
    // for (i = 0; i < all_usb->index; i++)
    //     tag_pr_debug("all_usb->data%d[%x]\n", i, all_usb->data[i]);

    if (0 != copy_to_user(ubuf, all_usb->data, all_usb->index)) {
        retval = -EFAULT;
        goto DONE;
    }

    retval = all_usb->index;
DONE:
    if (buf)
        kfree(buf);
    return retval;
}

//从procfs下发usb配置信息
static ssize_t devctl_ply_write(struct file *file, const char __user *ubuf,
                               size_t count, loff_t *off)
{
    char *tmp;
    struct usb_attr_set *set;
    int err = 0, ret = 0;
    struct dev_switch sw = {0};

    tmp = kzalloc(count + 1, GFP_KERNEL);
    if (!tmp)
        return -ENOMEM;

    if (0 != copy_from_user(tmp, ubuf, count)) {
        kfree(tmp);
        return -EFAULT;
    }

    // notice： 目前只考虑mass
	sw.type = DEV_TYPE_USB;
	sw.sub_type = USB_CLASS_MASS_STORAGE;
	type_switch_get(&sw);

    set = (struct usb_attr_set *)(tmp);
    switch (set->state) {
        case USB_STATE_ON:
            err = usb_wlist_add(&set->attr);
            if (err == 0 && sw.mode != SW_MODE_OFF) {
                // todo: make usb on
            }
            break;
        case USB_STATE_OFF:
            err = usb_wlist_del(&set->attr);
            tag_pr_debug("err[%d] sw.mode[%d]\n", err, sw.mode);

            if (err == 0 && sw.mode != SW_MODE_OFF) {
                // todo:因为可能多次关，需要检查后再关（当前状态或关时检查）
                find_usb_release_driver(&set->attr);  
            }
            break;
        default:
            ret = -EFAULT;
            break;
    }

    if (err == 0)
        ret = count;

    return ret;
}

const struct file_operations devctl_ply_ops = {
    .read = devctl_ply_read,
    .write = devctl_ply_write,
};

//下发开关配置
static ssize_t devctl_switch_write(struct file *file, const char __user *ubuf,
                            size_t count, loff_t *off)
{
	int retval = 0;
    int ret = 0;
    char *tmp;
    struct dev_switch sw_old = {0};
    struct dev_switch * sw_new = NULL;

    if (count < sizeof(struct dev_switch))
        return -EFAULT;

    tmp = kzalloc(count + 1, GFP_KERNEL);
    if (!tmp)
        return -ENOMEM;

    if (0 != copy_from_user(tmp, ubuf, count)) {
        retval = -EFAULT;
		goto DONE;
    }

    sw_new = (struct dev_switch *)tmp;

    ret = type_switch_set(sw_new);
    if (ret) {
		retval = ret;
		goto DONE;
    }

    // 当开关状态由 关->开，应该重新进行检测
	type_switch_get(&sw_old);
	if (sw_old.mode == SW_MODE_OFF && sw_new->mode != SW_MODE_OFF) {
        // 遍历当前插入的设备列表，对每一项从wlist中查找，如果找到，则加载驱动，如果未找到，则卸载驱动        
        int ret = 0;
        char *buf;
        struct all_usb_data *all_usb;
        unsigned int usb_cnt = 0;
        int i;
        struct usb_attr *attr = NULL;

        buf = kzalloc(PAGE_SIZE, GFP_KERNEL);
        if (!buf) 
            return -ENOMEM;

        all_usb = (struct all_usb_data *)buf;
        all_usb->data_len = PAGE_SIZE - sizeof(struct all_usb_data);
        all_usb->index = 0;

        usb_for_each_dev(all_usb, all_usb_foreach);
        
        usb_cnt = all_usb->index / sizeof (struct usb_attr);

        for (i = 0; i < usb_cnt; i++) {

            attr = &(((struct usb_attr*)all_usb->data)[i]);
            ret = usb_wlist_find(attr);
            if (ret) {
                // todo: make usb on
                tag_pr_debug("make usb on\n");
            } else {
                if (sw_new->mode == SW_MODE_CTRL || sw_new->mode == SW_MODE_AUDIT_CTRL) {
					tag_pr_debug("usb do ctrl\n");
					find_usb_release_driver(attr); 
				}

				if (sw_new->mode == SW_MODE_AUDIT || sw_new->mode == SW_MODE_AUDIT_CTRL) {
					tag_pr_debug("usb do audit\n");
                    audit_write_pac(&g_audit_pac, PAC_DEV_TYPE_USB, attr->pid, attr->vid, attr->clsid, PAC_ACTION_DEV_GOT);
				}
            }
        }
    }

	retval = count;
DONE:
	if (tmp)
    	kfree(tmp);
    return retval;
}

const struct file_operations devctl_switch_ops = {
    .write = devctl_switch_write,
};

static ssize_t devctl_audit_read(struct file *file, char __user *ubuf, size_t size,
    loff_t *off)
{
   return audit_q_read_buf(&g_audit_pac, ubuf);
}

const struct file_operations devctl_audit_ops = {
    .read = devctl_audit_read,
};

static int usb_ctrl_proc_create(void)
{
    nfssec_devctl_dir = proc_mkdir(NFSSEC_DEVCTL_DIR, NULL);
    if (!nfssec_devctl_dir) {
        tag_pr_err("create /proc/%s error!\n", NFSSEC_DEVCTL_DIR);
        return EFAULT;
    }

    devctrl_policy_file =
        proc_create(NFSSEC_USB_POLICY, 0666, nfssec_devctl_dir, &devctl_ply_ops);
    if (!devctrl_policy_file) {
        tag_pr_err("can not create /proc/%s\n", NFSSEC_USB_POLICY);
        return -EFAULT;
    }

    devctl_switch_file = proc_create(NFSSEC_DEVCTL_SWITCH, 0666, nfssec_devctl_dir,
                                  &devctl_switch_ops);
    if (!devctl_switch_file) {
        tag_pr_err("can not create /proc/%s\n", NFSSEC_DEVCTL_SWITCH);
        return -EFAULT;
    }

    devctl_audit_file = proc_create(NFSSEC_DEVCTL_AUDIT, 0666, nfssec_devctl_dir,
                                  &devctl_audit_ops);
    if (!devctl_audit_file) {
        tag_pr_err("can not create /proc/%s\n", NFSSEC_DEVCTL_AUDIT);
        return -EFAULT;
    }

    return 0;
}

static void usb_ctrl_proc_delete(void)
{
    if (devctl_switch_file)
        remove_proc_entry(NFSSEC_DEVCTL_SWITCH, nfssec_devctl_dir);

    if (devctrl_policy_file)
        remove_proc_entry(NFSSEC_USB_POLICY, nfssec_devctl_dir);

    if (devctl_audit_file)
        remove_proc_entry(NFSSEC_DEVCTL_AUDIT, nfssec_devctl_dir);

    if (nfssec_devctl_dir)
        remove_proc_entry(NFSSEC_DEVCTL_DIR, NULL);
}

static int __init __dev_ctrl_init(void)
{
     int ret;

    ret = usb_ctrl_proc_create();
    if (ret)
        return ret;

    usb_register_notify(&usb_notify_block);

	audit_q_init(&g_audit_pac, 200);

    return 0;
}

static void __exit dev_ctrl_exit(void)
{
    usb_ctrl_proc_delete();

    usb_unregister_notify(&usb_notify_block);

    usb_wlist_free();

    audit_q_clear(&g_audit_pac);
}

#ifdef NFSSEC_KMOD
module_init(dev_ctrl_init);
module_exit(dev_ctrl_exit);
MODULE_AUTHOR("nfssec");
MODULE_LICENSE("GPL");
#else

static struct delayed_work __delay_work;
static struct workqueue_struct *wq;
#define MY_FILE "/opt/nfssec/script/loadService.sh"

static void __nfssec_pac_init_delay(struct work_struct *work_req)
{
	struct file *filep = NULL;

	filep = filp_open(MY_FILE, O_RDWR | O_APPEND | O_CREAT, 0644);
	if (IS_ERR_OR_NULL(filep)) {
		mod_delayed_work(wq, &__delay_work, 100);
	} else {
		filp_close(filep, NULL);
		__dev_ctrl_init();
	}
}

int __init nfssec_pac_init(void)
{
	INIT_DELAYED_WORK(&__delay_work, __nfssec_pac_init_delay);
	wq = alloc_ordered_workqueue("nfssec_wq_pac", WQ_MEM_RECLAIM);
	queue_delayed_work(wq, &__delay_work, 100);

	return 0;
}
#endif