#include <project.h>
#include "common/version_type.h"
#include "common/sym_hook/lsm_if.h"
#include "lsm_impl.h"
#include "nfs_security.h"
#include "nfs_security_ops.h"
#include "common/nfs_mnt.h"

static int impl_bprm_check_security(struct linux_binprm *bprm)
{
    return NFS_SECURITY_OPS(bprm_check, bprm);
}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 32)
static int impl_file_open(struct dentry *dentry, struct vfsmount *mnt, const int flags)
#else
static int impl_file_open(struct file *file)
#endif
{
    int retval;

    struct dentry * __dentry = NULL;
    struct vfsmount * __mnt = NULL;
    int __flags = 0;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 32)
    __dentry = dentry;
    __mnt = mnt;
    __flags = flags;
#else
// #ifndef CONFIG_SECURITY_PATH
//     mnt = get_dentry_mnt(dentry);
// #endif

    __mnt = get_file_vfsmount(file);
    __dentry = get_file_dentry(file);
    __flags = file->f_flags;
#endif
    retval = NFS_SECURITY_OPS(open_check, __dentry, __mnt, __flags);

    // if (mnt)
    //     mntput(mnt);
    return retval;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)
static int impl_task_kill(struct task_struct *p, struct kernel_siginfo *info,
                         int sig, const struct cred *secid)
#else
static int impl_task_kill(struct task_struct *p, struct siginfo *info, int sig,
                         const struct cred *cred)
#endif
{
    int retval = 0;
    pid_t pid = p->pid;

    retval = NFS_SECURITY_OPS(kill_check, pid, sig);

    return retval;
}

static int impl_inode_create(struct inode *dir, struct dentry *dentry, umode_t mode)
{
    int retval;
    struct vfsmount *mnt = NULL;

#ifndef CONFIG_SECURITY_PATH
    mnt = get_dentry_mnt(dentry);
#endif

    retval = NFS_SECURITY_OPS(mknod_check, dentry, mnt, mode, 0);

    if (mnt)
        mntput(mnt);
    return retval;
}

#ifdef CONFIG_SECURITY_PATH
static int impl_path_mknod(const struct path *dir, struct dentry *dentry,
                          umode_t mode, unsigned int dev)
{
    return NFS_SECURITY_OPS(mknod_check, dentry, dir->mnt, mode, dev);
}

static int impl_path_mkdir(const struct path *dir, struct dentry *dentry,
                          umode_t mode)
{
    return NFS_SECURITY_OPS(mkdir_check, dentry, dir->mnt, mode);
}

static int impl_path_rmdir(const struct path *dir, struct dentry *dentry)
{
    return NFS_SECURITY_OPS(rmdir_check, dentry, dir->mnt);
}

static int impl_path_unlink(const struct path *dir, struct dentry *dentry)
{
    return NFS_SECURITY_OPS(unlink_check, dentry, dir->mnt);
}

static int impl_path_symlink(const struct path *dir, struct dentry *dentry,
                            const char *old_name)
{
    return NFS_SECURITY_OPS(symlink_check, dentry, dir->mnt, old_name);
}

static int impl_path_rename(const struct path *old_dir, struct dentry *old_dentry,
                    const struct path *new_dir, struct dentry *new_dentry)
{
    return NFS_SECURITY_OPS(rename_check, old_dentry, new_dentry, old_dir->mnt);
}

static int impl_path_link(struct dentry *old_dentry, const struct path *new_dir,
                         struct dentry *new_dentry)
{
    return NFS_SECURITY_OPS(link_check, old_dentry, new_dentry, new_dir->mnt);
}

static int impl_path_truncate(const struct path *path)
{
    return NFS_SECURITY_OPS(truncate_check, path->dentry, path->mnt);
}

static int impl_path_chown(const struct path *path, kuid_t user, kgid_t group)
{
    return NFS_SECURITY_OPS(chown_check, path->dentry, path->mnt, user, group);
}

// int impl_path_chmod(const struct path *path, umode_t mode)
// {
//     return NFS_SECURITY_OPS(chmod_check, path->dentry, path->mnt, mode);
// }

// int impl_path_chroot(const struct path *path)
// {
//     return NFS_SECURITY_OPS(chroot_check, path);
// }

#else

static int impl_inode_mknod(struct inode *dir, struct dentry *dentry,
                           umode_t mode, dev_t dev)
{
    int retval;
    struct vfsmount *mnt = NULL;

#ifndef CONFIG_SECURITY_PATH
    mnt = get_dentry_mnt(dentry);
#endif

    retval = NFS_SECURITY_OPS(mknod_check, dentry, mnt, mode, dev);

    if (mnt)
        mntput(mnt);
    return retval;
}

static int impl_inode_rename(struct inode *old_dir, struct dentry *old_dentry,
                            struct inode *new_dir, struct dentry *new_dentry)
{
    int retval = 0;
    struct vfsmount *mnt = NULL;

#ifndef CONFIG_SECURITY_PATH
    mnt = get_dentry_mnt(old_dentry);
#endif

    retval = NFS_SECURITY_OPS(rename_check, old_dentry, new_dentry, mnt);

    if (mnt)
        mntput(mnt);
    return retval;
}

static int impl_inode_mkdir(struct inode *dir, struct dentry *dentry,
                           umode_t mode)
{
    int retval;
    struct vfsmount *mnt = NULL;

#ifndef CONFIG_SECURITY_PATH
    mnt = get_dentry_mnt(dentry);
#endif

    retval = NFS_SECURITY_OPS(mkdir_check, dentry, mnt, mode);

    if (mnt)
        mntput(mnt);
    return retval;
}

static int impl_inode_rmdir(struct inode *dir, struct dentry *dentry)
{
    int retval = 0;
    struct vfsmount *mnt = NULL;

#ifndef CONFIG_SECURITY_PATH
    mnt = get_dentry_mnt(dentry);
#endif

    retval = NFS_SECURITY_OPS(rmdir_check, dentry, mnt);

    if (mnt)
        mntput(mnt);
    return retval;
}

static int impl_inode_symlink(struct inode *dir, struct dentry *dentry,
                             const char *old_name)
{
    int retval;
    struct vfsmount *mnt = NULL;

#ifndef CONFIG_SECURITY_PATH
    mnt = get_dentry_mnt(dentry);
#endif

    retval = NFS_SECURITY_OPS(symlink_check, dentry, mnt, old_name);

    if (mnt)
        mntput(mnt);
    return retval;
}

static int impl_inode_link(struct dentry *old_dentry, struct inode *dir,
                          struct dentry *new_dentry)
{
    int retval;
    struct vfsmount *mnt = NULL;

#ifndef CONFIG_SECURITY_PATH
    mnt = get_dentry_mnt(new_dentry);
#endif

    retval = NFS_SECURITY_OPS(link_check, old_dentry, new_dentry, mnt);

    if (mnt)
        mntput(mnt);
    return retval;
}                      

static int impl_inode_unlink(struct inode *dir, struct dentry *dentry)
{
    int retval;
    struct vfsmount *mnt = NULL;

#ifndef CONFIG_SECURITY_PATH
    mnt = get_dentry_mnt(dentry);
#endif

    retval = NFS_SECURITY_OPS(unlink_check, dentry, mnt);

    if (mnt)
        mntput(mnt);
    return retval;
}
#endif

struct security_hook_list nfssec_prot_hooks[] = {

   	MY_HOOK_INIT(bprm_check_security, impl_bprm_check_security),

    MY_HOOK_INIT(task_kill, impl_task_kill),

    MY_HOOK_INIT(file_open, impl_file_open), 

    // for creat syscall(normal file)
    MY_HOOK_INIT(inode_create, impl_inode_create),

#ifdef CONFIG_SECURITY_PATH
    MY_HOOK_INIT(path_mknod, impl_path_mknod),
    MY_HOOK_INIT(path_mkdir, impl_path_mkdir),
    MY_HOOK_INIT(path_rmdir, impl_path_rmdir),
    MY_HOOK_INIT(path_unlink, impl_path_unlink),
    MY_HOOK_INIT(path_symlink, impl_path_symlink),
    MY_HOOK_INIT(path_rename, impl_path_rename),
    MY_HOOK_INIT(path_link, impl_path_link),
    MY_HOOK_INIT(path_truncate, impl_path_truncate),
    MY_HOOK_INIT(path_chown, impl_path_chown),
    // MY_HOOK_INIT(path_chmod, impl_path_chmod),
    // MY_HOOK_INIT(path_chroot, impl_path_chroot),
#else
    // for mknod syscall(dev file)
    MY_HOOK_INIT(inode_mknod, impl_inode_mknod),
    MY_HOOK_INIT(inode_rename, impl_inode_rename),
    MY_HOOK_INIT(inode_mkdir, impl_inode_mkdir),
    MY_HOOK_INIT(inode_rmdir, impl_inode_rmdir),
    MY_HOOK_INIT(inode_symlink, impl_inode_symlink),
    MY_HOOK_INIT(inode_link, impl_inode_link),
    MY_HOOK_INIT(inode_unlink, impl_inode_unlink),
#endif
};

int nfssec_prot_hook_size = sizeof(nfssec_prot_hooks) / sizeof(nfssec_prot_hooks[0]);
