#include <project.h>
#include <linux/fs.h>
#include <linux/mount.h>

#include "nfs_security.h"
#include "protect_check.h"

int nfs_bprm_check(struct linux_binprm *bprm)
{
    return bprm_check(bprm);
}

int nfs_kill_check(pid_t pid, int sig)
{
    int retval = 0;

    if (9 != sig && 15 != sig) 
        goto DONE;

    if (0 != task_kill_check(pid, sig)) {
        retval = -EACCES;
        goto DONE;        
    }

DONE:
    return retval;
}

int nfs_open_check(struct dentry *dentry, struct vfsmount *mnt, const int flags)
{
    int retval = 0;

    if (!dentry || !mnt)
        goto DONE;

    // 只处理追加操作（写操作可以直接通过截断钩子）
    if (!((O_APPEND | O_TRUNC | O_CREAT) & flags)) 
        goto DONE;

    if (0 != file_acl_check(dentry, mnt, TRUE, FALSE, __SUBTYPE_WRITEFILE)) {
        retval = -EACCES;
        goto DONE;
    }

DONE:
    return retval;
}

int nfs_mknod_check(struct dentry *dentry, struct vfsmount *mnt,
                         unsigned int mode, unsigned int dev)
{
    int retval = 0;

    if (!dentry)
        goto DONE;

    // todo : 可以细化为创建文件，创建设备文件
    if (0 != file_acl_check(dentry, mnt, TRUE, FALSE, __SUBTYPE_CREATEFILE)) {
        retval = -EACCES;
        goto DONE;
    }

DONE:
    return retval;
}

int nfs_rename_check(struct dentry *old_dentry, struct dentry *new_dentry, 
    struct vfsmount *mnt)
{
    int retval = 0;

    if (!mnt || !old_dentry || !old_dentry->d_inode)
        goto DONE;

    if ((!S_ISREG(old_dentry->d_inode->i_mode)) &&
        (!S_ISDIR(old_dentry->d_inode->i_mode)) &&
        (!S_ISLNK(old_dentry->d_inode->i_mode)))
        goto DONE;

    if (!new_dentry)
        goto DONE;

    if (0 != file_acl_check(old_dentry, mnt, TRUE, FALSE, __SUBTYPE_RENAMEFILE)) {
        retval = -EACCES;
        goto DONE;
    }

    if (0 != file_acl_check(new_dentry, mnt, TRUE, FALSE, __SUBTYPE_RENAMEFILE)) {
        retval = -EACCES;
        goto DONE;
    }

DONE:
    return retval;
}

int nfs_mkdir_check(struct dentry *dentry, struct vfsmount *mnt,
                         unsigned int mode)
{
    int retval = 0;
        
    if (!mnt || !dentry)
        goto DONE;

    if (0 != file_acl_check(dentry, mnt, TRUE, FALSE, __SUBTYPE_CREATEFOLDER)) {
        retval = -EACCES;
        goto DONE;
    }

DONE:
    return retval;
}


int nfs_rmdir_check(struct dentry *dentry, struct vfsmount *mnt)
{
    int retval = 0;

    if (!mnt || !dentry || !dentry->d_inode)
        goto DONE;

    if ((!S_ISREG(dentry->d_inode->i_mode)) &&
        (!S_ISDIR(dentry->d_inode->i_mode)))
        goto DONE;
        
    if (0 != file_acl_check(dentry, mnt, TRUE, FALSE, __SUBTYPE_DELETEFOLDER)) {
        retval = -EACCES;
        goto DONE;
    }

DONE:
    return retval;
}

int nfs_symlink_check(struct dentry *dentry, struct vfsmount *mnt,
                           const char *from)
{
    int retval = 0;

    if (!mnt || !dentry)
        goto DONE;

    if (0 != file_acl_check(dentry, mnt, TRUE, FALSE, __SUBTYPE_CREATEFILE)) {
        retval = -EACCES;
        goto DONE;
    }

DONE:
    return retval;
}

int nfs_link_check(struct dentry *old_dentry, struct dentry *new_dentry,
                        struct vfsmount *mnt)
{
    int retval = 0;

    if (!mnt || !old_dentry || !old_dentry->d_inode)
        goto DONE;

    if ((!S_ISREG(old_dentry->d_inode->i_mode)) &&
        (!S_ISDIR(old_dentry->d_inode->i_mode)))
        goto DONE;

    if (!new_dentry)
        goto DONE;

    if (0 != file_acl_check(old_dentry, mnt, TRUE, FALSE, __SUBTYPE_RENAMEFILE)) {
        retval = -EACCES;
        goto DONE;
    }

    if (0 != file_acl_check(new_dentry, mnt, TRUE, FALSE, __SUBTYPE_RENAMEFILE)) {
        retval = -EACCES;
        goto DONE;
    }

DONE:
    return retval;
}

int nfs_unlink_check(struct dentry *dentry, struct vfsmount *mnt)
{
    int retval = 0;

    if (!mnt || !dentry || !dentry->d_inode)
        goto DONE;

    if ((!S_ISREG(dentry->d_inode->i_mode)) &&
        (!S_ISDIR(dentry->d_inode->i_mode)) &&
        (!S_ISLNK(dentry->d_inode->i_mode)))
        goto DONE;

    if (0 != file_acl_check(dentry, mnt, TRUE, FALSE, __SUBTYPE_DELETEFILE)) {
        retval = -EACCES;
        goto DONE;
    }
DONE:
    return retval;
}


int nfs_truncate_check(struct dentry *dentry, struct vfsmount *mnt)
{
    int retval = 0;

    if (!mnt || !dentry || !dentry->d_inode)
        goto DONE;

    if ((!S_ISREG(dentry->d_inode->i_mode)) &&
        (!S_ISDIR(dentry->d_inode->i_mode)) &&
        (!S_ISLNK(dentry->d_inode->i_mode)))
        goto DONE;

    if (0 != file_acl_check(dentry, mnt, TRUE, FALSE, __SUBTYPE_WRITEFILE)) {
        retval = -EACCES;
        goto DONE;
    }
DONE:
    return retval;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
int nfs_chown_check(struct dentry *dentry, struct vfsmount *mnt,
                         kuid_t user, kgid_t group)
#else
int nfs_chown_check(struct dentry *dentry, struct vfsmount *mnt,
                         uid_t user, gid_t group)
#endif
{
    int retval = 0;

    if (!mnt || !dentry || !dentry->d_inode)
        goto DONE;

    if ((!S_ISREG(dentry->d_inode->i_mode)) &&
        (!S_ISDIR(dentry->d_inode->i_mode)) &&
        (!S_ISLNK(dentry->d_inode->i_mode)))
        goto DONE;

    if (0 != file_acl_check(dentry, mnt, TRUE, FALSE, __SUBTYPE_WRITEFILE)) {
        retval = -EACCES;
        goto DONE;
    }

DONE:
    return retval;
}

