#ifndef __NFS_SECURITY_OPS_H_
#define __NFS_SECURITY_OPS_H_

#include <project.h>
#include <linux/mount.h>
#include <linux/fs.h>
#include <linux/fs_struct.h>
#include <linux/list.h>
#include <linux/net.h>


struct nfs_security_ops {
    int (*bprm_check)(struct linux_binprm *bprm);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 25)
    int (*chroot_check)(const struct path *path);
    int (*pivot_root_check)(const struct path *old_path,
                                 const struct path *new_path);
    int (*mount_check)(const char *dev_name, const struct path *path,
                            const char *type, unsigned long flags,
                            void *data_page);
#else
    int (*chroot_check)(struct nameidata *nd);
    int (*pivot_root_check)(struct nameidata *old_nd,
                                 struct nameidata *new_nd);
    int (*mount_check)(const char *dev_name, struct nameidata *nd,
                            const char *type, unsigned long flags,
                            void *data_page);
#endif
    int (*umount_check)(struct vfsmount *mnt, int flags);
    _Bool (*lport_reserved)(const u16 port);
   
    void (*save_open_mode)(int mode);
    void (*clear_open_mode)(void);

    int (*open_check)(struct dentry *dentry, struct vfsmount *mnt,
                           const int flags);

    int (*file_alloc_security)(struct file *file);
    void (*file_free_security)(struct file *file);
    int (*file_check)(struct file *file, int mask);

    int (*ptrace_check)(long request, long pid);
    int (*ioctl_check)(struct file *filp, unsigned int cmd,
                            unsigned long arg);
    int (*parse_table)(int __user *name, int nlen, void __user *oldval,
                       void __user *newval, struct ctl_table *table);
    _Bool (*capable)(const u8 operation);
    int (*mknod_check)(struct dentry *dentry, struct vfsmount *mnt,
                            unsigned int mode, unsigned int dev);
    int (*mkdir_check)(struct dentry *dentry, struct vfsmount *mnt,
                            unsigned int mode);
    int (*rmdir_check)(struct dentry *dentry, struct vfsmount *mnt);
    int (*unlink_check)(struct dentry *dentry, struct vfsmount *mnt);
    int (*symlink_check)(struct dentry *dentry, struct vfsmount *mnt,
                              const char *from);
    int (*truncate_check)(struct dentry *dentry, struct vfsmount *mnt);
    int (*rename_check)(struct dentry *old_dentry,
                             struct dentry *new_dentry, struct vfsmount *mnt);
    int (*link_check)(struct dentry *old_dentry, struct dentry *new_dentry,
                           struct vfsmount *mnt);

    int (*fcntl_check)(struct file *file, unsigned int cmd,
                            unsigned long arg);
    int (*kill_check)(pid_t pid, int sig);
    int (*tgkill_check)(pid_t tgid, pid_t pid, int sig);
    int (*tkill_check)(pid_t pid, int sig);
    int (*socket_create_check)(int family, int type, int protocol);
    int (*socket_listen_check)(struct socket *sock);
    int (*socket_connect_check)(struct socket *sock, struct sockaddr *addr,
                                     int addr_len);
    int (*socket_bind_check)(struct socket *sock, struct sockaddr *addr,
                                  int addr_len);
    int (*socket_post_accept_check)(struct socket *sock,
                                         struct socket *newsock);
    int (*socket_sendmsg_check)(struct socket *sock, struct msghdr *msg,
                                     int size);
    int (*socket_post_recvmsg_check)(struct sock *sk, struct sk_buff *skb,
                                          int flags);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
    int (*chown_check)(struct dentry *dentry, struct vfsmount *mnt,
                            kuid_t user, kgid_t group);
#else
    int (*chown_check)(struct dentry *dentry, struct vfsmount *mnt,
                            uid_t user, gid_t group);
#endif
    int (*chmod_check)(struct dentry *dentry, struct vfsmount *mnt,
                            mode_t mode);
    int (*getattr_check)(struct vfsmount *mnt, struct dentry *dentry);
    int (*sigqueue_check)(pid_t pid, int sig);
    int (*tgsigqueue_check)(pid_t tgid, pid_t pid, int sig);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
    int (*search_binary_handler)(struct linux_binprm *bprm);
#else
    int (*search_binary_handler)(struct linux_binprm *bprm,
                                 struct pt_regs *regs);
#endif
#ifdef CONFIG_CCSECURITY_USE_EXTERNAL_TASK_SECURITY
    int (*alloc_task_security)(const struct task_struct *task);
    void (*free_task_security)(const struct task_struct *task);
#endif
    int (*task_create)(unsigned long clone_flags);

    _Bool disabled;
};

extern struct nfs_security_ops g_nfs_security_ops;

#define NFS_SECURITY_OPS(func, args...) \
    ({ \
        int r; \
        r = g_nfs_security_ops.func ? g_nfs_security_ops.func(args) : 0; \
        if (r) \
            tag_pr_debug("securit_ops[%s] return [%d]\n", #func, r); \
        r; \
    })

// init for g_nfs_security_ops
void nfs_security_ops_init(void);

#endif