#include <linux/module.h>
#include <linux/syscalls.h>
#include <project.h>
#include "nfs_syscall_hook.h"
#include "common/syscall/pt_regs.h"
#include "common/syscall/syscall_hook.h"
#include "common/config/sec_ply_if.h"
#include "common/kmod_check_table.h"
#include "common/audit/sec_audit.h"

asmlinkage long (*__original_sys_delete_module)(const struct pt_regs *);

// hook状态 0/1 未hook/ 已hook
static int syscall_hook_st = 0;

extern struct delayed_work my_work;

asmlinkage long hook_sys_delete_module(const struct pt_regs *regs)
{
    long retval = 0;
    int ret = 0;
    char name[MODULE_NAME_LEN] = {0};
    const char *module_name = (char *)PT_REGS_PARM1(regs);
    // unsigned int flags = PT_REGS_PARM2(regs);

    KMOD_CHECK_RESULT result;

    sec_ply_if_t * ply = NULL;

    ret = copy_from_user(name, module_name, MODULE_NAME_LEN - 1);
    if (ret != 0) {
        retval = __original_sys_delete_module(regs);
        goto DONE;
    }

    ply = get_ply_if(SEC_PLY_PROTECT_KMOD);
    if (!ply || is_sec_ply_empty(ply)) {
        retval = __original_sys_delete_module(regs);
        goto DONE;
    }

    memset(&result, 0, sizeof(result));

    mutex_lock(&ply->sec_ply_lock);
    ret = kmod_check(ply->ply_head, name, &result);
    mutex_unlock(&ply->sec_ply_lock);

    // 找到策略，且开控制，则禁止卸载模块
    if (result.checkFlags[__POLICY_TYPE_KMOD] && ply->sw.b_ctrl) {
        retval = -1;
        goto DONE;
    }

// 内核模块时，需要禁止卸载当前模块，以免出现问题
// 纯内核层时，不存在这个情况
#ifdef NFSSEC_KMOD
    // 禁止卸载当前模块，因为会导致调用delete_module时，调用到module_exit，导致错误
    if (!strcmp(name, THIS_MODULE->name)) {
        tag_pr_warning(
            "Can't rmmod this module, please turn off kmod protect first!\n");

        retval = 0;
    } else {
        retval = __original_sys_delete_module(regs);
    }
#else
    retval = __original_sys_delete_module(regs);
#endif

DONE:
    // todo 审计条件
    if ((1 == result.checkFlags[__POLICY_TYPE_KMOD]) && ply->sw.b_audit)
        AUDIT_SYS_PROTECT(current->comm, current->pid, 
            AUDIT_OPE_FILE_RMMOD, AUDIT_SUBTYPE_PROC_PROTECT, retval ? AUDIT_RESULT_FAILED : AUDIT_RESULT_SUCCESS, name);

    return retval;
}

int nfs_syscall_hook_init(void)
{
    int retval = -1;
    int ret = 0;

    sec_ply_if_t * ply = NULL;

    ply = get_ply_if(SEC_PLY_PROTECT_KMOD);
    if (!ply) 
        goto DONE;

    // 只在打开模块保护的情况下，才挂该hook
    if (!is_sw_enable(&ply->sw)) {
        tag_pr_debug("protect kmod disable, no need to hook kmod.\n");
        return 0;
    }

    // 防止重复hook
    if (syscall_hook_st == 1) {
        tag_pr_debug("protect kmod hooked, no need to multi.\n");
        return 0;
    }

    ret = syscall_hook_init();
    if (ret != 0) {
        tag_pr_err("syscall_hook_init err.\n");
        goto DONE;
    }

    ret = hook_replace_system_call(
        __NR_delete_module, (unsigned long *)hook_sys_delete_module,
        (unsigned long **)&__original_sys_delete_module);
    if (ret != 0) {
        tag_pr_err("hook_replace_system_call[%s] err.\n", "sys_delete_module");
        goto DONE;
    }

    syscall_hook_st = 1;

    retval = 0;
DONE:
    return retval;
}

void nfs_syscall_hook_exit(void)
{
    int ret = 0;

    if (syscall_hook_st == 0)
        return;

    if (__original_sys_delete_module != NULL) {
        ret = hook_restore_system_call(
            __NR_delete_module, (unsigned long *)hook_sys_delete_module,
            (unsigned long **)&__original_sys_delete_module);
        if (ret != 0)
            tag_pr_err("hook_restore_system_call[%s] err.\n",
                       "sys_delete_module");
    }

    syscall_hook_exit();

    syscall_hook_st = 0;
    
    return;
}
